﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;

namespace PAT.Common.Classes.DataStructure
{
    /// <summary>
    /// This class manipulates difference bound matrix for explicit timing requirements.
    /// </summary>
    public class FullDBM
    {
        public static int[] ClockMaxValues;
        public static int[] ClockLowerValues;
        public static int[] ClockUpperValues;
        
        public bool IsEmpty
        {
            get
            {
                return dimention == 1;
            }
        }

        //matrix to store the 
        public int[] Matrix;
        public bool[] MatrixStrictness;

        //if FullDBM is empty, it is in connoicalform
        public bool IsCanonicalForm = true;
        public int dimention; 

        public FullDBM(int dim)
        {
            dimention = dim;
            Matrix = new int[dimention*dimention];
            MatrixStrictness = new bool[Matrix.Length];
        }

        public FullDBM(int dim, int[] matrix, bool[] matrixStrictness, bool isCanonical)
        {
            dimention = dim;
            Matrix = matrix;
            MatrixStrictness = matrixStrictness;
            IsCanonicalForm = isCanonical;
        }

        public int GetTimerUpper(short timerID)
        {
            Debug.Assert(IsCanonicalForm);
            return Matrix[timerID * dimention];
        }

        public void ResetTimer(short timerid)
        {
            Debug.Assert(IsCanonicalForm);

            Matrix[timerid * dimention + 0] = 0;
            Matrix[0 * dimention + timerid] = 0;
            MatrixStrictness[timerid * dimention + 0] = false;
            MatrixStrictness[0 * dimention + timerid] = false;

            for (int i = 1; i < dimention; i++)
            {
                Matrix[i * dimention + timerid] = Matrix[i * dimention + 0];
                Matrix[timerid * dimention + i] = Matrix[0 * dimention + i];

                MatrixStrictness[i * dimention + timerid] = MatrixStrictness[i * dimention + 0];
                MatrixStrictness[timerid * dimention + i] = MatrixStrictness[0 * dimention + i];
            }
        }

        private void GetCanonicalForm()
        {
            //int dimention = Matrix.Count;
            for (int k = 0; k < dimention; k++)
            {
                for (int i = 0; i < dimention; i++)
                {
                    if (i != k)
                    {
                        for (int j = 0; j < dimention; j++)
                        {
                            //check for the overflow problem
                            if (Matrix[i * dimention + k] != int.MaxValue && Matrix[k * dimention + j] != int.MaxValue)
                            {
                                //Attension, Matrix[i* dimention + k] + Matrix[k* dimention + j] is bigger than int.MaxValue, there is a possbility of overflow.
                                //Matrix[i* dimention + j] = Math.Min(Matrix[i* dimention + j], Matrix[i* dimention + k] + Matrix[k* dimention + j]);

                                if (Matrix[i * dimention + j] > Matrix[i * dimention + k] + Matrix[k * dimention + j])
                                {
                                    Matrix[i * dimention + j] = Matrix[i * dimention + k] + Matrix[k * dimention + j];
                                    MatrixStrictness[i * dimention + j] = MatrixStrictness[i * dimention + k] || MatrixStrictness[k * dimention + j];
                                }
                                else if (Matrix[i * dimention + j] == Matrix[i * dimention + k] + Matrix[k * dimention + j])
                                {
                                    MatrixStrictness[i * dimention + j] = MatrixStrictness[i * dimention + j] || (MatrixStrictness[i * dimention + k] || MatrixStrictness[k * dimention + j]);
                                }
                            }
                        }

                        if (Matrix[i * dimention + i] < 0 || MatrixStrictness[0 * dimention + 0])
                        {
                            IsCanonicalForm = true;
                            Matrix[0 * dimention + 0] = -1;
                            return;
                        }
                    }
                }
            }

            if (!SpecificationBase.IsSimulation && Matrix[0] >= 0)
            {
                for (int i = 0; i < dimention; i++)
                {
                    for (int j = 0; j < dimention; j++)
                    {
                        //=======================classic extrapolation======================
                        if (Matrix[i * dimention + j] > ClockMaxValues[i]) //i > 0 && ClockMaxValues[i] > 0 && 
                        {
                            //Debug.Assert(Matrix[i* dimention + j] <= ClockMaxValues[i - 1]);
                            Matrix[i * dimention + j] = int.MaxValue;                            
                        }

                        if (-1 * Matrix[i * dimention + j] > ClockMaxValues[j]) //j > 0 && ClockMaxValues[j] > 0 &&  
                        {
                            //Debug.Assert(-1*Matrix[i* dimention + j] <= ClockMaxValues[j - 1]);

                            Matrix[i * dimention + j] = -1 * ClockMaxValues[j];
                            MatrixStrictness[i * dimention + j] = true;
                        }
                        //=======================classic extrapolation======================


                        ////=======================diaganol extrapolation======================
                        if (-1 * Matrix[0 * dimention + i] > ClockMaxValues[i]) //i > 0 && ClockMaxValues[i] > 0 &&  
                        {
                            //Debug.Assert(-1 * Matrix[0* dimention + i] <= ClockMaxValues[i - 1]);
                            Matrix[i* dimention + j] = int.MaxValue;
                        }

                        if (i != 0 && -1 * Matrix[0 * dimention + j] > ClockMaxValues[j]) //j > 0 && ClockMaxValues[j] > 0 && 
                        {
                            //Debug.Assert(-1 * Matrix[0* dimention + j] <= ClockMaxValues[j - 1]);
                            Matrix[i * dimention + j] = int.MaxValue;
                        }

                        if (i == 0 && -1 * Matrix[i * dimention + j] > ClockMaxValues[j]) //j > 0 && ClockMaxValues[j] > 0 && 
                        {
                            //Debug.Assert(-1 * Matrix[i* dimention + j] <= ClockMaxValues[j - 1]);
                            Matrix[i * dimention + j] = -1 * ClockMaxValues[i];
                            MatrixStrictness[i * dimention + j] = true;
                        }
                        ////=======================diaganol extrapolation======================

                        ////=======================L/U extrapolation======================
                        if (Matrix[i * dimention + j] > ClockLowerValues[i]) //i > 0 && ClockLowerValues[i] > 0 && 
                        {
                            //Debug.Assert(Matrix[i* dimention + j] <= ClockLowerValues[i - 1]);
                            Matrix[i * dimention + j] = int.MaxValue;
                        }

                        if (-1 * Matrix[i * dimention + j] > ClockUpperValues[j]) //j > 0 && ClockUpperValues[j] > 0 && 
                        {
                            //Debug.Assert(-1*Matrix[i* dimention + j] <= ClockUpperValues[j - 1]);
                            Matrix[i * dimention + j] = -1 * ClockUpperValues[j];
                            MatrixStrictness[i * dimention + j] = true;
                        }
                        ////=======================L/U extrapolation======================

                        ////=======================diaganol L/U extrapolation======================
                        if (-1 * Matrix[0 * dimention + i] > ClockLowerValues[i]) //i > 0 && ClockLowerValues[i] > 0 && 
                        {
                            //Debug.Assert(-1 * Matrix[0* dimention + i] <= ClockLowerValues[i - 1]);
                            Matrix[i * dimention + j] = int.MaxValue;
                        }

                        if (i != 0 && -1 * Matrix[0 * dimention + j] > ClockUpperValues[j]) //j > 0 && ClockUpperValues[j] > 0 && 
                        {
                            //Debug.Assert(-1 * Matrix[0* dimention + j] <= ClockUpperValues[j - 1]);
                            Matrix[i * dimention + j] = int.MaxValue;
                        }

                        if (i == 0 && -1 * Matrix[i * dimention + j] > ClockUpperValues[j]) //j > 0 && ClockUpperValues[j] > 0 &&  
                        {
                            //Debug.Assert(-1 * Matrix[i* dimention + j] <= ClockUpperValues[j - 1]);
                            Matrix[i * dimention + j] = -1 * ClockUpperValues[j];
                            MatrixStrictness[i * dimention + j] = true;
                        }
                        ////=======================diaganol L/U extrapolation======================
                    }
                }
            }

            IsCanonicalForm = true;
        }

        /// <summary>
        /// Update the FullDBM with a new constraint
        /// </summary>
        /// <param name="timer">which timer the constraint is on</param>
        /// <param name="op">0 for equal; 1 for >=; -1 for <=</param>
        /// <param name="constant"></param>
        public void AddConstraint(short timer, TimerOperationType op, int constant)
        {
            Debug.Assert(timer > 0);

            switch (op)
            {
                case TimerOperationType.Equals:
                    if (Matrix[timer * dimention + 0] > constant)
                    {
                        Matrix[timer * dimention + 0] = constant;
                    }
                    if (Matrix[0 * dimention + timer] > -1 * constant)
                    {
                        Matrix[0 * dimention + timer] = -1 * constant;
                    }
                    break;
                case TimerOperationType.GreaterThanOrEqualTo:
                    if (Matrix[0 * dimention + timer] > -1 * constant)
                    {
                        Matrix[0 * dimention + timer] = -1 * constant;
                    }
                    break;
                case TimerOperationType.LessThanOrEqualTo:
                    if (Matrix[timer * dimention + 0] > constant)
                    {
                        Matrix[timer * dimention + 0] = constant;
                    }
                    break;
                case TimerOperationType.GreaterThan:
                    if (Matrix[0 * dimention + timer] > -1 * constant)
                    {
                        Matrix[0 * dimention + timer] = -1 * constant;
                        MatrixStrictness[0 * dimention + timer] = true;
                    }
                    break;
                case TimerOperationType.LessThan:
                    if (Matrix[timer * dimention + 0] > constant)
                    {
                        Matrix[timer * dimention + 0] = constant;
                        MatrixStrictness[timer* dimention + 0] = true;
                    }
                    break;
            }

            IsCanonicalForm = false;
        }

        public void Delay()
        {
            Debug.Assert(IsCanonicalForm); /* chengbin */
            
            for (int i = 0; i < dimention; i++)
            {

                if(i != 0)
                {
                    Matrix[i*dimention + 0] = int.MaxValue;
                    MatrixStrictness[i * dimention + 0] = false;
                }
                else
                {
                    Matrix[i * dimention + 0] = 0;
                    MatrixStrictness[i * dimention + 0] = false;
                }
            }           
        }

        public FullDBM KeepTimers(List<byte> activeTimer)
        {
            Debug.Assert(this.IsCanonicalForm); /* chengbin */

            int[] newMatrix = new int[Matrix.Length];
            bool[] newMatrixStrictness = new bool[Matrix.Length];


            foreach (int i in activeTimer)
            {
                foreach (int j in activeTimer)
                {
                    newMatrix[i * dimention + j] = Matrix[i * dimention + j];
                    newMatrixStrictness[i * dimention + j] = MatrixStrictness[i * dimention + j];
                }
            }

            return new FullDBM(dimention, newMatrix, newMatrixStrictness, true);
        }

        public bool IsConstraintNotSatisfied()
        {
            if (IsEmpty)
            {
                return false;
            }

            if (!IsCanonicalForm)
            {
                GetCanonicalForm();
            }

            return Matrix[0* dimention + 0] < 0 || MatrixStrictness[0* dimention + 0];
        }

        public void Conjunction(FullDBM myDBM)
        {
            if (myDBM.IsEmpty)
            {
                return;
            }

            if (IsEmpty)
            {
                Matrix = myDBM.Matrix;
                MatrixStrictness = myDBM.MatrixStrictness;
                IsCanonicalForm = myDBM.IsCanonicalForm;
            }
            else if (!myDBM.IsEmpty)
            {
                //assert the dimention of the two dbm are same
                for (int i = 0; i < dimention; i++)
                {
                    for (int j = 0; j < dimention; j++)
                    {
                        if (Matrix[i* dimention + j] > myDBM.Matrix[i* dimention + j])
                        {
                            Matrix[i* dimention + j] = myDBM.Matrix[i* dimention + j];
                            MatrixStrictness[i* dimention + j] = myDBM.MatrixStrictness[i* dimention + j];
                        }
                        else if (Matrix[i* dimention + j] == myDBM.Matrix[i* dimention + j])
                        {
                            MatrixStrictness[i* dimention + j] = MatrixStrictness[i* dimention + j] || myDBM.MatrixStrictness[i* dimention + j];
                        }
                    }
                }

                IsCanonicalForm = false;                
            }
        }

        public String ToString(Dictionary<string, short> clockMapping)
        {
            if (IsEmpty)
            {
                return "";
            }

            StringBuilder sb = new StringBuilder();

            foreach (KeyValuePair<string, short> pair in clockMapping)
            {
                sb.AppendLine(pair.Key + (MatrixStrictness[0* dimention + pair.Value] ? ":(" : ":[") +
                              (Matrix[0* dimention + pair.Value] == int.MaxValue ? "-" + Constants.INFINITE : (Matrix[0* dimention + pair.Value] * -1).ToString()) +
                              "," + (Matrix[pair.Value* dimention + 0] == int.MaxValue ? Constants.INFINITE : Matrix[pair.Value* dimention + 0].ToString()) +
                              (MatrixStrictness[pair.Value* dimention + 0] ? ");" : "];"));
            }


            return sb.ToString();
        }

        public override String ToString()
        {
            if (IsEmpty)
            {
                return "";
            }

            StringBuilder sb = new StringBuilder();
            
            for (int i = 1; i < dimention; i++)
            {
                sb.AppendLine("clock" + i + (MatrixStrictness[0* dimention + i] ? ":(" : ":[") + (Matrix[0* dimention + i] == int.MaxValue ? "-" + Constants.INFINITE : (Matrix[0* dimention + i]*-1).ToString()) + "," + (Matrix[i* dimention + 0] == int.MaxValue ? Constants.INFINITE : Matrix[i* dimention + 0].ToString()) + (MatrixStrictness[i* dimention + 0] ? ");" : "];"));                
            }

            return sb.ToString();
        }

        public FullDBM Clone()
        {
            if (IsEmpty)
            {
                return this;
            }

            return new FullDBM(dimention, (int[])Matrix.Clone(), (bool[])MatrixStrictness.Clone(), IsCanonicalForm);
        }

        public string GetID()
        {
            if (IsEmpty)
            {
                return "";
            }

            string toReturn = "";

            for (int i = 1; i < dimention; i++)
            {
                toReturn += (MatrixStrictness[0* dimention + i] ? "(" : "[") + Matrix[0* dimention + i] * -1 + "," + Matrix[i* dimention + 0] + (MatrixStrictness[i* dimention + 0] ? ")" : "]");
            }

            return toReturn;
        }
    }
}
