using System.Collections.Generic;


namespace PAT.Common.Classes.Expressions.ExpressionClass
{
    public sealed class Assignment : Expression
    {
        public string LeftHandSide;
        public Expression RightHandSide;
        
        public Assignment(string l, Expression r)
        {
            LeftHandSide = l;
            RightHandSide = r;
            ExpressionType = ExpressionType.Assignment;
            HasVar = true;

            expressionID = LeftHandSide + "=" + RightHandSide.ExpressionID;
        }

        public override string ToString()
        {
            return LeftHandSide + " = " + RightHandSide.ToString() + ";";
        }

        public override List<string> GetVars()
        {
            List<string> vars = this.RightHandSide.GetVars();
            if (!vars.Contains(LeftHandSide))
            {
                vars.Add(LeftHandSide);
            }

            return vars;
        }

        public override Expression ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new Assignment((constMapping.ContainsKey(LeftHandSide)? constMapping[LeftHandSide].ExpressionID: LeftHandSide), RightHandSide.ClearConstant(constMapping));
        }

        /// <summary>
        /// a = 1; a = a + 1;
        /// </summary>
        public static bool IsComplexUpdate(List<int> updatedVariablesBefore, List<int> updatedVariablesInNewUpdate, List<int> usedVariableInNewUpdate)
        {
            return (updatedVariablesInNewUpdate.Count > 0 && usedVariableInNewUpdate.Contains(updatedVariablesInNewUpdate[0]) &&
                        updatedVariablesBefore.Contains(updatedVariablesInNewUpdate[0]));
        }



        public override bool HasExternalLibraryCall() 
        {
            return this.RightHandSide.HasExternalLibraryCall();
        }
    }
}