using System;
using System.Collections.Generic;

namespace PAT.Common.Classes.Expressions.ExpressionClass
{
    public sealed class If : Expression
    {    
        public Expression Condition, ThenPart, ElsePart;

        public If(Expression c, Expression t, Expression e)
        {
            Condition = c;
            ThenPart = t;
            ElsePart = e;
            
            if (e != null)
            {
                this.HasVar = c.HasVar || t.HasVar || e.HasVar;

                expressionID = "I@" + Condition.ExpressionID + "{" + ThenPart.ExpressionID + "^" + ElsePart.ExpressionID + "}";
            }
            else
            {
                this.HasVar = c.HasVar || t.HasVar;
                expressionID = "I@" + Condition.ExpressionID + "{" + ThenPart.ExpressionID + "}";
            }

            ExpressionType = ExpressionType.If;
        }

        public override String ToString()
        {
            if (ElsePart != null)
            {
                return " if (" + Condition + ") {" + ThenPart + "} else {" + ElsePart + "}";
            }
            else
            {
                return " if (" + Condition + ") {" + ThenPart + "}";
            }
        }

        //public override String GetID()
        //{
        //    if (ElsePart != null)
        //    {
        //        return "I@" + Condition.GetID() + "{" + ThenPart.GetID() + "^" + ElsePart.GetID() + "}";
        //    }
        //    else
        //    {
        //        return "I@" + Condition.GetID() + "{" + ThenPart.GetID() + "}";
        //    }
        //}

        public override List<string> GetVars()
        {
            List<string> vars = this.Condition.GetVars();
            Ultility.Ultility.AddList(vars, this.ThenPart.GetVars());
            if (ElsePart != null)
            {
                Ultility.Ultility.AddList(vars, this.ElsePart.GetVars());
            }
            return vars;
        }

        public override Expression ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new If(Condition.ClearConstant(constMapping), ThenPart.ClearConstant(constMapping), ElsePart == null ? null : ElsePart.ClearConstant(constMapping));
        }


        public override bool HasExternalLibraryCall()
        {
            if (ElsePart != null)
            {
                return Condition.HasExternalLibraryCall() || ThenPart.HasExternalLibraryCall() ||
                       ElsePart.HasExternalLibraryCall();
            }
            else
            {
                return Condition.HasExternalLibraryCall() || ThenPart.HasExternalLibraryCall();
            }
        }
    }
}