using System;
using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.Ultility;

namespace PAT.Common.Classes.Expressions.ExpressionClass
{
    public sealed class StaticMethodCall : Expression
    {
        public string MethodName;
        public Expression[] Arguments;

        public StaticMethodCall(String method, Expression[] args)
        {
            ExpressionType = ExpressionType.StaticMethodCall;
            MethodName = method;
            Arguments = args;
            StringBuilder sb = new StringBuilder("C@" + MethodName + "(");
            foreach (Expression exp in Arguments)
            {
                HasVar = HasVar || exp.HasVar;
                sb.Append(exp.ExpressionID + ",");
            }
            expressionID = sb.ToString();
        }

        public override String ToString()
        {
            if (Arguments.Length > 0)
            {
                StringBuilder sb = new StringBuilder("call(" + MethodName + ",");
                for (int i = 0; i < Arguments.Length; i++)
                {
                    if (i == Arguments.Length - 1)
                    {
                        sb.Append(Arguments[i].ToString());
                    }
                    else
                    {
                        sb.Append(Arguments[i].ToString() + ", ");
                    }
                }
                sb.Append(")");
                return sb.ToString();
            }
            else
            {
                return "call(" + MethodName + ")";
            }
        }
        
        //public override String GetID()
        //{
        //    StringBuilder sb = new StringBuilder("C@" + MethodName + "(");
        //    for (int i = 0; i < Arguments.Length; i++)
        //    {
        //        if (i == Arguments.Length - 1)
        //        {
        //            sb.Append(Arguments[i].GetID());
        //        }
        //        else
        //        {
        //            sb.Append(Arguments[i].GetID() + ",");
        //        }
        //    }
        //    sb.Append(")");
        //    return sb.ToString();
        //}

        public override List<string> GetVars()
        {
            List<string> vars = new List<string>(16);
            foreach (Expression exp in Arguments)
            {
                Ultility.Ultility.AddList(vars, exp.GetVars());
            }
            return vars;
        }

        public override Expression ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Expression[] newArgs = new Expression[Arguments.Length];

            for (int i = 0; i < Arguments.Length; i++)
            {
                newArgs[i] = Arguments[i].ClearConstant(constMapping);
            }

            return new StaticMethodCall(MethodName, newArgs);
        }

        public override bool HasExternalLibraryCall()
        {
            //BDD can support call to channel properties
            return (MethodName != Constants.cfull && MethodName != Constants.cempty && MethodName != Constants.ccount
                    && MethodName != Constants.csize && MethodName != Constants.cpeek);
        }

    }
}