using System;
using System.Collections.Generic;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.Ultility;
using PAT.PAR.LTS;
using PAT.PAR.Ultility;

namespace PAT.PAR.Assertions
{
    public class PARAssertionDivergenceTime : AssertionDivergence
    {
        private DefinitionRef Process;

        public PARAssertionDivergenceTime(DefinitionRef processDef)
            : base()
        {
            Process = processDef;
        }

        public override void Initialize(SpecificationBase spec)
        {
            PARAssertion.Initialize(this, Process, spec);
        }

        public override string ToString()
        {
            return StartingProcess + " divergencefree <T>";
        }

        public override void RunVerification()
        {
            if (!Process.IsTime())
            {
                VerificationOutput.CounterExampleTrace = null;
                VerificationOutput.VerificationResult = VerificationResultType.VALID;

                return;
            }

            if (ShortestPath)
            {
                BFSVerification();
            }
            else
            {
                DFSVerification();
            }
        }



        private void DFSVerification()
        {
            StringDictionary<List<Constraint>> Visited = new StringDictionary<List<Constraint>>(Common.Classes.Ultility.Ultility.MC_INITIAL_SIZE);
            StringHashTable DivChecked = new StringHashTable(Common.Classes.Ultility.Ultility.MC_INITIAL_SIZE);

            Stack<PARConfiguration> pendingImpl = new Stack<PARConfiguration>(1000);

            //The following are for identifying a counterexample trace. 
            Stack<int> depthStack = new Stack<int>(1000);
            depthStack.Push(0);
            List<int> depthList = new List<int>(1000);
            //The above are for identifying a counterexample trace. 

            //implementation initial state
            PARConfiguration currentImpl = InitialStep as PARConfiguration;
            pendingImpl.Push(currentImpl);

            PARUltility.Visit(Visited, currentImpl);

            while (pendingImpl.Count > 0)
            {
                if (CancelRequested)
                {
                    VerificationOutput.NoOfStates = Visited.Count;
                    return;
                }

                currentImpl = pendingImpl.Pop();

                //The following are for identifying a counterexample trace. 
                int depth = depthStack.Pop();

                while (depth > 0 && depthList[depthList.Count - 1] >= depth)
                {
                    int lastIndex = depthList.Count - 1;
                    depthList.RemoveAt(lastIndex);
                    VerificationOutput.CounterExampleTrace.RemoveAt(lastIndex);
                }

                VerificationOutput.CounterExampleTrace.Add(currentImpl);
                depthList.Add(depth);

                if (IsDivergent(DivChecked, currentImpl))
                {
                    VerificationOutput.VerificationResult = VerificationResultType.INVALID;
                    VerificationOutput.NoOfStates = Visited.Count;
                    return;
                }

                List<PARConfiguration> list = currentImpl.MakeOneMoveLocal();

                VerificationOutput.Transitions += list.Count;

                for (int i = list.Count - 1; i >= 0; i--)
                {
                    PARConfiguration step = list[i];
                    if (!PARUltility.Visit(Visited, step))
                    {
                        pendingImpl.Push(step);
                        depthStack.Push(depth + 1);
                    }
                }
            }

            VerificationOutput.CounterExampleTrace = null;
            VerificationOutput.VerificationResult = VerificationResultType.VALID;
            VerificationOutput.NoOfStates = Visited.Count;
        }




        private void BFSVerification()
        {
            StringDictionary<List<Constraint>> Visited = new StringDictionary<List<Constraint>>(Common.Classes.Ultility.Ultility.MC_INITIAL_SIZE);
            StringHashTable DivChecked = new StringHashTable(Common.Classes.Ultility.Ultility.MC_INITIAL_SIZE);

            Queue<PARConfiguration> working = new Queue<PARConfiguration>(1024);
            Queue<List<ConfigurationBase>> paths = new Queue<List<ConfigurationBase>>(1024);

            PARUltility.Visit(Visited, InitialStep as PARConfiguration);

            working.Enqueue(InitialStep as PARConfiguration);
            List<ConfigurationBase> path = new List<ConfigurationBase>();
            path.Add(InitialStep);
            paths.Enqueue(path);

            do
            {
                if (CancelRequested)
                {
                    VerificationOutput.NoOfStates = Visited.Count;
                    return;
                }

                PARConfiguration current = working.Dequeue();
                List<ConfigurationBase> currentPath = paths.Dequeue();
                List<PARConfiguration> list = current.MakeOneMoveLocal();

                VerificationOutput.Transitions += list.Count;

                if (IsDivergent(DivChecked, current))
                {
                    VerificationOutput.VerificationResult = VerificationResultType.INVALID;
                    VerificationOutput.NoOfStates = Visited.Count;
                    VerificationOutput.CounterExampleTrace = currentPath;
                    return;
                }

                for (int i = list.Count - 1; i >= 0; i--)
                {
                    PARConfiguration step = list[i];

                    //if (!Visited.ContainsKey(stepID))
                    if (!PARUltility.Visit(Visited, step))
                    {
                        working.Enqueue(step);

                        List<ConfigurationBase> newPath = new List<ConfigurationBase>(currentPath);
                        newPath.Add(step);
                        paths.Enqueue(newPath);
                    }
                }
            } while (working.Count > 0);

            if (MustAbstract)
            {
                VerificationOutput.VerificationResult = VerificationResultType.UNKNOWN;
            }
            else
            {
                VerificationOutput.VerificationResult = VerificationResultType.VALID;
            }

            VerificationOutput.NoOfStates = Visited.Count;
            return;
        }


        /// <summary>
        /// To get all states which can be reached by tau-VerificationOutput.Transitions only; meanwhile, identify which of them are divergent states.
        /// </summary>
        /// <returns></returns>
        private static bool IsDivergent(StringHashTable Visited, PARConfiguration States)
        {
            string ID = States.GetID();

            if (Visited.ContainsKey(ID))
            {
                return false;
            }

            Dictionary<string, int> visited = new Dictionary<string, int>();
            Dictionary<int, List<int>> Transitions = new Dictionary<int, List<int>>();
            Dictionary<string, bool> scc = new Dictionary<string, bool>();

            Stack<PARConfiguration> TaskStack = new Stack<PARConfiguration>();

            TaskStack.Push(States);
            visited.Add(ID, 0);
            Transitions.Add(0, new List<int>());

            Dictionary<int, int> preorder = new Dictionary<int, int>();
            Dictionary<int, int> lowlink = new Dictionary<int, int>();
            Stack<PARConfiguration> stepStack = new Stack<PARConfiguration>();
            List<int> scc_found = new List<int>();
            //# Preorder counter 
            int ii = 0;

            //store the expended event step of a node to avoid multiple invocation of the make one move.
            Dictionary<int, List<PARConfiguration>> ExpendedNode = new Dictionary<int, List<PARConfiguration>>();

            do
            {
                PARConfiguration pair = TaskStack.Peek();
                ID = pair.GetID();
                if (!Visited.ContainsKey(ID))
                {
                    int v = visited[ID];

                    List<int> outgoing = Transitions[v];

                    if (!preorder.ContainsKey(v))
                    {
                        preorder.Add(v, ii);
                        ii++;
                    }

                    bool done = true;

                    if (ExpendedNode.ContainsKey(v))
                    {
                        List<PARConfiguration> list = ExpendedNode[v];
                        if (list.Count > 0)
                        {
                            //transverse all steps
                            for (int k = list.Count - 1; k >= 0; k--)
                            {
                                PARConfiguration step = list[k];

                                if (step != null && step.Event == Constants.TAU)
                                {
                                    int t = visited[step.GetID()];

                                    //if the step is a unvisited step
                                    if (!preorder.ContainsKey(t))
                                    {
                                        //only add the first unvisited step
                                        //for the second or more unvisited steps, ignore at the monent
                                        if (done)
                                        {
                                            TaskStack.Push(step);

                                            //procPath.Add(step.GetID());

                                            done = false;
                                            //list.RemoveAt(k);
                                            list[k] = null;
                                        }
                                    }
                                    else
                                    {
                                        //list.RemoveAt(k);
                                        list[k] = null;
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        //List<RTSConfiguration> product = pair.MakeOneMoveLocal(Constants.TAU); 
                        List<PARConfiguration> product = pair.MakeOneMoveLocal();

                        for (int k = product.Count - 1; k >= 0; k--)
                        {
                            PARConfiguration step = product[k];

                            if (step.Event == Constants.TAU)
                            {
                                string stateString = step.GetID();
                                if (visited.ContainsKey(stateString))
                                {
                                    int t = visited[stateString];
                                    outgoing.Add(t);

                                    //if this node is still not visited
                                    if (!preorder.ContainsKey(t))
                                    {
                                        //only put the first one to the work list stack.
                                        //if there are more than one node to be visited, 
                                        //simply ignore them and keep its event step in the list.
                                        if (done)
                                        {
                                            TaskStack.Push(step);
                                            //procPath.Add(step.GetID());
                                            done = false;
                                            //product.RemoveAt(k);
                                            product[k] = null;
                                        }
                                        else
                                        {
                                            product[k] = step;
                                        }
                                    }
                                    //this node is truly visited. can be removed
                                    else
                                    {
                                        //product.RemoveAt(k);
                                        product[k] = null;
                                    }
                                }
                                else
                                {
                                    int stateID = visited.Count;
                                    visited.Add(stateString, stateID);
                                    //reachable.Add(step);
                                    //if (!isDiv.ContainsKey(step.GetID()))
                                    //{
                                    //    isDiv.Add(step.GetID(), false);
                                    //}

                                    Transitions.Add(stateID, new List<int>(8));
                                    outgoing.Add(stateID);
                                    //only put the first one into the stack.
                                    if (done)
                                    {
                                        TaskStack.Push(step);

                                        //procPath.Add(step.GetID());

                                        done = false;
                                        //product.RemoveAt(k);
                                        product[k] = null;
                                    }
                                    else
                                    {
                                        product[k] = step;
                                    }
                                }
                            }
                        }

                        //create the remaining steps as the expending list for v
                        ExpendedNode.Add(v, product);
                    }

                    if (done)
                    {
                        lowlink[v] = preorder[v];

                        bool selfLoop = false;
                        for (int j = 0; j < outgoing.Count; j++)
                        {
                            int w = outgoing[j];
                            if (w == v)
                            {
                                selfLoop = true;
                            }

                            if (!scc_found.Contains(w) && preorder.ContainsKey(w))
                            {
                                if (preorder[w] > preorder[v])
                                {
                                    lowlink[v] = Math.Min(lowlink[v], lowlink[w]);
                                }
                                else
                                {
                                    lowlink[v] = Math.Min(lowlink[v], preorder[w]);
                                }
                            }
                        }

                        TaskStack.Pop();
                        //procPath.RemoveAt(procPath.Count - 1);

                        if (lowlink[v] == preorder[v])
                        {
                            scc.Add(pair.GetID(), pair.IsTimeConsuming);
                            //pair.HidenEvent != null && .HidenEvent.Contains("[") &&
                            //!pair.HidenEvent.Contains("[0"));
                            scc_found.Add(v);

                            //todo: remove the ad hoc implementation of the checking
                            while (stepStack.Count > 0 && preorder[visited[stepStack.Peek().GetID()]] > preorder[v])
                            {
                                PARConfiguration s = stepStack.Pop();
                                string tmp = s.GetID();
                                int k = visited[tmp];
                                scc.Add(tmp, s.IsTimeConsuming);
                                //s.HidenEvent != null && s.HidenEvent.Contains("[") &&
                                //!s.HidenEvent.Contains("[0"));
                                scc_found.Add(k);
                            }

                            //outgoing.Count == 0 --> deadlock, we need to check //outgoing.Count == 0
                            //StronglyConnectedComponets.Count > 1 || selfLoop -> non-trivial case, we need to check
                            if (scc.Count > 1 || selfLoop)
                            {
                                foreach (bool b in scc.Values)
                                {
                                    if (b)
                                    {
                                        return true;
                                    }
                                }
                            }

                            foreach (string componet in scc.Keys)
                            {
                                ExpendedNode.Remove(visited[componet]);
                            }

                            scc.Clear();
                        }
                        else
                        {
                            stepStack.Push(pair);
                        }
                    }
                }
                else
                {
                    TaskStack.Pop();
                }
            } while (TaskStack.Count > 0);

            //return new NormalizedState(reachable);
            foreach (string key in visited.Keys)
            {
                if (!Visited.ContainsKey(key))
                {
                    Visited.Add(key);
                }
            }

            return false;
        }




        public override string StartingProcess
        {
            get
            {
                return Process.ToString();
            }
        }
    }
}