using System;
using System.Collections.Generic;
using PAT.Common.Classes.BA;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.SemanticModels.LTS.Assertion;
using PAT.Common.Classes.Ultility;
using PAT.PAR.LTS;

namespace PAT.PAR.Assertions
{
    public class PARAssertionLTL : AssertionLTL
    {
        private DefinitionRef Process;
        protected bool ApplyZenoCheck;

        public PARAssertionLTL(DefinitionRef processDef, string ltl) : base(ltl)
        {
            Process = processDef;
        }

        //public override void UIInitialize(ISynchronizeInvoke target, FairnessType type, bool por, bool verbose, bool parallel, bool shortestPath, bool useBDD, bool applyZenoCheck, params object[] ExtraOptions)
        //{
        //    ApplyZenoCheck = applyZenoCheck;
        //    base.UIInitialize(target, type, por, verbose, parallel, shortestPath, useBDD, applyZenoCheck, ExtraOptions);
        //}

        public override void Initialize(SpecificationBase spec)
        {


            Specification Spec = spec as Specification;

            List<string> varList = Process.GetGlobalVariables();

            BA.Initialize(Spec.DeclarationDatabase);

            foreach (KeyValuePair<string, Expression> pair in BA.DeclarationDatabase)
            {
                varList.AddRange(pair.Value.GetVars());
            }

            Valuation GlobalEnv = Spec.SpecValuation.GetVariableChannelClone(varList, Process.GetChannels());

            //Initialize InitialStep
            PARConfiguration initialStep = new PARConfiguration(Process, Constants.INITIAL_EVENT, null, GlobalEnv, false, false, Constraint.constraintWithNonNegativeParameters()); // trueConstraint
            
            InitialStep = initialStep;
            (InitialStep as PARConfiguration).SetTimer();

            MustAbstract = Process.MustBeAbstracted();
        }

        public override bool IsProcessLevelFairnessApplicable()
        {
            PARProcess nextProcess = Process.GetTopLevelConcurrency(new List<string>());
            if (MustAbstract)
            {
                if (nextProcess is IndexInterleaveAbstract)
                {
                    IndexInterleaveAbstract interleave = nextProcess as IndexInterleaveAbstract;
                    foreach (PARProcess p in interleave.Processes)
                    {
                        if (p.MustBeAbstracted())
                        {
                            return false;
                        }
                    }

                    return true;
                }
            }
            else
            {
                if (nextProcess is IndexInterleave || nextProcess is IndexParallel || nextProcess is IndexInterleaveAbstract)
                {
                    return true;
                }
            }

            return false;
        }

        public override void RunVerification()
        {
            if (!ApplyZenoCheck)
            {
                base.RunVerification();
                return;
            }

            if (IsSafety)
            {
                RunVerificationSafety();
                return;
            }

            //todo: the LTL-Negate is to be implemented
            if (!IsNegateLiveness && FairnessType == FairnessType.NO_FAIRNESS && !ApplyZenoCheck)
            {
                RunVerificationNegate();
                return;
            }

            //OutgoingTransitionTable.Clear();
            //TaskStack.Clear();

            VerificationOutput.LoopIndex = -1;
            //reset counter example trace to null, because if a spurious counter example is found, we will store the prefix of the SCC to VerificationOutput.CounterExampleTrace
            VerificationOutput.CounterExampleTrace = null;

            TarjanModelChecking();
        }

        protected void TarjanModelChecking()
        {
            Dictionary<string, int[]> DFSData = new Dictionary<string, int[]>();
            List<KeyValuePair<PARConfiguration, string>> initials = new List<KeyValuePair<PARConfiguration, string>>();
            Dictionary<string, List<string>> OutgoingTransitionTable = new Dictionary<string, List<string>>();

            foreach (string s in BA.InitialStates)
            {
                List<string> next = BA.MakeOneMove(s, InitialStep);

                foreach (string var in next)
                {
                    initials.Add(new KeyValuePair<PARConfiguration, string>(InitialStep as PARConfiguration, var));
                }
            }

            Stack<KeyValuePair<PARConfiguration, string>> TaskStack = new Stack<KeyValuePair<PARConfiguration, string>>();

            if (initials.Count == 0)
            {
                VerificationOutput.VerificationResult = VerificationResultType.VALID;
                return;
            }

            for (int z = 0; z < initials.Count; z++)
            {
                KeyValuePair<PARConfiguration, string> initState = initials[z];
                TaskStack.Push(initState);
                string ID = initState.Key.GetIDWithEvent() + Constants.SEPARATOR + initState.Value;
                DFSData.Add(ID, new int[] { VISITED_NOPREORDER, 0 });
                OutgoingTransitionTable.Add(ID, new List<string>(8));
            }

            List<string> StronglyConnectedComponets = new List<string>(1024);
            Stack<KeyValuePair<PARConfiguration, string>> stepStack = new Stack<KeyValuePair<PARConfiguration, string>>(1024);

            //# Preorder counter 
            int i = 0;

            //store the expended event step of a node to avoid multiple invocation of the make one move.
            Dictionary<string, List<KeyValuePair<PARConfiguration, string>>> ExpendedNode = new Dictionary<string, List<KeyValuePair<PARConfiguration, string>>>();

            do
            {
                if (CancelRequested)
                {
                    VerificationOutput.NoOfStates = DFSData.Count;
                    return;
                }

                KeyValuePair<PARConfiguration, string> pair = TaskStack.Peek();
                PARConfiguration evt = pair.Key;
                string BAState = pair.Value;

                string v = pair.Key.GetIDWithEvent() + Constants.SEPARATOR + pair.Value;

                List<string> outgoing = OutgoingTransitionTable[v];

                int[] nodeData = DFSData[v];

                if (nodeData[0] == VISITED_NOPREORDER)
                {
                    nodeData[0] = i;
                    i++;
                }

                bool done = true;

                if (ExpendedNode.ContainsKey(v))
                {
                    List<KeyValuePair<PARConfiguration, string>> list = ExpendedNode[v];
                    if (list.Count > 0)
                    {
                        //transverse all steps
                        for (int k = list.Count - 1; k >= 0; k--)
                        {
                            KeyValuePair<PARConfiguration, string> step = list[k];

                            //int t = step.ID; // StepHashingTable[stateString];

                            //if the step is a unvisited step
                            //if (!preorder.ContainsKey(t))
                            string tmp = step.Key.GetIDWithEvent() + Constants.SEPARATOR + step.Value;
                            if (DFSData[tmp][0] == VISITED_NOPREORDER)
                            {
                                //only add the first unvisited step
                                //for the second or more unvisited steps, ignore at the monent
                                if (done)
                                {
                                    TaskStack.Push(step);

                                    //procPath.Add(step.configuration.GetID());

                                    done = false;
                                    list.RemoveAt(k);
                                }
                            }
                            else
                            {
                                list.RemoveAt(k);
                            }
                        }
                    }
                }
                else
                {
                    List<PARConfiguration> list = evt.MakeOneMoveLocal();

                    List<KeyValuePair<PARConfiguration, string>> product = new List<KeyValuePair<PARConfiguration, string>>();

                    foreach (PARConfiguration rtsConfiguration in list)
                    {
                        List<string> states = BA.MakeOneMove(BAState, rtsConfiguration);

                        for (int j = 0; j < states.Count; j++)
                        {
                            product.Add(new KeyValuePair<PARConfiguration, string>(rtsConfiguration, states[j]));
                        }
                    }

                    //count the transitions visited
                    VerificationOutput.Transitions += product.Count;

                    for (int k = product.Count - 1; k >= 0; k--)
                    {
                        KeyValuePair<PARConfiguration, string> step = product[k];

                        string tmp = step.Key.GetIDWithEvent() + Constants.SEPARATOR + step.Value;

                        int[] data;
                        if (DFSData.TryGetValue(tmp, out data))
                        {
                            outgoing.Add(tmp);

                            //if this node is still not visited
                            if (data[0] == VISITED_NOPREORDER)
                            {
                                if (done)
                                {
                                    TaskStack.Push(step);
                                    done = false;
                                    product.RemoveAt(k);
                                }
                                else
                                {
                                    product[k] = step;
                                }
                            }
                            //this node is truly visited. can be removed
                            else
                            {
                                product.RemoveAt(k);
                            }
                        }
                        else
                        {
                            DFSData.Add(tmp, new int[] { VISITED_NOPREORDER, 0 });
                            OutgoingTransitionTable.Add(tmp, new List<string>(8));
                            outgoing.Add(tmp);

                            if (done)
                            {
                                TaskStack.Push(step);
                                done = false;
                                product.RemoveAt(k);
                            }
                            else
                            {
                                product[k] = step;
                            }
                        }
                    }

                    //create the remaining steps as the expending list for v
                    ExpendedNode.Add(v, product);
                }

                if (done)
                {
                    int lowlinkV = nodeData[0];
                    int preorderV = lowlinkV;

                    bool selfLoop = false;
                    for (int j = 0; j < outgoing.Count; j++)
                    {
                        string w = outgoing[j];
                        if (w == v)
                        {
                            selfLoop = true;
                        }

                        int[] wdata = DFSData[w];
                        if (wdata[0] != SCC_FOUND)
                        {
                            if (wdata[0] > preorderV)
                            {
                                lowlinkV = Math.Min(lowlinkV, wdata[1]);
                            }
                            else
                            {
                                lowlinkV = Math.Min(lowlinkV, wdata[0]);
                            }
                        }
                    }
                    nodeData[1] = lowlinkV;

                    TaskStack.Pop();

                    if (lowlinkV == preorderV)
                    {
                        bool HasTimeTransition = !evt.IsInstantaneous();

                        //store the always active clockes
                        List<byte> alwaysClock = new List<byte>();

                        StronglyConnectedComponets.Add(v);
                        nodeData[0] = SCC_FOUND;
                        alwaysClock.AddRange(evt.TheConstraint.GetNewClocks());

                        bool BuchiFair = pair.Value.EndsWith(Constants.ACCEPT_STATE);

                        if (stepStack.Count > 0)
                        {
                            KeyValuePair<PARConfiguration, string> valuePair = stepStack.Peek();
                            string k = valuePair.Key.GetIDWithEvent() + Constants.SEPARATOR + valuePair.Value;

                            while (stepStack.Count > 0 && DFSData[k][0] > preorderV)
                            {
                                stepStack.Pop();
                                StronglyConnectedComponets.Add(k);
                                DFSData[k][0] = SCC_FOUND;

                                if (!BuchiFair && valuePair.Value.EndsWith(Constants.ACCEPT_STATE))
                                {
                                    BuchiFair = true;
                                }

                                PARConfiguration step = valuePair.Key;
                                alwaysClock = Common.Classes.Ultility.Ultility.Intersect(alwaysClock, step.TheConstraint.GetNewClocks());
                                HasTimeTransition = HasTimeTransition || !step.IsInstantaneous();
                                if (stepStack.Count > 0)
                                {
                                    valuePair = stepStack.Peek();
                                    k = valuePair.Key.GetIDWithEvent() + Constants.SEPARATOR + valuePair.Value;
                                }
                            }
                        }

                        int SCCSize = StronglyConnectedComponets.Count;

                        if (BuchiFair && (evt.IsDeadLock || SCCSize > 1 || selfLoop) && alwaysClock.Count == 0 && HasTimeTransition)
                        {
                            VerificationOutput.VerificationResult = VerificationResultType.INVALID;
                            VerificationOutput.NoOfStates = DFSData.Count;

                            while (TaskStack.Count > 0 && TaskStack.Peek().Key.Event != Constants.INITIAL_EVENT)
                            {
                                TaskStack.Pop();
                            }

                            string startID = null;
                            if (TaskStack.Count > 0)
                            {
                                startID = TaskStack.Peek().Key.GetIDWithEvent() + Constants.SEPARATOR +
                                          TaskStack.Peek().Value;
                            }

                            //GetCounterExample(StronglyConnectedComponets, startID, OutgoingTransitionTable);
                            VerificationOutput.CounterExampleTrace = GetConcreteTrace(GetCounterExample(StronglyConnectedComponets, startID, OutgoingTransitionTable));
                            return;
                        }

                        foreach (string componet in StronglyConnectedComponets)
                        {
                            ExpendedNode.Remove(componet);
                        }

                        StronglyConnectedComponets.Clear();
                    }
                    else
                    {
                        stepStack.Push(pair);
                    }
                }
            } while (TaskStack.Count > 0);

            VerificationOutput.VerificationResult = VerificationResultType.VALID;
            VerificationOutput.NoOfStates = DFSData.Count;
        }
        
        public override string StartingProcess
        {
            get
            {
                return Process.ToString();
            }
        }

        public override string GetResultString()
        {
            if (ApplyZenoCheck)
            {
                return base.GetResultString() + "NonZenoness: Applied\r\n";
            }
            else
            {
                return base.GetResultString() + "NonZenoness: Not Applied\r\n";                
            }
        }
    }

    public class PAREventBAPair
    {
        public PARConfiguration configuration;
        public string state;
        //public string ID;

        public PAREventBAPair(PARConfiguration e, string s)
        {
            configuration = e;
            state = s;
        }

        public string GetCompressedState()
        {
            return configuration.GetIDWithEvent() + Constants.SEPARATOR + state;
        }

        /// <summary>
        /// Given one environment, get the initial states of the product of the system and the automata. Notice that the automata 
        /// is allowed to make one move first. This is necessary to check the very first state of the system. 
        /// </summary>
        /// <param name="BA"></param>
        /// <param name="initialStep"></param>
        /// <returns></returns>
        public static List<PAREventBAPair> GetInitialPairs(BuchiAutomata BA, PARConfiguration initialStep)
        {
            List<PAREventBAPair> toReturn = new List<PAREventBAPair>();
            HashSet<string> existed = new HashSet<string>();

            foreach (string s in BA.InitialStates)
            {
                List<string> next = BA.MakeOneMove(s, initialStep);

                foreach (string var in next)
                {
                    if (!existed.Contains(var))
                    {
                        existed.Add(var);
                        toReturn.Add(new PAREventBAPair(initialStep, var));
                    }
                }
            }

            return toReturn;
        }

        public static List<PAREventBAPair> Next(BuchiAutomata BA, PARConfiguration[] steps, string BAState)
        {
            List<PAREventBAPair> product = new List<PAREventBAPair>(steps.Length * BA.States.Length);

            for (int i = 0; i < steps.Length; i++)
            {
                PARConfiguration step = steps[i];

                List<string> states = BA.MakeOneMove(BAState, step);

                for (int j = 0; j < states.Count; j++)
                {
                    product.Add(new PAREventBAPair(step, states[j]));
                }
            }

            return product;
        }
    }
}