﻿//#define DEBUG_PRINT
//#define DEBUG_PRINT2

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;
using PAT.PAR.LTS;

namespace PAT.PAR.Assertions
{
    public class PARInverseMethod : AssertionBase
    {
        private DefinitionRef Process;
        private List<KeyValuePair<string, Expression>> ParameterValuation;
        private Constraint result;
        // Option: random selection of an incompatible inequality?
        private static bool RandomSelection = true;
        // Has the analysis been deterministic?
        private int RandomChoices;
        // Number of (finally) reachable states
        private int nbStates;
        // Number of states visited (but potentially merged)
        private int nbVisited;
        // Number of transitions (not represented explicitely so far)
        private int nbTransitions;
        // Maximum number of clocks met during the analysis
        private int maxNbClocks;
        private List<Constraint> inequalities;

        public PARInverseMethod(DefinitionRef processDef, List<KeyValuePair<string, Expression>> pars)
        {                       
            Process = processDef;
            ParameterValuation = pars;
        }

        public override void Initialize(SpecificationBase spec)
        {
            //todo: I am not sure what is the best names here
            //initialize model checking options, the default option is for deadlock/reachablity algorithms
            ModelCheckingOptions = new ModelCheckingOptions();
            List<string> DeadlockEngine = new List<string>();
            DeadlockEngine.Add(Constants.ENGINE_WITH_PAR_OPTIM);
            DeadlockEngine.Add(Constants.ENGINE_WITHOUT_PAR_OPTIM);
            ModelCheckingOptions.AddAddimissibleBehavior(Constants.COMPLETE_BEHAVIOR, DeadlockEngine);


            PARAssertion.Initialize(this, Process, spec);
        }


        // Convert the parameter valuation to a dictionary
        public Dictionary<string, int> ValuationToDictionary()
        {
            Dictionary<string, int> dictionary = new Dictionary<string, int> ();
            // Convert the parameter valuation to a dictionary
            foreach (KeyValuePair<string, Expression> item in ParameterValuation)
            {
                // Convert to int
                int value = Convert.ToInt32(item.Value.ToString());
                // Add to dictionary
                dictionary.Add(item.Key, value);
            }
            return dictionary;
        }


        // Convert the parameter valuation into an int array
        // SUPPOSE THAT THE ParameterValuation IS ORDERED BY INDEX ASC
        public int[] ValuationToIntArray()
        {
            int[] result = new int[ParameterValuation.Count];
            int param_index = 0;
            // Convert the parameter valuation to a dictionary
            foreach (KeyValuePair<string, Expression> item in ParameterValuation)
            {
                // Convert to int
                int value = Convert.ToInt32(item.Value.ToString());
                // Add to array
                result[param_index] = value;
                // Increment index
                param_index++;
            }
            return result;
        }



        /************************************************************
        * MAIN IMPLEMENTATION OF THE INVERSE METHOD ALGORITHM
        ************************************************************/
        public override void RunVerification()
        {
            // Cast for simplicity purpose
            PARConfiguration InitialState = InitialStep as PARConfiguration;

            // CLONE BECAUSE OF A BUG (?) SOMEWHERE IN PAT (running the same verification twice will start from the modified initial state)
            Constraint OldConstraint = InitialState.TheConstraint.Clone();

            // Put the reference valuation under the form of a dictionary
            Dictionary<string, int> dictionary = ValuationToDictionary();

            // Put the reference valuation under the form of an int array (for the pi-incompatible function)
            int[] pi0_int_array = ValuationToIntArray();

            // Define the hash for states already visited: state_id => state
            Dictionary<string, PARConfiguration> Visited = new Dictionary<string, PARConfiguration>();

            // Define the queue of states to be visited
            // TO DO: replace queue by set, to avoid unnecessary duplication of states
            Queue<PARConfiguration> queue = new Queue<PARConfiguration>();
            queue.Enqueue(InitialState);

            // Define the step (i.e., depth) of the algorithm (only for presentation purpose)
            //int step = 1;

            // TO DO: keep track of the (real) transitions, not only their number, for presentation purpose
            nbTransitions = 0;

            // Start with 'true'
            Constraint k = Constraint.constraintWithNonNegativeParameters();

            // No random selection so far
            RandomChoices = 0;

            // No state visited yet
            nbVisited = 0;

            // No clock yet
            maxNbClocks = 0;

            // List of inequalities (for pretty printing)
            inequalities = new List<Constraint>();

            // Seed the random choice in PPL
            Polyhedron.Seed();

            // Define the options
            if (SelectedEngineName == Constants.ENGINE_WITH_PAR_OPTIM)
            {
                Constraint.EQUIVALENCE_REDUCTION = true;
            }
            else
            {
                Constraint.EQUIVALENCE_REDUCTION = false;
            }

#if DEBUG_PRINT
            System.Console.Out.WriteLine("************************************************************");
            System.Console.Out.WriteLine("STARTING INVERSE METHOD");
            System.Console.Out.WriteLine("Initial configuration: \n" + InitialState);
            System.Console.Out.WriteLine("Reference valuation: " + ParameterValuationToString());
            System.Console.Out.WriteLine("K := " + k);
            System.Console.Out.WriteLine("Random mode? " + RandomSelection);
            System.Console.Out.WriteLine("************************************************************");

            int nbVisitedStates = 1;
#endif

            while (queue.Count() != 0)
            {
                // Consider one more state
                PARConfiguration current_state = queue.Dequeue();

                // Count to be visited
                nbVisited++;


                //System.Console.Out.WriteLine("***********************************************");
                //System.Console.Out.WriteLine("NB STATES = " + Visited.Count());
                //System.Console.Out.WriteLine("NB VISITED = " + nbVisited);
                //System.Console.Out.WriteLine("***********************************************");

#if DEBUG_PRINT
                System.Console.Out.WriteLine("************************************************************");
                System.Console.Out.WriteLine("Dequeued following state (#" + nbVisitedStates + ") :");
                System.Console.Out.WriteLine(current_state);
                nbVisitedStates++;

                System.Console.Out.WriteLine("About to add k = " + k);
#endif

                // Update the state by adding the intersection with k
                current_state.AddConstraint(k);
#if DEBUG_PRINT
                System.Console.Out.WriteLine("State ^ K:\n  " + current_state);
#endif
                // PATCH TO SOLVE A VERY STRANGE BUG (18/11/11)
                //string useless = current_state.TheConstraint.thePolyhedron.GetId();
                //System.Console.Out.WriteLine("1) Current state's constraint's ID: " + current_state.TheConstraint.thePolyhedron.GetId());


#if DEBUG_PRINT2
                System.Console.Out.WriteLine("About to check satisfiability...");
#endif
                // Test satisfiability (because the constraint may have changed since it was pushed in, and we added k)
                if (current_state.TheConstraint.IsSatisfiable())
                {

#if DEBUG_PRINT2
                    System.Console.Out.WriteLine("State is satisfiable.");
                    System.Console.Out.WriteLine("Current state: " + current_state);
#endif

                    //string useless456 = current_state.TheConstraint.thePolyhedron.GetId();
                    //System.Console.Out.WriteLine("2) Current state's constraint's ID: " + current_state.TheConstraint.thePolyhedron.GetId());
                    //System.Console.Out.WriteLine("2) State ^ K:\n  " + current_state);


                    //// !!!!! CHECK IF STILL COMPATIBLE (should be the case, but bug found)
                    //if (!current_state.TheConstraint.IsSatisfiable())
                    //{
                    //    System.Console.Out.WriteLine(" *********** Error during the inverse method: constraint is not satisfiable anymore.");
                    //    System.Console.Out.WriteLine("Current state: " + current_state);
                    //    throw new Exception("Error during the inverse method: constraint is not satisfiable anymore, 1st check (after " + Visited.Count() + " states).");
                    //}

                    //System.Console.Out.WriteLine("3) Current state's constraint's ID: " + current_state.TheConstraint.thePolyhedron.GetId());


                    // //!!!!! CHECK IF STILL COMPATIBLE (should be the case, but bug found)
                    //if (!current_state.TheConstraint.IsSatisfiable())
                    //{
                    //    System.Console.Out.WriteLine(" *********** Error during the inverse method: constraint is not satisfiable anymore.");
                    //    System.Console.Out.WriteLine("Current state: " + current_state);
                    //    throw new Exception("Error during the inverse method: constraint is not satisfiable anymore, 2nd check (after " + Visited.Count() + " states).");
                    //}


                    //System.Console.Out.WriteLine("4) Current state's constraint's ID: " + current_state.TheConstraint.thePolyhedron.GetId());


                    // Project onto the parameters
                    Constraint ConstraintOnParameters = current_state.TheConstraint.HideAllClocks();



//                    System.Console.Out.WriteLine("Current state: " + current_state);
                    // !!!!! CHECK IF STILL COMPATIBLE (should be the case, but bug found)
                    //if (!current_state.TheConstraint.IsSatisfiable())
                    //{
                    //    System.Console.Out.WriteLine(" *********** Error during the inverse method: constraint is not satisfiable anymore.");
                    //    System.Console.Out.WriteLine("Current state: " + current_state);
                    //    System.Console.Out.WriteLine("Current constraint: " + current_state.TheConstraint);
                    //    System.Console.Out.WriteLine("Constraint on parameters: " + ConstraintOnParameters);
                    //    throw new Exception("Error during the inverse method: constraint is not satisfiable anymore, after projection (after " + Visited.Count() + " states).");
                    //}
//                    System.Console.Out.WriteLine("Current state (copy) ID after projection" + current_state.TheConstraint.thePolyhedron.GetId());
#if DEBUG_PRINT
                    System.Console.Out.WriteLine("Current constraint on parameters: " + ConstraintOnParameters);
#endif


                    //// !!!!! CHECK IF STILL COMPATIBLE (should be the case, but bug found)
                    //if (!ConstraintOnParameters.IsSatisfiable())
                    //{
                    //    System.Console.Out.WriteLine(" *********** Error during the inverse method: constraint on parameters is not satisfiable anymore.");
                    //    System.Console.Out.WriteLine("Current state: " + current_state);
                    //    System.Console.Out.WriteLine("Current constraint: " + current_state.TheConstraint);
                    //    System.Console.Out.WriteLine("Constraint on parameters: " + ConstraintOnParameters);
                    //    throw new Exception("Error during the inverse method: constraint on parameters is not satisfiable anymore (after " + Visited.Count() + " states).");
                    //}
                    //else
                    //{
                    //    System.Console.Out.WriteLine("Current state: " + current_state);
                    //    System.Console.Out.WriteLine("Current constraint: " + current_state.TheConstraint);
                    //    System.Console.Out.WriteLine("Constraint on parameters: " + ConstraintOnParameters);
                    //}

                    // TEST COMPATIBILITY W.R.T. REFERENCE VALUATION
                    // Create a constraint for the result
                    Constraint negated_inequality = Constraint.constraintWithNonNegativeParameters();


#if DEBUG_PRINT
                    System.Console.Out.WriteLine("Creating fresh constraint for getting the result of incompatible-test: " + negated_inequality);
                    System.Console.Out.WriteLine("Now calling pi0-compatibility method.");
#endif
                    // Call the method
                    bool WasRandom = ConstraintOnParameters.Incompatible(RandomSelection, pi0_int_array, negated_inequality);
#if DEBUG_PRINT
                    System.Console.Out.WriteLine("Result of incompatible-test: " + negated_inequality);
#endif
                    // The constraint was pi0-incompatible if negated_inequality != false
                    if (negated_inequality.IsSatisfiable()) {
#if DEBUG_PRINT
                        System.Console.Out.WriteLine("Found a pi0-incompatible inequality. Negation: " + negated_inequality);
#endif
                        // Update the deterministic flag
                        if (WasRandom)
                        {
#if DEBUG_PRINT
                            System.Console.Out.WriteLine("There may have been some non-determinism in this selection.");
#endif
                            RandomChoices++;
                        }

//                    // Test compatibility w.r.t. reference valuation
//                    List<Constraint> Incompatible = ConstraintOnParameters.NegatedIncompatibleInequalities(dictionary);

//                    // Incompatible with reference valuation
//                    int NbIncompatible = Incompatible.Count();
//                    if (NbIncompatible != 0)
//                    {
//                        // Set the random indicator to true if more than 1 inequality
//                        if(NbIncompatible > 1){
//                            HasBeenDeterministic = false;
//                        }
//                        // Select an inequality
//                        Constraint negated_inequality;
//                        if (RandomSelection)
//                        {
//                            // Random selection
//                            Random random = new Random();
//                            int randomIndex = random.Next(0, Incompatible.Count()-1);
//                            negated_inequality = Incompatible[randomIndex];
//                        }else{
//                            // Pick up the first one
//                            negated_inequality = Incompatible[0];
//                        }
                         
                        // Update K
                        k.Conjunction(negated_inequality);
                        inequalities.Add(negated_inequality);
#if DEBUG_PRINT
                        System.Console.Out.WriteLine("K = " + k);
#endif
                        // Update ALL states visited by adding the newly computed inequality
                        Dictionary<string, PARConfiguration> VisitedUpdated = new Dictionary<string, PARConfiguration>();
                        // Add the negated inequality to all states
                        foreach (KeyValuePair<string, PARConfiguration> kvp in Visited)
                        {
                            // Retrieve the configuration
                            PARConfiguration state = kvp.Value;
                            // Add the negated inequality
                            state.AddConstraint(negated_inequality);
                            // Compute the new id
                            string id = state.GetID();
                            // Try to add to the new set of states
                            // (only try, because states can be merged at this point in case of partial orders)
#if DEBUG_PRINT
                            System.Console.Out.WriteLine("\n--Updating state of (new) id=" + id + ":\n  " + state);
#endif
                            try
                            {
                                VisitedUpdated.Add(id, state);
                            }
                            catch (ArgumentException)
                            {
#if DEBUG_PRINT
                                // Retrieve the former one (just to check that it is indeed the same one)
                                PARConfiguration already_visited;
                                VisitedUpdated.TryGetValue(id, out already_visited);
                                System.Console.Out.WriteLine("  State already visited: " + already_visited);
#endif
                            }
                            // TO DO: update the graph!
                        }
                        Visited = VisitedUpdated;
#if DEBUG_PRINT
                        System.Console.Out.WriteLine("Now " + Visited.Count() + " states visited.");
#endif

                    }
                    // Compatible with reference valuation
                    else
                    {
#if DEBUG_PRINT
                        System.Console.Out.WriteLine("State is compatible.");
#endif
                        // Get the 'id' of the configuration
                        string id = current_state.GetID();
#if DEBUG_PRINT
                        System.Console.Out.WriteLine("Computing id: " + id);
#endif

                        // TODO: update transition graph!
                        nbTransitions++;

                        // Test if the state was visited before
                        if (Visited.ContainsKey(id))
                        {
#if DEBUG_PRINT
                            System.Console.Out.WriteLine("State already visited. Skip.");
#endif
                        }
                        // If new: add it
                        else
                        {
                            Visited.Add(id, current_state);

                            // Update the maximum number of clocks
                            int current_nb_clocks = current_state.getNbClocks();
                            if (current_nb_clocks > maxNbClocks)
                            {
                                maxNbClocks = current_nb_clocks;
                            }
#if DEBUG_PRINT
                            System.Console.Out.WriteLine("Adding state to list of visited (now " + Visited.Count() + " states visited).");
#endif

                            // Compute the successors
                            PARConfiguration[] states = current_state.MakeOneMove() as PARConfiguration[];
                            // Add them to the queue
                            foreach (PARConfiguration state in states)
                            {
                                queue.Enqueue(state);
#if DEBUG_PRINT
                                System.Console.Out.WriteLine("\n--New state added to queue (now " + queue.Count() + " states in queue): \n" + state);
#endif
                            }
                        }
                    }
                }
                else
                {
#if DEBUG_PRINT
                    System.Console.Out.WriteLine("State is unsatisfiable. Discard.");
#endif
                } // If satisfiable
            } // while queue not empty

            // END OF THE ALGORITHM

            // Computes the intersection!
            foreach (KeyValuePair<string, PARConfiguration> kvp in Visited)
            {
                // Retrieve the configuration
                PARConfiguration state = kvp.Value;
                // Intersect!
                k.Conjunction(state.TheConstraint.HideAllClocks());
            }


            // Store the result
            result = k;
            nbStates = Visited.Count();
#if DEBUG_PRINT
            System.Console.Out.WriteLine("************************************************************");
            System.Console.Out.WriteLine("INVERSE METHOD SUCCESSFULLY TERMINATED");
            System.Console.Out.WriteLine("K = " + k);
            System.Console.Out.WriteLine("Random choices: " + RandomChoices);
            System.Console.Out.WriteLine(nbVisitedStates + " states visited.");
            System.Console.Out.WriteLine(Visited.Count() + " states reached: ");
            foreach (KeyValuePair<string, PARConfiguration> kvp in Visited)
            {
                System.Console.Out.WriteLine("------------------------------------------------------------");
                System.Console.Out.WriteLine("Id=" + kvp.Key);
                System.Console.Out.WriteLine(kvp.Value);
            }
            System.Console.Out.WriteLine("************************************************************");
#endif
            // Result always "valid" for IM
            this.VerificationOutput.VerificationResult = VerificationResultType.VALID;
            // Update result
            this.VerificationOutput.NoOfStates = Visited.Count;
            this.VerificationOutput.Transitions = nbTransitions;

            // REVERSE CLONE BECAUSE OF A BUG (?) SOMEWHERE IN PAT (running the same verification twice will start from the modified initial state)
            InitialState.TheConstraint = OldConstraint;
        }



        /*
        // MAIN IMPLEMENTATION OF THE REACHABILITY SYNTHESIS
        public override void RunVerification()
        {
            // Cast for simplicity purpose
            PARConfiguration InitialState = InitialStep as PARConfiguration;

            // Define the hash for states already visited
            Dictionary<string, PARConfiguration> Visited = new Dictionary<string, PARConfiguration>();

            // Define the queue of states to be visited
            // TO DO: replace queue by set, to avoid unnecessary duplication of states
            Queue<PARConfiguration> queue = new Queue<PARConfiguration>();
            queue.Enqueue(InitialState);

            // Set of constraints
            HashSet<Constraint> ConstraintSet = new HashSet<Constraint>();

#if DEBUG_PRINT
            System.Console.Out.WriteLine("************************************************************");
            System.Console.Out.WriteLine("STARTING SYNTHESIS W.R.T. REACHABILITY");
            System.Console.Out.WriteLine("Initial configuration: \n" + InitialState.ToString());
#endif

            while (queue.Count() != 0)
            {
                // Consider one more state
                PARConfiguration current_state = queue.Dequeue();
#if DEBUG_PRINT
                System.Console.Out.WriteLine("************************************************************");
                System.Console.Out.WriteLine("Dequeued following state:");
                System.Console.Out.WriteLine(current_state);
#endif

                // Compute the id
                string id = current_state.GetID();

                // Test whether the state was visited before
                bool visited = Visited.ContainsKey(id);
                    // TO DO: case option inclusion!
                if (visited)
                {
#if DEBUG_PRINT
                    System.Console.Out.WriteLine("State already visited. Skip.");
#endif
                }
                // If new: add it
                else
                {
                    Visited.Add(id, current_state);

                    // Test compatibility w.r.t. predicate
                    // TO DO

                    // If satisfies predicate: store the constraint
                    // Project onto the parameters
                    Constraint OnParam = current_state.TheConstraint.HideAllClocks();
                    bool Added = ConstraintSet.Add(OnParam);
#if DEBUG_PRINT
                    System.Console.Out.WriteLine("Adding constraint " + OnParam);
                    if (Added)
                    {
                        System.Console.Out.WriteLine("(already present)");
                    }
#endif
                    // OPTION TO DO: inclusion or equality

                    // If does not satisfy predicate: compute successors
                    PARConfiguration[] states = current_state.MakeOneMove() as PARConfiguration[];
                    // Add them to the queue
                    foreach (PARConfiguration state in states)
                    {
                        queue.Enqueue(state);
#if DEBUG_PRINT
                                System.Console.Out.WriteLine("\n--New state added to queue (now " + queue.Count() + " states in queue): \n" + state);
#endif
                    }
                }

            } // while queue not empty

            // END OF THE ALGORITHM
            // Store the result
            result = ConstraintSet;
#if DEBUG_PRINT
            System.Console.Out.WriteLine("************************************************************");
            System.Console.Out.WriteLine("REACHABILITY ANALYSIS SUCCESSFULLY TERMINATED");
            System.Console.Out.WriteLine(Visited.Count() + " states reached: ");
            foreach (KeyValuePair<string, PARConfiguration> kvp in Visited)
            {
                System.Console.Out.WriteLine("------------------------------------------------------------");
                System.Console.Out.WriteLine("Id=" + kvp.Key);
                System.Console.Out.WriteLine(kvp.Value);
            }
            System.Console.Out.WriteLine("************************************************************");
#endif

        }*/




        public override string StartingProcess
        {
            get
            {
                return Process.ToString();
            }
        }

        public override string ToString()
        {
            return StartingProcess + " with " + ParameterValuationToString();
        }


        // Convert the parameter valuation to a string
        private string ParameterValuationToString()
        {
            StringBuilder sb = new StringBuilder();
            foreach (KeyValuePair<string, Expression> pair in ParameterValuation)
            {
                sb.Append(pair.Key + "=" + pair.Value + ", ");
            }
            return sb.ToString().Substring(0, sb.ToString().Length - 2);
        }


        public override string GetResultString()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine("********Inverse Method Result********");
            sb.AppendLine(result.ToString());
            
            if (SelectedEngineName == Constants.ENGINE_WITH_PAR_OPTIM)
            {
                sb.AppendLine("Algorithm version: with state equivalence reduction [ALSD12].");
            }
            else
            {
                sb.AppendLine("Algorithm version: without state equivalence reduction [ALSD12].");
            }

            sb.AppendLine(nbStates + " states reachable.");
//            sb.AppendLine("These states were visited " + nbVisited + " times.");
            if (RandomChoices == 0)
            {
                sb.AppendLine("The analysis has been fully deterministic.");
            }
            else
            {
                sb.AppendLine("The analysis may have been non-deterministic (" + RandomChoices + " random choices).");
                if (!RandomSelection)
                {
                    sb.AppendLine("But analysis has been performed in no-random mode.");
                }
            }
//            sb.AppendLine("Number of transitions: " + nbTransitions);
            sb.AppendLine("Maximal number of clocks: " + maxNbClocks);
            sb.AppendLine("Inequalities added by IM:");
            // Add inequalities
            int i = 1;
            foreach (Constraint inequality in inequalities)
            {
                sb.AppendLine(i + ") " + inequality.ToString());
                i++;
            }
            return sb.ToString();
        }
    }
}
