﻿using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class AtomicProcess : PARProcess
    {
        public PARProcess Process;

        public AtomicProcess(PARProcess process)
        {
            if (process is AtomicProcess)
            {
                Process = (process as AtomicProcess).Process;
            }
            else
            {
                Process = process;
            }

            ProcessID = PARDataStore.DataManager.InitializeProcessID(Constants.ATOMIC + Process.ProcessID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> next = Process.MoveOneStep(eStep);

            foreach (PARConfiguration configuration in next)
            {
                configuration.Process = new AtomicProcess(configuration.Process);
                configuration.IsAtomic = true;
            }

            return next;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new AtomicProcess(Process.ClearConstant(constMapping));
        }

        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> steps = Process.SyncOutput(eStep);
            foreach (PARConfigurationWithChannelData step in steps)
            {
                step.Process = new AtomicProcess(step.Process);
                step.IsAtomic = true;
            }

            return steps;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> steps = Process.SyncInput(eStep);
            foreach (PARConfiguration step in steps)
            {
                step.Process = new AtomicProcess(step.Process);
                step.IsAtomic = true;
            }

            return steps;
        }

        public override string ToString()
        {
            return "atomic{" + Process + "}";
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return Process.GetAlphabets(visitedDefinitionRefs);
        }

        public override List<string> GetGlobalVariables()
        {
            return Process.GetGlobalVariables();
        }

        public override List<string> GetChannels()
        {
            return Process.GetChannels();
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return Process.GetTopLevelConcurrency(visitedDef);
        }

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = Process.SetTimer(timerID, clocks);
            if (process == null)
            {
                return null;
            }

            return new AtomicProcess(process);
        }
    }
}
