using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class CaseProcess : PARProcess
    {
        public PARProcess[] Processes;
        public Expression[] Conditions;

        public CaseProcess(PARProcess[] processes, Expression[] conds)
        {
            Processes = processes;
            Conditions = conds;

            //generate the process ID
            StringBuilder IDBuilder = new StringBuilder(Constants.CASE);
            for (int i = 0; i < Processes.Length; i++)
            {
                IDBuilder.Append(Conditions[i].ExpressionID);
                IDBuilder.Append(Constants.CASECONDITIONAL);
                IDBuilder.Append(Processes[i].ProcessID);
                IDBuilder.Append(";");
            }

            ProcessID = PARDataStore.DataManager.InitializeProcessID(IDBuilder.ToString());
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = new List<string>();

            for (int i = 0; i < Conditions.Length; i++)
            {
                Common.Classes.Ultility.Ultility.AddList(Variables, Processes[i].GetGlobalVariables());
                Common.Classes.Ultility.Ultility.AddList(Variables, Conditions[i].GetVars());
            }

            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channels = new List<string>();

            for (int i = 0; i < Conditions.Length; i++)
            {
                List<string> vars = Processes[i].GetChannels();
                foreach (string var in vars)
                {
                    if (!channels.Contains(var))
                    {
                        channels.Add(var);
                    }
                }
            }

            return channels;
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> ReturnList = new List<PARConfiguration>();

            for (int i = 0; i < Processes.Length; i++)
            {
                if ((EvaluatorDenotational.Evaluate(Conditions[i], eStep.GlobalEnv) as BoolConstant).Value)
                {
                    ReturnList.Add(new PARConfiguration(Processes[i], Constants.TAU, "[" + Conditions[i] + "]", eStep.GlobalEnv, false, true, eStep.TheConstraint));
                    return ReturnList;
                }
            }

            ReturnList.Add(new PARConfiguration(new Skip(), Constants.TAU, "[default]", eStep.GlobalEnv, false, true, eStep.TheConstraint));
            return ReturnList;
        }



        public override string ToString()
        {
            if (Processes.Length == 1)
            {
                return "if" + Conditions[0] + "{" + Processes[0].ToString() + "}";
            }

            StringBuilder s = new StringBuilder();
            s.AppendLine("case {");
            for (int i = 0; i < Processes.Length; i++)
            {
                Expression con = Conditions[i];
                PARProcess process = Processes[i];
                s.AppendLine(con.ToString() + ":" + process.ToString());
            }

            s.AppendLine("}");
            return s.ToString();
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> toReturn = new HashSet<string>();

            for (int i = 0; i < Processes.Length; i++)
            {
                PARProcess process = Processes[i];
                toReturn.UnionWith(process.GetAlphabets(visitedDefinitionRefs));
            }

            return toReturn;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            PARProcess[] newProcesses = new PARProcess[Processes.Length];
            Expression[] newConditions = new Expression[Processes.Length];

            for (int i = 0; i < Processes.Length; i++)
            {
                newConditions[i] = Conditions[i].ClearConstant(constMapping);
                newProcesses[i] = Processes[i].ClearConstant(constMapping);                
            }

            return new CaseProcess(newProcesses, newConditions);
        }

        public override bool MustBeAbstracted()
        {
            for (int i = 0; i < Processes.Length; i++)
            {
                if (Processes[i].MustBeAbstracted())
                {
                    return true;
                }
            }

            return false;
        }

        //public override bool IsTime()
        //{
        //    for (int i = 0; i < Processes.Length; i++)
        //    {
        //        if (Processes[i].IsTime())
        //        {
        //            return true;
        //        }
        //    }

        //    return false;
        //}
    }
}