using System;
using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class ChannelInput : PARProcess
    {
        public string ChannelName;
        public Expression[] ExpressionList;
        public PARProcess Process;
        public bool IsUrgent;

        public ChannelInput(string evtName, Expression[] exp, PARProcess process, bool isUrgent)
        {
            ChannelName = evtName;
            IsUrgent = isUrgent;
            ExpressionList = exp;
            Process = process;
            String ID = ChannelName + "?" + Common.Classes.Ultility.Ultility.PPIDListDot(ExpressionList) +
                   (IsUrgent ? Constants.EVENTPREFIX_URGENT : Constants.EVENTPREFIX) + Process.ProcessID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID.ToString());
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> list = new List<PARConfiguration>();
            Valuation globalEnv = eStep.GlobalEnv;

            ChannelQueue Buffer = null;

            if (globalEnv.Channels.TryGetValue(this.ChannelName, out Buffer))
            {
                if (Buffer.Count > 0)
                {
                    ChannelQueue newBuffer = Buffer.Clone();
                    ExpressionValue[] values = newBuffer.Dequeue();

                    if (values.Length == ExpressionList.Length)
                    {
                        Dictionary<string, Expression> mapping = new Dictionary<string, Expression>(values.Length);

                        Valuation newEnv = eStep.GlobalEnv.GetChannelClone();
                        newEnv.Channels[ChannelName] = newBuffer;
                        string eventName = ChannelName + "?";
                        string eventID = ChannelName + "?";

                        for (int i = 0; i < ExpressionList.Length; i++)
                        {
                            ExpressionValue v = values[i];
                            if (i == ExpressionList.Length - 1)
                            {
                                eventName += v;
                                eventID += v.ExpressionID;
                            }
                            else
                            {
                                eventName += v + ".";
                                eventID += v.ExpressionID + ".";
                            }

                            if (ExpressionList[i] is Variable)
                            {
                                mapping.Add(ExpressionList[i].ExpressionID, v);
                            }
                            else
                            {
                                if (v.ExpressionID != ExpressionList[i].ExpressionID)
                                {
                                    return list;
                                }
                            }
                        }

                        PARProcess newProcess = mapping.Count > 0 ? Process.ClearConstant(mapping) : Process;
                        list.Add(new PARConfiguration(newProcess, eventID, eventID != eventName ? eventName : null,
                                                      newEnv, false, IsUrgent, eStep.TheConstraint));
                    }
                }
            }

            return list;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> ReturnList = new List<PARConfiguration>(1);
            if (eStep.ChannelName == ChannelName && eStep.Expressions.Length == ExpressionList.Length)
            {
                Dictionary<string, Expression> mapping = new Dictionary<string, Expression>(eStep.Expressions.Length);

                for (int i = 0; i < ExpressionList.Length; i++)
                {
                    Expression v = eStep.Expressions[i];


                    if (ExpressionList[i] is Variable)
                    {
                        mapping.Add(ExpressionList[i].ExpressionID, v);
                    }
                    else
                    {
                        if (v.ExpressionID != ExpressionList[i].ExpressionID)
                        {
                            return ReturnList;
                        }
                    }
                }

                PARProcess newProc = Process;
                if (mapping.Count > 0)
                {
                    newProc = Process.ClearConstant(mapping);
                }

                ReturnList.Add(new PARConfiguration(newProc, null, null, eStep.GlobalEnv, false, IsUrgent, eStep.TheConstraint));
            }

            return ReturnList;                
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {      
            return Process.GetAlphabets(visitedDefinitionRefs);
        }

        public override List<string> GetGlobalVariables()
        {
            return Process.GetGlobalVariables();
        }

        public override List<string> GetChannels()
        {
            List<string> toReturn = Process.GetChannels();

            if (!toReturn.Contains(ChannelName))
            {
                toReturn.Add(ChannelName);
            }

            return toReturn;
        }

        public override string ToString()
        {
            return ChannelName + "?" + Common.Classes.Ultility.Ultility.PPStringListDot(ExpressionList).TrimStart('.') + (IsUrgent ? "->>" : "->") + Process;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Expression[] newExpression = new Expression[ExpressionList.Length];
            for (int i = 0; i < ExpressionList.Length; i++)
            {
                if (ExpressionList[i] is ExpressionValue)
                {
                    newExpression[i] = ExpressionList[i];                        
                }
                else
                {
                    newExpression[i] = ExpressionList[i].ClearConstant(constMapping);
                    //evaluate the value after the clearance, to make sure there only single variable or single value for each expression
                    if (!newExpression[i].HasVar)
                    {
                        newExpression[i] = EvaluatorDenotational.Evaluate(newExpression[i], null);
                    }
                }                
            }

            return new ChannelInput(ChannelName, newExpression, Process.ClearConstant(constMapping), IsUrgent);
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        //public override bool IsTime()
        //{
        //    return Process.IsTime();
        //}

    }
}