using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class ConditionalChoice : PARProcess
    {
        public  PARProcess FirstProcess;
        public  PARProcess SecondProcess;
        public  Expression ConditionalExpression;

        public ConditionalChoice(PARProcess firstProcess, PARProcess secondProcess, Expression conditionExpression)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;
            ConditionalExpression = conditionExpression;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.CONDITIONAL_CHOICE +
                                                                      conditionExpression.ExpressionID + Constants.CONDITIONAL_CHOICE +
                                                                      SecondProcess.ProcessID);
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();

            ExpressionValue v = EvaluatorDenotational.Evaluate(ConditionalExpression, eStep.GlobalEnv);

            if ((v as BoolConstant).Value)
            {
                returnList.Add(new PARConfiguration(FirstProcess, Constants.TAU, "[if(" + ConditionalExpression + ")]", eStep.GlobalEnv, false, false, eStep.TheConstraint));
            }
            else
            {
                returnList.Add(new PARConfiguration(SecondProcess, Constants.TAU, "[else(" + ConditionalExpression + ")]", eStep.GlobalEnv, false, false, eStep.TheConstraint));
            }

            return returnList;
        }



        public override string ToString()
        {
            return "if " + ConditionalExpression + " {" + FirstProcess + "} else {" + SecondProcess + "}";
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = SecondProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, FirstProcess.GetGlobalVariables());
            Common.Classes.Ultility.Ultility.AddList(Variables, ConditionalExpression.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channels = SecondProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList(channels, FirstProcess.GetChannels());
            return channels;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Expression newCon = ConditionalExpression.ClearConstant(constMapping);
            return new ConditionalChoice(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping), newCon);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        //public override bool IsTime()
        //{
        //    return FirstProcess.IsTime() || SecondProcess.IsTime();
        //}
    }
}