﻿//#define DEBUGETIENNE
//#define DEBUGCONSOLE

using System.Linq;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Text.RegularExpressions;
using PAT.Common.Classes.Ultility;
using PAT.Common.Classes.Expressions.ExpressionClass;
using System.Runtime.InteropServices;
using System.IO;
using System.Windows.Forms;



namespace PAT.PAR.LTS
{
    // The operators
    public enum TimerOperationType : byte
    {
        Equals,
        LessThanOrEqualTo,
        GreaterThanOrEqualTo,
        LessThan,
        GreaterThan
    }


    public class Constraint
    {
        /* CONSTANTES */
        private const int FIRST_CLOCK = 1;

        /* MEMBERS */
        // The list of clocks
        private List<byte> ClockList;

#if DEBUGETIENNE
        private static int nbConstraints = 0;
        private int idConstraint;
        private string debugString;
#endif

        // constraint under the form of a string
        private string theConstraint;
        // True constraint with non-negative parameters
        public static string NonNegativeParametersConstraint = null;

        // Some strings (can be changed)
        public static string STRING_CLOCK_NAME      = "x";
        public static string STRING_CLOCK_NAME_TMP  = "##";
        // Some strings (depend on the C++ interface; do not change!)
        public static string STRING_UNDEFINED       = "undefined";
        public static string STRING_TRUE            = "true";
        public static string STRING_FALSE           = "false";
        public static string STRING_SEP_AND         = "&";
        public static string STRING_SEP_SET         = ",";
        public static string STRING_OP_LEQ          = "<=";
        public static string STRING_OP_EQ           = "=";
        public static string STRING_OP_L            = "<";



        private Constraint()
        {
            // Initialize the list of clocks
            ClockList = new List<byte>();

//            MessageBox.Show("start");
            // Initialize the constraint
            theConstraint = STRING_UNDEFINED;

#if DEBUGETIENNE
        debugString = STRING_UNDEFINED + "\n";
        idConstraint = nbConstraints;
        nbConstraints++;
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    (New empty constraint)" + this);
#endif
        }


        // Create a new constraint on parameters only (no clock)
        private static Constraint ConstraintOnParameters(string TheConstraintOnParam)
        {
            Constraint result = new Constraint();
            result.theConstraint = TheConstraintOnParam;
            return result;
        }


        // Static method to compute (only once per model) the non-negative parameters constraint
        public static void InitParameters()
        {
            bool first = true;
            string theConstraint = "";
            // Get the parameters
            foreach (string name in Specification.Parameters)
            {
                // Add '&'
                if (first)
                {
                    first = false;
                }
                else
                {
                    theConstraint += STRING_SEP_AND;
                }
                // Add inequality
                theConstraint += "0<=" + name;
            }
            // Copy to a static field
            NonNegativeParametersConstraint = theConstraint;

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Updated constraint: " + NonNegativeParametersConstraint);
#endif
            
        }

        /************************************************************
         * CONSTRUCTORS
         ***********************************************************/

        //public static Constraint trueConstraint()
        //{
        //    Constraint c = new Constraint();
        //    c.theConstraint = STRING_TRUE;
        //    return c;
        //}

        /**
         * Creates a new constraint, with a set of inequalities 'p >= 0', for all parameters 'p'
         */
        public static Constraint constraintWithNonNegativeParameters()
        {
            // Create and update the constraint
            Constraint c = new Constraint();

            #if DEBUGETIENNE
            c.debugString = "NEW CONSTRAINT OF ID = " + c.idConstraint + " : \n" + NonNegativeParametersConstraint + "\n";
            #endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    New constraint: \n" + c.ToString());
            System.Console.Out.WriteLine("    ====================");
#endif

            c.theConstraint = NonNegativeParametersConstraint;
            return c;
        }



        /**
         * Creates a new constraint under the form of an inequality clockID <= constant
         */
        public static Constraint inequality(byte clockID, int constant)
        {
            // Create the constraint
            Constraint c = new Constraint();
            // Add the inequality
            c.theConstraint = ClockToString(clockID) + opToString(TimerOperationType.LessThanOrEqualTo) + constant;

#if DEBUGETIENNE
            c.debugString = "NEW CONSTRAINT OF ID = " + c.idConstraint
                + " : \n" + c.theConstraint + "\n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    New inequality: \n" + c.ToString());
            System.Console.Out.WriteLine("    ====================");
#endif
            return c;
        }


        /**
         * Creates a new constraint under the form of a parametric inequality clockID <= parameter
         */
        public static Constraint parInequality(byte clockID, string parameter)
        {
            // Create the constraint
            Constraint c = new Constraint();
            // Add the inequality
            c.theConstraint = ClockToString(clockID) + opToString(TimerOperationType.LessThanOrEqualTo) + parameter;

#if DEBUGETIENNE
            c.debugString = "NEW CONSTRAINT OF ID = " + c.idConstraint
                + " : \n" + c.theConstraint + "\n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    New par inequality: \n" + c.ToString());
            System.Console.Out.WriteLine("    ====================");
#endif

            return c;
        }

        //public static Constraint falseConstraint()
        //{
        //    Constraint c = new Constraint();
        //    c.theConstraint = STRING_FALSE;
        //    return c;
        //}




        //public byte getLatestClock()
        //{
        //    return LatestClock;
        //}


        /************************************************************
        * ACCESS WITH NO MODIFICATION
        ***********************************************************/
        

        public int GetNbClocks()
        {
            return ClockList.Count();
        }



        public bool HasClocks()
        {
            return ClockList.Count() != 0;
        }


        public bool ContainsClock(byte clockID)
        {
            return ClockList.Contains(clockID);
        }



        // Generate a new clock ID if not present yet
        public byte GetNewClockID()
        {
            // Case no clock: first clock index
            if (ClockList.Count() == 0)
            {
                return FIRST_CLOCK;
            }


            // ETIENNE SAYS: THIS MAY BE MORE EFFICIENT THAN THE SAME OPERATION IN DBM CLASS (n log n againt n*n)
            // Otherwise: look for the smallest unused clock index
            byte i = FIRST_CLOCK;
            // Copy the list
            List<byte> ClockListCopy = new List<byte>(ClockList);
            // Sort it
            ClockListCopy.Sort();
            // Look for the first index not present in the array
            while (i <= ClockList.Count())
            {
                if (ClockListCopy[i - 1] != i)
                {
                    return i;
                }
                i++;
            }

            // If no free clock index was found, return the next one
            return i;
        }



        public bool IsSatisfiable()
        {
            // DANGEROUS IMPLEMENTATION (BUT SHOULD BE OK)
            return theConstraint != STRING_FALSE;
            // ok because the constraints are always simplified by PPL, thus an insatisfiable constraint will be false anyway
        }



        public List<byte> GetNewClocks()
        {
            if (ClockList.Count() == 0)
            {
                return new List<byte>();
            }

            // No Clone!
            return ClockList;
        }





        // Check if a constraint satisfies an inequality "clock ~ constant"
        public bool ImpliesInequality(byte clockID, TimerOperationType op, int constant)
        {
            string inequality = ClockToString(clockID) + opToString(op) + constant.ToString();
            return PPL.contains(inequality, this.theConstraint);
        }




        // Check if a constraint satisfies a parametric inequality "clock ~ parameter"
        public bool ImpliesParInequality(byte clockID, TimerOperationType op, string parameter)
        {
            string inequality = ClockToString(clockID) + opToString(op) + parameter;
            return PPL.contains(inequality, this.theConstraint);
        }


        // Retrieve the value associated to a parameter; no test performed!
        private static int EvaluateParameter(string parameter, Dictionary<string, int> ParameterValuation)
        {
            int value = 0;
            ParameterValuation.TryGetValue(parameter, out value);
            return value;
        }


        // Parse an inequality, evalute it w.r.t. the reference valuation; if incompatible, negate it and return it
        // If compatible, return null
        private static Constraint NegateInequalityIfIncompatible(string inequality, Dictionary<string, int> ParameterValuation)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("      Entering NegateInequalityIfIncompatible(" + inequality + ")");
#endif
            // Get the operator of the inequality
            TimerOperationType op;
            if (inequality.Contains(STRING_OP_LEQ))
            {
                op = TimerOperationType.LessThanOrEqualTo;
            }
            else if (inequality.Contains(STRING_OP_L))
            {
                op = TimerOperationType.LessThan;
            }
            else if (inequality.Contains(STRING_OP_EQ))
            {
                op = TimerOperationType.Equals;
            }
            else
            {
                throw new Exception("Inequality " + inequality + " is not well-formed: no operator found.");
            }

            // Remove '<=' and split
            string[] split = inequality.Split(new Char[] { '<', '=' });

            // Retrieve the members
            string member_left = "";
            string member_right = "";
            // Depend on the form of the split
            if (split.Length == 2){
                member_left = split[0];
                member_right = split[1];
            }
            else if (split.Length == 3){
                member_left = split[0];
                member_right = split[2];
            }
            // Check that the form is correct
            if (member_left == "" || member_right == "")
            {
                string error = "Inequality " + inequality + " is not well-formed.\n Split = ";
                for(int i = 0; i < split.Length; i++)
                {
                    error += i + " : " + split[i] + " - ";
                }
                throw new Exception(error);
            }

            List<string> members = new List<string>();
            members.Add(member_left);
            members.Add(member_right);

            // To differentiate between left and right member
            bool left = true;
            // To compute left and right evaluation
            int left_valuation = 0;
            int right_valuation = 0;
            foreach (string member in members)
            {
#if DEBUGCONSOLE             
                System.Console.Out.WriteLine("      Split member " + member);
#endif
                // Split w.r.t. '+'
                string[] terms = member.Split(new Char[] { '+' });

                foreach (string term in terms)
                {
                    if (term != "")
                    {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("        Split term " + term);
#endif

                        // Look for the initial digits
                        Regex pattern = new Regex(@"\b(\d*)");
                        Match m = pattern.Match(term);
                        if (m.Success && m.Length != 0)
                        {
                            string digits = m.Groups[1].ToString();
                            // If only digit: constant
                            if (digits.Length == term.Length)
                            {
                                // Update term
                                if (left)
                                {
                                    left_valuation += Convert.ToInt32(digits);
#if DEBUGCONSOLE
                                    System.Console.Out.WriteLine("        Adding constant on the left: " + digits);
#endif
                                }
                                else
                                {
                                    right_valuation += Convert.ToInt32(digits);
#if DEBUGCONSOLE
                                    System.Console.Out.WriteLine("        Adding constant on the right: " + digits);
#endif
                                }
                            }
                                // Coeff + parameter
                            else
                            {
                                // Get the parameter name
                                string parameter = term.Substring(digits.Length);
                                // Get the parameter valuation
                                int value = EvaluateParameter(parameter, ParameterValuation);
#if DEBUGCONSOLE
                                System.Console.Out.WriteLine("        Found parameter:" + parameter + " = " + value);
#endif

                                // Update term
                                int coeff = Convert.ToInt32(digits);
                                if (left)
                                {
                                    left_valuation += coeff * value;
#if DEBUGCONSOLE
                                    System.Console.Out.WriteLine("        Adding on the left: " + coeff + "*" + value);
#endif
                                }
                                else
                                {
                                    right_valuation += coeff * value;
#if DEBUGCONSOLE
                                    System.Console.Out.WriteLine("        Adding on the right: " + coeff + "*" + value);
#endif
                                }
                            }
                        }
                            // Only a parameter
                        else
                        {
                            // Get the parameter valuation
                            int value = EvaluateParameter(term, ParameterValuation);
#if DEBUGCONSOLE
                            System.Console.Out.WriteLine("        Found parameter: " + term + " = " + value);
#endif

                            // Update term
                            if (left)
                            {
                                left_valuation += value;
#if DEBUGCONSOLE
                                System.Console.Out.WriteLine("        Adding on the left: " + value);
#endif
                            }
                            else
                            {
                                right_valuation += value;
#if DEBUGCONSOLE
                                System.Console.Out.WriteLine("        Adding on the right: " + value);
#endif
                            }
                        }
                    }
                } // foreach term

                // First: left, then right
                if(left)
                {
                    left = false;
                }
            } // For each member

            // Evaluate the inequality
            switch (op)
            {
                case TimerOperationType.LessThanOrEqualTo:
                    // Incompatible
                    if (left_valuation > right_valuation)
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("      Incompatible because " + left_valuation + " > " + right_valuation);
#endif
                        // Add the negation of the inequality
                        return ConstraintOnParameters(member_right + opToString(TimerOperationType.LessThan) + member_left);
                    }
                    // Compatible
                    else
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("      Compatible because " + left_valuation + " <= " + right_valuation);
#endif
                        return null;
                    }
                case TimerOperationType.LessThan:
                    // Incompatible
                    if (left_valuation >= right_valuation)
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("      Incompatible because " + left_valuation + " >= " + right_valuation);
#endif
                        // Add the negation of the inequality
                        return ConstraintOnParameters(member_right + opToString(TimerOperationType.LessThanOrEqualTo) + member_left);
                    }
                    // Compatible
                    else
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("      Compatible because " + left_valuation + " < " + right_valuation);
#endif
                        return null;
                    }
                case TimerOperationType.Equals:
                    // Incompatible
                    if (left_valuation > right_valuation)
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("      Incompatible because " + left_valuation + " > " + right_valuation);
#endif
                        // Add the negation of the inequality
                        return ConstraintOnParameters(member_right + opToString(TimerOperationType.LessThan) + member_left);
                    }
                    else if (right_valuation > left_valuation)
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("      Incompatible because " + right_valuation + " > " + left_valuation);
#endif
                        // Add the negation of the inequality
                        return ConstraintOnParameters(member_left + opToString(TimerOperationType.LessThan) + member_right);
                    }
                    // Compatible
                    else
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("      Compatible because " + left_valuation + " = " + right_valuation);
#endif
                        return null;
                    }
            }
            // This can't happen (from the parsing)
            throw new Exception("The function should have returned already.");
        }

        // Return a list of inequalities from the constraint that are incompatible with a reference valuation, and negate them
        // The constraint must be satisfiable
        // If the list is empty, the constraint is satisfied by the valuation
        public List<Constraint> NegatedIncompatibleInequalities(Dictionary<string, int> ParameterValuation)
        {
            List<Constraint> list = new List<Constraint>();

            // No false constraint allowed here
            if (theConstraint == STRING_FALSE)
            {
                throw new Exception("Attempting to test the compatibility of a false constraint with a reference valuation.");
            }

            // True constraint is satisfiable: empty list
            if (theConstraint == STRING_TRUE)
            {
                return list;
            }

            // Split between inequalities
            string[] split = theConstraint.Split(new Char[] { '&' });

            // Check and negate inequalities
            foreach(string inequality in split)
            {
                if (inequality != "")
                {
                    // Negate the inequality
                    Constraint negation = NegateInequalityIfIncompatible(inequality, ParameterValuation);
                    // If null: already compatible
                    if (negation != null)
                    {
                        list.Add(negation);
                    }
                }
            }
            return list;
        }


        /************************************************************
        * OPERATIONS RETURNING A NEW OBJECT (no modification on this)
        ***********************************************************/


        // Remove the clocks not in "clocks"; keep the others and the parameters
        public Constraint KeepClocks(List<byte> ClocksToKeep)
        {
            // Test just in case; should not happen in normal mode, but may be called for exception and debug print
            if (theConstraint == STRING_FALSE)
            {
                Constraint resultConstraint = new Constraint();
                resultConstraint.theConstraint = STRING_FALSE;
                return resultConstraint;
            }

            // Compute the list of clocks to remove
            List<byte> ClocksToRemove = ListDiff(ClockList, ClocksToKeep);

#if DEBUGETIENNE
            debugString = "ABOUT TO KEEP CLOCKS {" + SomeClockstoString(ClocksToKeep)
                + "} I.E., HIDE CLOCKS {" + SomeClockstoString(ClocksToRemove)
                + "}, IN (ID = " + idConstraint + ") : [\n" + debugString + "\n] // hide \n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Keeping clocks {" + SomeClocksToString(ClocksToKeep)
                + "} i.e., hide {" + SomeClocksToString(ClocksToRemove)
                + "} in: \n" + this);
#endif

            // OPTIMIZATION: no hiding in true & no hiding of an empty set of clocks
            if (theConstraint == NonNegativeParametersConstraint || ClocksToRemove.Count() == 0)
            {
                Constraint toReturn = this.Clone();
#if DEBUGETIENNE
                debugString += "RESULT OF KEEPING = " + theConstraint + " (UNCHANGED, BUT NEW ID " + toReturn.idConstraint + " )\n";
#endif
                return toReturn;
            }

            string result = PPL.hide(theConstraint, SomeClocksToString(ClocksToRemove));

            Constraint newConstraint = new Constraint();
            newConstraint.theConstraint = result;

            // Update the list of clocks
            newConstraint.ClockList = new List<byte>(ClocksToKeep);

#if DEBUGETIENNE
            newConstraint.debugString = debugString;
            newConstraint.debugString += "RESULT OF KEEPING : \n" + result + " (ID = " + idConstraint + ") \n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + newConstraint.ToString());
            System.Console.Out.WriteLine("    ====================");
#endif
            return newConstraint;
        }


        // HIDE ALL CLOCKS IN THE CONSTRAINT (CLONE VERSION, this REMAINS UNCHANGED)
        public Constraint HideAllClocks()
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    About to hide all clocks {" + ClocksToString() + "} in: \n" + this);
#endif

            // OPTIMIZATION: no hiding in true & no hiding of an empty set of clocks
            if (theConstraint == NonNegativeParametersConstraint || ClockList.Count() == 0)
            {
                Constraint toReturn = this.Clone();
#if DEBUGETIENNE
                toReturn.debugString = "ABOUT TO HIDE ALL CLOCKS {" + ClocksToString() + "} IN : [\n" + debugString + "\n] // hide \n";
                toReturn.debugString += "RESULT OF HIDING = " + toReturn.theConstraint + " (UNCHANGED, BUT NEW ID " + toReturn.idConstraint + " )\n";
#endif
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    (unchanged)");
                System.Console.Out.WriteLine("    ====================");
#endif
                return toReturn;
            }

            string result = PPL.hide(theConstraint, ClocksToString());

            Constraint newConstraint = new Constraint();
            newConstraint.theConstraint = result;

            // Empty the list of clocks
            newConstraint.ClockList = new List<byte>();

#if DEBUGETIENNE
            newConstraint.debugString = "ABOUT TO HIDE ALL CLOCKS {" + ClocksToString() + "} IN : [\n" + debugString + "\n] // hide \n";
            newConstraint.debugString += "RESULT OF HIDING : \n" + result + " (ID = " + idConstraint + ") \n";
            newConstraint.debugString = ""; // TO GET PRINTING EASIER
#endif

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + newConstraint);
            System.Console.Out.WriteLine("    ====================");
#endif
            return newConstraint;
        }


        /************************************************************
        * OPERATIONS MODIFYING THE OBJECT
        ***********************************************************/

        // Timed delay on the constraint
        public void Delay()
        {
            #if DEBUGETIENNE
            debugString = "ABOUT TO DELAY : [\n" + debugString + "\n] // delay \n";
            #endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Delaying: \n" + this.ToString());
#endif
            if (theConstraint == STRING_TRUE)
            {
#if DEBUGETIENNE
                debugString += "RESULT OF DELAY = true (UNCHANGED)\n";
#endif
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    (unchanged)");
                System.Console.Out.WriteLine("    ====================");
#endif
                return; // OPTIMIZATION
            }
            theConstraint = PPL.time_elapsing(theConstraint, ClocksToString());

            #if DEBUGETIENNE
            debugString += "RESULT OF DELAY = \n" + theConstraint + " (ID = " + idConstraint + ") \n";
            #endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }




        // Add a list of inequalities of the form 'x >= 0', for all 'x' in 'ClockList'
        public void EnsureClocksNonnegative(List<byte> ClockList)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Ensuring that clocks are non negative in: \n" + this);
#endif
            // Create the inequalities
            bool first = true;
            string inequalities = "";
            foreach (byte clockId in ClockList)
            {
                // Add the separator
                if (first)
                {
                    first = false;
                }
                else
                {
                    inequalities += STRING_SEP_AND;
                }
                // Add the inequality
                inequalities += 0 + opToString(TimerOperationType.LessThanOrEqualTo) + ClockToString(clockId);
            }
            // Perform intersection
            theConstraint = PPL.intersection(this.theConstraint, inequalities);
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + theConstraint);
            System.Console.Out.WriteLine("    ====================");
#endif
        }


        // Test if all elements of l1 belong to l2
        private static bool ListIncluded(List<byte> l1, List<byte> l2){
            foreach (byte elem in l1)
            {
                if (!l2.Contains(elem)){
                    return false;
                }
            }
            return true;
        }



        // Computes the difference between l1 \ l2, i.e., the elements of l1 minus the elements of l2
        private static List<byte> ListDiff(List<byte> l1, List<byte> l2)
        {
            List<byte> toReturn = new List<byte>();
            foreach (byte elem in l1)
            {
                if (!l2.Contains(elem))
                {
                    toReturn.Add(elem);
                }
            }
            return toReturn;
        }



        // Conjunction between 2 constraints
        public void Conjunction(Constraint otherConstraint){
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    About to perform conjunction of \n" + this + "\n     with \n" + otherConstraint);
#endif

            // OPTIMIZATION 1
            if (otherConstraint.theConstraint == STRING_TRUE || otherConstraint.theConstraint == theConstraint)
            {
#if DEBUGETIENNE
                debugString = otherConstraint.debugString;
                debugString = "PERFORMED INTERSECTION BETWEEN TRUE OR THIS (ID "
                    + otherConstraint.idConstraint + ") AND [\n " + debugString + "\n] // intersection \n";
#endif
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    (unchanged)");
                System.Console.Out.WriteLine("    ====================");
#endif
                return;
            }

            // OPTIMIZATION 2
            if (theConstraint == STRING_TRUE) {
                theConstraint = otherConstraint.theConstraint;
                ClockList = otherConstraint.ClockList;
#if DEBUGETIENNE
                debugString = otherConstraint.debugString;
                debugString = "PERFORMED INTERSECTION BETWEEN TRUE  (ID "
                    + otherConstraint.idConstraint + ") AND [\n " + debugString + "\n] // intersection\n";
#endif
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    Result (assignation): \n" + this);
                System.Console.Out.WriteLine("    ====================");
#endif
                return;
            }

            // Verifying the equality of the clock list
            if (!(ListIncluded(ClockList, otherConstraint.ClockList)
                && ListIncluded(otherConstraint.ClockList, ClockList)))
            {
                throw new InvalidOperationException("Constraints " + ToString() +
                    " and " + otherConstraint.ToString() + " have different clock lists: {"
                    + ClocksToString() +
                    "} and {" + otherConstraint.ClocksToString() + "}.");
            }

            theConstraint = PPL.intersection(otherConstraint.theConstraint, this.theConstraint);

#if DEBUGETIENNE
            debugString = "(" + debugString + ") (ID "
                    + idConstraint + ") AND (\n" + otherConstraint.debugString + ") (ID "
                    + otherConstraint.idConstraint + ") \n";
            debugString += "RESULT OF INTERSECTION: " + theConstraint + " (ID "
                    + idConstraint + ") \n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }




        // USELESS ACCORDING TO DBM.cs
        //// Check if this is a subset of otherConstraint (or the converse ?!!)
        //public bool IsSubSetOf(Constraint otherConstraint)
        //{
        //    //Debug.Assert(IsCanonicalForm); /* chengbin */

        //    // ETIENNE SAYS: not convinced about those two optimizations
        //    // If other is true: yes
        //    if (otherConstraint.ClockList.Count() == 0)
        //    {
        //        return true;
        //    }

        //    // If this is true: no
        //    if (ClockList.Count() == 0)
        //    {
        //        return false;
        //    }

        //    //string operation = "included(" + this.theConstraint + ", " + otherConstraint.theConstraint + ")";
        //    return PPL.contains(otherConstraint.theConstraint, this.theConstraint);
        //}



        // Addition of a new clock set to 0 (if clockID already in, throws exception)
        public void AddClock(byte clockID)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Adding clock " + clockID + " to : \n" + this);
#endif

            // If the clock already exists, do nothing
            if (ClockList.Contains(clockID))
            {
                throw new InvalidOperationException("Attempt to add clock " + clockID.ToString()
                    + " to the list of clocks '" + ClocksToString() + "'.");
            }
            //}

            // Add the clock
            ClockList.Add(clockID);

            // Add the constraint 'clock = 0'
            AddInequality(clockID, TimerOperationType.Equals, 0);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result of addition: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif

#if DEBUGETIENNE
            debugString = "ADDING CLOCK " + ClockToString(clockID) + " TO [\n"
                + debugString + "] // adding \n RESULT OF ADDING: \n" + theConstraint + "(ID = " + idConstraint + ")";
#endif

        }



        // Add an inequality of the form "clock ~ int"
        public void AddInequality(byte clockID, TimerOperationType op, int constant)
        {
#if DEBUGETIENNE
            debugString = "ADDING INEQUALITY: "
                    + ClockToString(clockID) + opToString(op) + constant
                    + " TO [\n" + debugString + "] // adding ineq \n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Adding inequality " + ClockToString(clockID) + opToString(op) + constant + " to : \n" + this);
#endif



            string inequality = "will be assigned later (function AddInequality)";
            switch (op)
            {
                case TimerOperationType.Equals:
                    // Add two inequalities
                    inequality = constant.ToString() + opToString(TimerOperationType.LessThanOrEqualTo) + ClockToString(clockID)
                        + STRING_SEP_AND
                        + ClockToString(clockID) + opToString(TimerOperationType.LessThanOrEqualTo) + constant.ToString();
                    break;
                case TimerOperationType.GreaterThanOrEqualTo:
                    // Reverse the inequality
                    inequality = constant.ToString() + opToString(TimerOperationType.LessThanOrEqualTo) + ClockToString(clockID);
                    break;
                case TimerOperationType.LessThanOrEqualTo:
                    // Keep the inequality
                    inequality = ClockToString(clockID) + opToString(TimerOperationType.LessThanOrEqualTo) + constant.ToString();
                    break;
                case TimerOperationType.GreaterThan:
                    throw new InvalidOperationException(
                            "Invalid operator '" + opToString(op)
                            + "' detected in an inequality. Allowed: "
                            + opToString(TimerOperationType.LessThanOrEqualTo) + " only."
                        );
                case TimerOperationType.LessThan:
                    throw new InvalidOperationException(
                            "Invalid operator '" + opToString(op)
                            + "' detected in an inequality. Allowed: "
                            + opToString(TimerOperationType.LessThanOrEqualTo) + " only."
                        );
            }
            this.theConstraint = PPL.intersection(this.theConstraint, inequality);

#if DEBUGETIENNE
            debugString += "RESULT OF ADDING INEQUALITY: \n" + theConstraint + " (ID = " + idConstraint + ") \n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }



        // Add an inequality of the form "clock ~ parameter"
        public void AddParInequality(byte clockID, TimerOperationType op, string parameter)
        {
#if DEBUGETIENNE
            debugString = "ADDING PAR INEQUALITY: "
                    + ClockToString(clockID) + opToString(op) + parameter
                    + " TO [\n" + debugString + "] // adding parineq \n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Adding par inequality " + ClockToString(clockID) + opToString(op) + parameter + " to : \n" + this);
#endif

            string inequality = "will be assigned later (function AddInequality)";
            switch (op)
            {
                case TimerOperationType.Equals:
                    inequality = parameter + opToString(TimerOperationType.Equals) + ClockToString(clockID);
                    break;
                case TimerOperationType.GreaterThanOrEqualTo:
                    // Reverse the inequality
                    inequality = parameter + opToString(TimerOperationType.LessThanOrEqualTo) + ClockToString(clockID);
                    break;
                case TimerOperationType.LessThanOrEqualTo:
                    // Keep the inequality
                    inequality = ClockToString(clockID) + opToString(TimerOperationType.LessThanOrEqualTo) + parameter;
                    break;
                case TimerOperationType.GreaterThan:
                    // Reverse the inequality
                    inequality = parameter + opToString(TimerOperationType.LessThan) + ClockToString(clockID);
                    break;
                case TimerOperationType.LessThan:
                    // Keep the inequality
                    inequality = ClockToString(clockID) + opToString(TimerOperationType.LessThan) + parameter;
                    break;
            }
            this.theConstraint = PPL.intersection(this.theConstraint, inequality);

#if DEBUGETIENNE
            debugString += "RESULT OF ADDING PARINEQ: \n" + theConstraint + " (ID = " + idConstraint + ") \n";
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }

        // Add a list of constraints
        // ETIENNE SAYS: there seem to be something strange here; contrarily to Conjunction, we do not test the list of clocks :(
        public void AddConstraints(List<Constraint> constraints)
        {
            // Creation of the string constraint
            string otherConstraint = "";
            bool first = true;
            foreach(Constraint constraint in constraints){
                // OPTIMIZATION
                if (constraint.theConstraint != STRING_TRUE)
                {
                    if (first)
                    {
                        first = false;
                    }
                    else
                    {
                        otherConstraint += STRING_SEP_AND;
                    }
                    otherConstraint += constraint.theConstraint;
                }

#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    ====================");
                System.Console.Out.WriteLine("    Adding inequalities " + otherConstraint + " to : \n" + this);
#endif
            }
#if DEBUGETIENNE
            debugString = "ADDING CONSTRAINTS: "
                    + otherConstraint
                    + " TO [\n" + debugString + "] // adding constraints \n";
#endif
            // Perform actual intersection
            this.theConstraint = PPL.intersection(this.theConstraint, otherConstraint);
#if DEBUGETIENNE
            debugString += "RESULT OF ADDITION OF CONSTRAINTS: \n" + theConstraint + " (ID = " + idConstraint + ") \n";
#endif

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }

        /************************************************************
        * CONVERSION TO STRING
        ***********************************************************/
        

        public static string ClockToString(byte clockID)
        {
            return STRING_CLOCK_NAME + clockID.ToString();
        }


        public static string opToString(TimerOperationType op)
        {
            string result = "will be assigned later (function opToString)";
            switch (op)
            {
                case TimerOperationType.Equals:
                    result = "="; break;
                case TimerOperationType.GreaterThanOrEqualTo:
                    result = ">="; break;
                case TimerOperationType.LessThanOrEqualTo:
                    result = "<="; break;
                case TimerOperationType.GreaterThan:
                    result = ">"; break;
                case TimerOperationType.LessThan:
                    result = "<"; break;
            }
            return result;
        }


        // Conversion to string
        public override string ToString()
        {
            // IMPORTANT: copy
            string theString = String.Copy(theConstraint);

            // Add a mention if true
            if (theString == NonNegativeParametersConstraint)
            {
                theString += "\n (TRUE)";
            }
            
            // Add some blanks
            theString = theString.Replace(STRING_SEP_AND, "\n" + STRING_SEP_AND + " ");
            // Temp replace
            theString = theString.Replace("<=", "~");
            theString = theString.Replace("<", " < ");
            theString = theString.Replace("=", " = ");
            theString = theString.Replace("~", " <= ");
#if DEBUGETIENNE
            theString += "\nID = " + idConstraint + "\n";
            //theString += "\nDEBUG: " + debugString + "\n";
#endif
#if DEBUGCONSOLE
            theString += "\nClocks: " + ClocksToString();
#endif
            return "    " + theString;
        }



        // Convert a list of clocks to a string
        public static string SomeClocksToString(List<byte> ClockList)
        {
            // Case null
            if (ClockList == null) return "null";

            // Convert each clock to a string
            List<string> clocks_string = new List<string>(ClockList.Count());
            foreach (byte clockID in ClockList)
            {
                clocks_string.Add(ClockToString(clockID));
            }
            // Convert everything with a separator
            return string.Join(STRING_SEP_SET, clocks_string.ToArray());
        }



        // Convert all clocks to string
        public string ClocksToString()
        {
            return SomeClocksToString(ClockList);
        }


        /*
        // Return a unique identifier
        // Note that the constraints are (should be?) under a kind of canonical form
        public string GetID()
        {
            return theConstraint;
        }
        */
        
        // Return a unique identifier
        // Note that the constraints are (should be?) under a kind of canonical form
        // Clocks are "collapsed", so that constraint equivalent but with different clocks permutations still have the same id
        public string GetID(List<byte> OldClockList)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("\nCREATION OF CONSTRAINT ID");
#endif

            // 1) Compute the correspondence old -> new
            Dictionary<byte, byte> correspondence = new Dictionary<byte, byte>();
            byte CurrentNewId = FIRST_CLOCK;
            foreach (byte OldClockId in OldClockList)
            {
                // Check if already captured (indeed, the list ClockList may contain the same clock several times!)
                if (!correspondence.ContainsKey(OldClockId))
                {
                    correspondence.Add(OldClockId, CurrentNewId);
                    // Update the new id for the next clock
                    CurrentNewId++;
                }
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("  Correspondance:");
            foreach (byte OldClockId in OldClockList)
            {
                // Find the new value in the correspondence table
                byte NewId = 0;
                correspondence.TryGetValue(OldClockId, out NewId);
                System.Console.Out.WriteLine("    " + OldClockId + " –> " + NewId);
            }
#endif


            // 2) Check that all clocks are captured by the correspondence
            // ETIENNE SAYS: this is just a test, should be removed if functional!
            foreach (byte ClockId in ClockList)
            {
                if (!correspondence.ContainsKey(ClockId))
                {
                    //throw new Exception("Inconsistent clock lists in GetID()."
                    //    + "\n ClockList as a parameter: " + SomeClocksToString(OldClockList)
                    //    + "\n ClockList in Constraint: " + ClocksToString()
                    //    + "\n Constraint: " + theConstraint);
                }
            }

            // 3) Create the list of new clocks
            string Id = "";
            foreach (byte clockID in OldClockList)
            {
                // Find the new value in the correspondence table
                byte NewId = 0;
                correspondence.TryGetValue(clockID, out NewId);
                Id += NewId + Constants.SEPARATOR;
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("  Id: " + Id);
#endif

            // 4) Replace the clock names in constraint
            // ETIENNE SAYS: very ugly, and very dangerous! (if "x1" is used as a parameter, for instance!)
            // First replace the clock names with an independent clock name
            // Pattern explanation:
            // - ([^A-Za-z_]|^) : one non letter and non '_' sign, or the beginning of the string
            // - STRING_CLOCK_NAME : the letter used to distinguish strings
            // - (\d+) : several digits
            //// - (\W|$) : one non letter, non '_' and non digit sign, or the end of the string
            // IMPERFECT SOLUTION: will react to x31z - but I have nothing better to suggest
            string pattern = @"([^A-Za-z_]|^)" + STRING_CLOCK_NAME + @"(\d+)"; // (\W|$)
            //theConstraint = "x23+x+bz32+x57+bx27+x31z<=p32+px21+x+x213+xp57+x317&x2<=x3&x2<=x1";
            string replacement = "$1" + STRING_CLOCK_NAME_TMP + "$2"; // $3
            Regex rgx = new Regex(pattern);
            string newConstraint = rgx.Replace(theConstraint, replacement);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("  Replacement 1: '" + theConstraint + "'  –>  '" + newConstraint + "'");
#endif


            // For each clock, replace by its new value
            foreach (byte OldClockId in ClockList)
            {
                // Get the new id
                byte NewId = 0;
                correspondence.TryGetValue(OldClockId, out NewId);
                // Replace it in string
                newConstraint = newConstraint.Replace(STRING_CLOCK_NAME_TMP + OldClockId, STRING_CLOCK_NAME + NewId);
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("  Replacement 2: " + newConstraint);
#endif

            // Put the constraint under normal form
            newConstraint = PPL.simplify(newConstraint);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("  Simplification: " + newConstraint);
#endif

            Id += newConstraint;

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("END ID: " + Id);
#endif
            // 5) Return
            return Id;
        }


        /************************************************************
        * CLONE
        ***********************************************************/

        // Copy
        public Constraint Clone()
        {
            Constraint newConstraint = new Constraint();
            newConstraint.theConstraint = theConstraint;
            
            // Copy the list of clocks
            newConstraint.ClockList = new List<byte>(ClockList);

#if DEBUGETIENNE
            newConstraint.debugString = debugString;
#endif
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Cloning: \n" + this);
            System.Console.Out.WriteLine("    Result: \n" + newConstraint);
            System.Console.Out.WriteLine("    ====================");
#endif
            return newConstraint;
        }



    }
}
