using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class DataOperationPrefix : PARProcess
    {
        public  Event Event;
        public  PARProcess Process;
        public  Expression AssignmentExpr;
        public  string[] LocalVariables;
        public bool IsUrgent;

        public DataOperationPrefix(Event e, Expression assignment, PARProcess process, string[] localvar, bool isUrgent)
        {
            Event = e;
            AssignmentExpr = assignment;
            Process = process;
            LocalVariables = localvar;
            IsUrgent = isUrgent;

            ProcessID = PARDataStore.DataManager.InitializeProcessID(Event.GetID() + "{" + AssignmentExpr.ExpressionID + "}" + (IsUrgent ? Constants.EVENTPREFIX_URGENT : Constants.EVENTPREFIX) + Process.ProcessID);              
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> list = new List<PARConfiguration>();
            string eventID = Event.GetEventID(eStep.GlobalEnv);

            Valuation newGlobleEnv = eStep.GlobalEnv.GetVariableClone();

            EvaluatorDenotational.Evaluate(AssignmentExpr, newGlobleEnv);

            if (LocalVariables != null)
            {
                Valuation tempEnv = eStep.GlobalEnv.GetVariableClone();
                for (int i = 0; i < tempEnv.Variables._entries.Length; i++)
                {
                    StringDictionaryEntryWithKey<ExpressionValue> pair = tempEnv.Variables._entries[i];
                    if (pair != null)
                    {
                        pair.Value = newGlobleEnv.Variables[pair.Key];
                    }
                }

                newGlobleEnv = tempEnv;
            }

            string eventName = Event.GetEventName(eStep.GlobalEnv);

            list.Add(new PARConfiguration(Process, eventID, eventID != eventName ? eventName : null, newGlobleEnv, true,
                                          IsUrgent, eStep.TheConstraint));
            return list;
        }

  
        public override string ToString()
        {
            return "(" + Event + "{" + AssignmentExpr + "}" + (IsUrgent ? "->>" : "->") + Process.ToString() + ")";
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return Process.GetAlphabets(visitedDefinitionRefs);
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = Process.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, AssignmentExpr.GetVars());

            if (Event.ExpressionList != null)
            {
                foreach (Expression expression in Event.ExpressionList)
                {
                    Common.Classes.Ultility.Ultility.AddList(Variables, expression.GetVars());
                }
            }

            return Variables;
        }

        public override List<string> GetChannels()
        {
            return Process.GetChannels();
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Expression newAssign = AssignmentExpr.ClearConstant(constMapping);
            return new DataOperationPrefix(Event.ClearConstant(constMapping), newAssign, Process.ClearConstant(constMapping), LocalVariables, IsUrgent);
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        //public override bool IsTime()
        //{
        //    return Process.IsTime();
        //}

      }
}