using System.Text;
using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class DeadlineProcess : PARProcess
    {
        public  PARProcess Process;
        public  Expression DelayTime;
        public byte TimerID;

        public DeadlineProcess(PARProcess process, Expression cond, byte timerID)
        {
            Process = process;
            DelayTime = cond;
            TimerID = timerID;
            
            StringBuilder ID = new StringBuilder(Process.ProcessID);
            ID.Append(Constants.DEADLIME + "[" + cond.ExpressionID + "]");

            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID.ToString());
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = Process.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());

            return Variables;
        }

        public override List<string> GetChannels()
        {
            return Process.GetChannels();
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> steps = Process.MoveOneStep(eStep);

            int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;
            eStep.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event != Constants.TERMINATION)
                {
                    vm.Process = new DeadlineProcess(vm.Process, DelayTime, TimerID);
                }
            }

            return steps;
        }


        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> steps = Process.SyncOutput(eStep);

            foreach (PARConfigurationWithChannelData vm in steps)
            {
                vm.Process = new DeadlineProcess(vm.Process, DelayTime, TimerID);
            }

            return steps;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> steps = Process.SyncInput(eStep);

            foreach (PARConfiguration vm in steps)
            {
                vm.Process = new DeadlineProcess(vm.Process, DelayTime, TimerID);
            }

            return steps;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + Process + " deadline[" + DelayTime + "])";
            }
            else
            {
                return "(" + Process + " deadline[" + DelayTime + "]" + Constraint.ClockToString(TimerID) + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return Process.GetAlphabets(visitedDefinitionRefs);
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new DeadlineProcess(Process.ClearConstant(constMapping), DelayTime.ClearConstant(constMapping), TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, Process.GetActiveTimer());//eStep));
        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = Process.SetTimer(timerID, clocks);

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddInequality(TimerID, DelayTime);
                    return new DeadlineProcess(process, DelayTime, TimerID);
                }

                clocks.AddInequality(timerID, DelayTime);
                return new DeadlineProcess(process, DelayTime, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddInequality(TimerID, DelayTime);
                return null;
            }

            clocks.AddInequality(timerID, DelayTime);
            return new DeadlineProcess(Process, DelayTime, timerID);
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return Process.GetTopLevelConcurrency(visitedDef);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}