using System.Collections.Generic;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;

namespace PAT.PAR.LTS
{
    public sealed class Definition
    {
        public string Name;
        public PARProcess Process;
        public string[] LocalVariables;

        //=========================Fileds for Static Analysis================================
        public List<Definition> SubDefinitions;
        public List<string> SubDefinitionNames;
        //public bool IsTimed;

        public List<string> GlobalVars;
        public List<string> Channels;
        public bool MustAbstract;
        //public bool IsProbabilistic;

        public EventCollection AlphabetEvents;
        public HashSet<string> Alphabets;

        //after static analysis, AlphabetsCalculable is true if and only if Alphabets are not null.
        public bool AlphabetsCalculable;
        public bool UsedInParallel;
        //=========================Fileds for Static Analysis================================


        public Definition(string name, string[] vars, PARProcess process)
        {
            Name = name;
            Process = process;
            LocalVariables = vars;

            SubDefinitions = new List<Definition>();
            SubDefinitionNames = new List<string>();

            Channels = new List<string>();
            GlobalVars = new List<string>();

            AlphabetsCalculable = true;
            Alphabets = new HashSet<string>();           
        }

        public override string ToString()
        {
            return Process.ToString();
        }

        public string GetFullDefinition()
        {
            return Name + "(" + Common.Classes.Ultility.Ultility.PPStringList(LocalVariables) + ")=" + Process + ";";
        }

        public Definition ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Process = Process.ClearConstant(constMapping);
            return this;
        }

        /// <summary>
        /// To Perform the static analysis on a single definition first.
        /// </summary>
        public void StaticAnalysis()
        {
            MustAbstract = Process.MustBeAbstracted();

            Channels = Process.GetChannels();

            GlobalVars = Process.GetGlobalVariables();

            //IsTimed = Process.IsTime();
            //IsProbabilistic = Process.IsProbabilistic();

            if (AlphabetEvents != null)
            {
                if (AlphabetEvents.ContainsVariable())
                {
                    AlphabetsCalculable = false;
                    Alphabets = null;                    
                }
                else
                {
                    Alphabets = new HashSet<string>(new EventCollection(AlphabetEvents).EventNames);
                    AlphabetsCalculable = true;                    
                }
                //try
                //{
                //    Alphabets = new EventCollection(AlphabetEvents).EventNames;
                //    AlphabetsCalculable = true;
                //}
                //catch (Exception)
                //{
                //    AlphabetsCalculable = false;
                //    Alphabets = null;
                //}                
            }
            else
            {
                if (AlphabetsCalculable)
                {
                    Alphabets = Process.GetAlphabets(null);
                }
            }            
        }
    }
}