using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class IndexChoice : PARProcess
    {
        public List<PARProcess> Processes;
        public IndexedProcess IndexedProcessDefinition;

        /// <summary>
        /// constructor for indexed choice.
        /// </summary>
        /// <param name="definition"></param>
        public IndexChoice(IndexedProcess definition)
        {
            IndexedProcessDefinition = definition;
        }

        public IndexChoice(List<PARProcess> processes)
        {
            Processes = processes;
            StringBuilder ID = new StringBuilder();
            foreach (PARProcess processBase in Processes)
            {
                ID.Append(Constants.GENERAL_CHOICE);
                ID.Append(processBase.ProcessID);
            }

            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID.ToString());
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> toReturn = new List<PARConfiguration>();

            for (int i = 0; i < Processes.Count; i++)
            {
                toReturn.AddRange(Processes[i].MoveOneStep(eStep));
            }

            return toReturn;
        }

        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> toReturn = new List<PARConfigurationWithChannelData>();

            for (int i = 0; i < Processes.Count; i++)
            {
                toReturn.AddRange(Processes[i].SyncOutput(eStep));
            }
            return toReturn;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> toReturn = new List<PARConfiguration>();

            for (int i = 0; i < Processes.Count; i++)
            {
                toReturn.AddRange(Processes[i].SyncInput(eStep));
            }

            return toReturn;
        }


        public override string ToString()
        {
            if (Processes == null)
            {
                return "[]" + IndexedProcessDefinition;
            }

            string result = "(" + Processes[0];
            for (int i = 1; i < Processes.Count; i++)
            {
                result += "[]";
                result += Processes[i].ToString();
            }
            result += ")";
            return result;
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetAlphabets(visitedDefinitionRefs);
            }

            HashSet<string> list = new HashSet<string>();
            for (int i = 0; i < Processes.Count; i++)
            {
                list.UnionWith(Processes[i].GetAlphabets(visitedDefinitionRefs));
            }
            return list;
        }

        public override List<string> GetGlobalVariables()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetGlobalVariables();
            }

            List<string> Variables = new List<string>();
            for (int i = 0; i < Processes.Count; i++)
            {
                List<string> temp = Processes[i].GetGlobalVariables();
                Common.Classes.Ultility.Ultility.AddList<string>(Variables, temp);
            }
            return Variables;
        }

        public override List<string> GetChannels()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetChannels();
            }

            List<string> channels = new List<string>();
            for (int i = 0; i < Processes.Count; i++)
            {
                List<string> temp = Processes[i].GetChannels();
                Common.Classes.Ultility.Ultility.AddList<string>(channels, temp);
            }
            return channels;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            List<PARProcess> newnewListProcess = Processes;
            if (Processes == null)
            {
                if (Specification.IsParsing)
                {
                    return new IndexChoice(IndexedProcessDefinition.ClearConstant(constMapping));
                }

                //return new IndexChoice(IndexedProcessDefinition.GetIndexedProcesses(constMapping));
                newnewListProcess = IndexedProcessDefinition.GetIndexedProcesses(constMapping);
            }
            List<PARProcess> newListProcess = new List<PARProcess>();
            for (int i = 0; i < newnewListProcess.Count; i++)
            {
                PARProcess newProc = newnewListProcess[i].ClearConstant(constMapping);
                if (!(newProc is Stop))
                {
                    newListProcess.Add(newProc);
                }
            }

            //all processes are stop processes
            if (newListProcess.Count == 0)
            {
                return new Stop();
            }

            return new IndexChoice(newListProcess);
        }

        public override bool MustBeAbstracted()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.MustBeAbstracted();
            }

            for (int i = 0; i < Processes.Count; i++)
            {
                if (Processes[i].MustBeAbstracted())
                {
                    return true;
                }
            }

            return false;
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> activeTimers = new List<byte>();

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        Common.Classes.Ultility.Ultility.Union(activeTimers, Processes[i].GetActiveTimer());
        //    }

        //    return activeTimers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            List<PARProcess> newListProcess = new List<PARProcess>();

            bool anyNeedtimer = false;
            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess newProc = Processes[i].SetTimer(timerID, clocks);
                if (newProc == null)
                {
                    newListProcess.Add(Processes[i]);
                }
                else
                {
                    anyNeedtimer = true;
                    newListProcess.Add(newProc);
                }
            }

            //all processes are stop processes
            if (anyNeedtimer)
            {
                IndexChoice index = new IndexChoice(newListProcess);
                return index;
            }

            return null;
        }

        //public override bool IsTime()
        //{
        //    if (Processes == null)
        //    {
        //        return IndexedProcessDefinition.Process.IsTime();
        //    }

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        if (Processes[i].IsTime())
        //        {
        //            return true;
        //        }
        //    }
        //    return false;
        //}

    }
}