using System.Collections.Generic;
using System.Diagnostics;
using PAT.Common;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class IndexInterleaveAbstract : PARProcess
    {
        public List<PARProcess> Processes;
        public Dictionary<string, int> ProcessesCounter;
        public Dictionary<string, int> ProcessesActualSize;
        public Expression RangeExpression;
        private bool MustAbstract;

        public IndexInterleaveAbstract(PARProcess processBase, Expression range)
        {
            Processes = new List<PARProcess>();
            Processes.Add(processBase);
            RangeExpression = range;
        }

        public IndexInterleaveAbstract(PARProcess processBase, int size)
        {
            Processes = new List<PARProcess>();
            Processes.Add(processBase);

            ProcessesCounter = new Dictionary<string, int>();
            ProcessesCounter.Add(processBase.ProcessID, size);

            ProcessesActualSize = new Dictionary<string, int>();
            ProcessesActualSize.Add(processBase.ProcessID, size);
            MustAbstract = size == -1;
        }

        public IndexInterleaveAbstract(List<PARProcess> processes, Dictionary<string, int> counters)
        {
            Processes = new List<PARProcess>();
            ProcessesCounter = new Dictionary<string, int>();
            //IdleTimeConstraints = new List<ClockConstraints>();
 
            List<string> tmp = new List<string>(processes.Count);
            for (int i = 0; i < processes.Count; i++)
            {
                string id = processes[i].ProcessID;
                if (counters[id] != 0)
                {
                    tmp.Add(id);
                    Processes.Add(processes[i]);
                    //IdleTimeConstraints.AddRange(processes[i].IdleTimeConstraints);
                    ProcessesCounter.Add(id, counters[id]);
                }
            }

            tmp.Sort();

            ProcessID = Constants.INTERLEAVE + tmp[0] + "*" + ProcessesCounter[tmp[0]];
            for (int i = 1; i < Processes.Count; i++)
            {
                ProcessID += Constants.SEPARATOR + tmp[i] + "*" + ProcessesCounter[tmp[i]];
            }

            ProcessID = PARDataStore.DataManager.InitializeProcessID(ProcessID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();
            List<Dictionary<string, int>> nextProcessCounters = null;
            List<List<PARConfiguration>> terminateSteps = new List<List<PARConfiguration>>();

            for (int i = 0; i < Processes.Count; i++)
            {
                terminateSteps.Add(new List<PARConfiguration>());
                PARProcess process = Processes[i];
                List<PARConfiguration> list1 = process.MoveOneStep(eStep);

                if (list1.Count > 0)
                {
                    nextProcessCounters = Common.Classes.Ultility.Ultility.ProcessCounterDecrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, ProcessesCounter, process.ProcessID, 1);
                }

                for (int j = 0; j < list1.Count; j++)
                {
                    PARConfiguration step = list1[j];

                    if (step.Event != Constants.TERMINATION)
                    {
                        foreach (Dictionary<string, int> ints in nextProcessCounters)
                        {
                            Dictionary<string, int> listInstance = new Dictionary<string, int>(ints);

                            List<PARProcess> newProcess = new List<PARProcess>(Processes);
                            AddOneProcess(newProcess, step.Process, listInstance);
                            IndexInterleaveAbstract interleave = new IndexInterleaveAbstract(newProcess, listInstance);

                            PARConfiguration newStep = new PARConfiguration(interleave, step.Event, step.DisplayName, step.GlobalEnv, step.IsDataOperation, step.IsUrgent, step.TheConstraint);
                            newStep.IsAtomic = step.IsAtomic;

                            if (AssertionBase.CalculateParticipatingProcess)
                            {
                                newStep.ParticipatingProcesses = new string[] { process.ProcessID };
                            }

                            returnList.Add(newStep);
                        }
                    }
                    else
                    {
                        terminateSteps[i].Add(step);
                    }
                }

                //to check whether there are synchoronous channel input/output
                if (SpecificationBase.HasSyncrhonousChannel)
                {
                    SynchronousChannelInputOutput(returnList, i, eStep, null);
                }
            }

            bool allReadyToTerminate = true;
            foreach (List<PARConfiguration> rtsConfigurations in terminateSteps)
            {
                if (rtsConfigurations.Count == 0)
                {
                    allReadyToTerminate = false;
                    break;
                }
            }

            if (allReadyToTerminate)
            {
                returnList.AddRange(TerminationSteps(eStep, terminateSteps));
            }
            return returnList;
        }




        private List<PARConfiguration> TerminationSteps (PARConfiguration eStep, List<List<PARConfiguration>> terminiateSteps) 
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();
            List<bool> isAtomic = new List<bool>();

            //maps an event to the list of resulting processes.
            isAtomic.Add(false);

            List<List<Constraint>> syncedConstraints = new List<List<Constraint>>();
            syncedConstraints.Add(new List<Constraint>());

            for (int i = 0; i < Processes.Count; i++)
            {
                List<PARConfiguration> steps = terminiateSteps[i];

                if (steps.Count == 0)
                {
                    return returnList;
                }

                foreach (PARConfiguration step in steps)
                {
                    if (syncedConstraints[0].Count == i)
                    {
                        for (int j = 0; j < syncedConstraints.Count; j++)
                        {
                            syncedConstraints[j].Add(step.TheConstraint);

                            if (step.IsAtomic)
                            {
                                for (int k = 0; k < isAtomic.Count; k++)
                                {
                                    isAtomic[k] = true;
                                }
                            }
                        }
                    }
                    else
                    {
                        //If there non-determinism, clone and then add.
                        List<List<Constraint>> toAddConstraint = new List<List<Constraint>>(syncedConstraints.Count);

                        for (int j = 0; j < syncedConstraints.Count; j++)
                        {
                            List<Constraint> newConstraintList = new List<Constraint>();
                            List<Constraint> Constraintlist = syncedConstraints[j];

                            for (int k = 0; k < Constraintlist.Count - 1; k++)
                            {
                                newConstraintList.Add(Constraintlist[k]);
                            }

                            newConstraintList.Add(step.TheConstraint);
                            toAddConstraint.Add(newConstraintList);
                            isAtomic.Add(step.IsAtomic);
                        }

                        syncedConstraints.AddRange(toAddConstraint);
                    }
                }
            }

            for (int i = 0; i < syncedConstraints.Count; i++)
            {
                Constraint newconstraint = Constraint.constraintWithNonNegativeParameters(); // trueConstraint
                bool isAtomicBool = false;
                for (int k = 0; k < syncedConstraints[i].Count; k++)
                {
                    newconstraint.Conjunction(syncedConstraints[i][k]);
                    isAtomicBool = isAtomicBool || isAtomic[i];
                }

                PARConfiguration tmpStep = new PARConfiguration(new Stop(), Constants.TERMINATION, null, eStep.GlobalEnv, false, false, newconstraint);
                tmpStep.IsAtomic = isAtomicBool;

                if (AssertionBase.CalculateParticipatingProcess)
                {
                    tmpStep.ParticipatingProcesses = new string[Processes.Count];
                    for (int j = 0; j < Processes.Count; j++)
                    {
                        tmpStep.ParticipatingProcesses[j] = Processes[j].ProcessID;
                    }
                }

                returnList.Add(tmpStep);
            }

            return returnList;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> returnList = new List<PARConfigurationWithChannelData>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                List<PARConfigurationWithChannelData> list1 = process.SyncOutput(eStep);

                if (list1.Count > 0)
                {
                    List<Dictionary<string, int>> nextProcessCounters = Common.Classes.Ultility.Ultility.ProcessCounterDecrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, ProcessesCounter, process.ProcessID, 1);

                    for (int j = 0; j < list1.Count; j++)
                    {
                        PARConfiguration step = list1[j];

                        foreach (Dictionary<string, int> ints in nextProcessCounters)
                        {
                            List<PARProcess> newProcess = new List<PARProcess>(Processes);

                            Dictionary<string, int> listInstance = new Dictionary<string, int>(ints);
                            AddOneProcess(newProcess, step.Process, listInstance);
                            IndexInterleaveAbstract interleave = new IndexInterleaveAbstract(newProcess, listInstance);

                            PARConfigurationWithChannelData newStep = new PARConfigurationWithChannelData(interleave, step.Event, step.DisplayName, step.GlobalEnv, step.IsDataOperation, step.IsUrgent, list1[j].ChannelName, list1[j].Expressions, list1[j].TheConstraint, eStep.TheConstraint);
                            newStep.IsAtomic = step.IsAtomic;
                            returnList.Add(newStep);
                        }
                    }
                }
            }

            return returnList;
        }




        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();

            List<Dictionary<string, int>> nextProcessCounters = null;

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                List<PARConfiguration> list1 = process.SyncInput(eStep);

                if (list1.Count > 0)
                {
                    nextProcessCounters = Common.Classes.Ultility.Ultility.ProcessCounterDecrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, ProcessesCounter,
                                                                           process.ProcessID, 1);
                }

                for (int j = 0; j < list1.Count; j++)
                {
                    PARConfiguration step = list1[j];

                    foreach (Dictionary<string, int> ints in nextProcessCounters)
                    {
                        Dictionary<string, int> listInstance = new Dictionary<string, int>(ints);

                        List<PARProcess> newProcess = new List<PARProcess>(Processes);

                        AddOneProcess(newProcess, step.Process, listInstance);

                        IndexInterleaveAbstract interleave = new IndexInterleaveAbstract(newProcess, listInstance);

                        PARConfiguration newStep = new PARConfiguration(interleave, step.Event, step.DisplayName,
                                                                        step.GlobalEnv, step.IsDataOperation, step.IsUrgent, step.TheConstraint);
                        newStep.IsAtomic = step.IsAtomic;

                        if (AssertionBase.CalculateParticipatingProcess)
                        {
                            newStep.ParticipatingProcesses = new string[] { process.ProcessID };
                        }

                        returnList.Add(newStep);
                    }
                }
            }

            return returnList;
        }




        private void SynchronousChannelInputOutput(List<PARConfiguration> returnList, int i, PARConfiguration step, string evt)
        {
            List<PARConfigurationWithChannelData> outputs = Processes[i].SyncOutput(step);

            List<Dictionary<string, int>> nextProcessCounters1 =
                Common.Classes.Ultility.Ultility.ProcessCounterDecrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, ProcessesCounter, Processes[i].ProcessID, 1);

            foreach (PARConfigurationWithChannelData vm in outputs)
            {
                if (evt != null & vm.Event != evt)
                {
                    continue;
                }

                for (int k = 0; k < Processes.Count; k++)
                {
                    string id = Processes[k].ProcessID;

                    if (k != i || (k == i && (ProcessesCounter[id] > 1 || ProcessesCounter[id] == -1)))
                    {
                        List<PARConfiguration> syncedProcess = Processes[k].SyncInput(vm);
                        if (syncedProcess.Count > 0)
                        {
                            if (k == i)
                            {
                                List<Dictionary<string, int>> nextProcessCountersInner = Common.Classes.Ultility.Ultility.ProcessCounterDecrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, ProcessesCounter, Processes[i].ProcessID, 2);

                                foreach (PARConfiguration p in syncedProcess)
                                {
                                    foreach (Dictionary<string, int> ints in nextProcessCountersInner)
                                    {
                                        Dictionary<string, int> dictionaryNew =
                                            new Dictionary<string, int>(ints);

                                        List<PARProcess> newProcess = new List<PARProcess>(Processes);

                                        AddOneProcess(newProcess, p.Process, dictionaryNew);
                                        AddOneProcess(newProcess, vm.Process, dictionaryNew);
                                        Constraint constraint = step.TheConstraint.Clone();
                                        constraint.Conjunction(p.TheConstraint);

                                        PARConfiguration newStep = new PARConfiguration(new IndexInterleaveAbstract(newProcess, dictionaryNew), vm.Event, vm.DisplayName, vm.GlobalEnv, false, vm.IsUrgent || p.IsUrgent, constraint);
                                        newStep.IsAtomic = vm.IsAtomic || p.IsAtomic;

                                        //todo: the following doesn't make sense
                                        if (AssertionBase.CalculateParticipatingProcess)
                                        {
                                            newStep.ParticipatingProcesses = new string[]
                                                                             {
                                                                                 newProcess[i].ProcessID,
                                                                                 newProcess[k].ProcessID
                                                                             };
                                        }

                                        returnList.Add(newStep);
                                    }
                                }
                            }
                            else
                            {
                                foreach (Dictionary<string, int> ints in nextProcessCounters1)
                                {
                                    List<Dictionary<string, int>> nextProcessCountersInner = Common.Classes.Ultility.Ultility.ProcessCounterDecrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, ints, id, 1);

                                    foreach (PARConfiguration p in syncedProcess)
                                    {
                                        foreach (Dictionary<string, int> dictionary in nextProcessCountersInner)
                                        {
                                            List<PARProcess> newProcess = new List<PARProcess>(Processes);

                                            AddOneProcess(newProcess, p.Process, dictionary);
                                            AddOneProcess(newProcess, vm.Process, dictionary);

                                            Constraint constraint = step.TheConstraint.Clone();
                                            constraint.Conjunction(p.TheConstraint);

                                            PARConfiguration newStep = new PARConfiguration(new IndexInterleaveAbstract(newProcess, dictionary), vm.Event, vm.DisplayName, vm.GlobalEnv, false, vm.IsUrgent || p.IsUrgent, constraint);
                                            newStep.IsAtomic = vm.IsAtomic || p.IsAtomic;

                                            if (AssertionBase.CalculateParticipatingProcess)
                                            {
                                                newStep.ParticipatingProcesses = new string[]
                                                                             {
                                                                                 Processes[i].ProcessID,
                                                                                 Processes[k].ProcessID
                                                                             };
                                            }

                                            returnList.Add(newStep);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        private void AddOneProcess(List<PARProcess> processes, PARProcess newProcess, Dictionary<string, int> counter)
        {
            if (newProcess is IndexInterleave)
            {
                IndexInterleave index = newProcess as IndexInterleave;

                foreach (PARProcess processe in index.Processes)
                {
                    string tmp = processe.ProcessID;

                    if (!counter.ContainsKey(tmp))
                    {
                        counter.Add(tmp, 1);
                        processes.Add(processe);
                    }
                    else
                    {
                        counter[tmp] = Common.Classes.Ultility.Ultility.ProcessCounterIncrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, counter[tmp], 1);
                    }
                }
            }
            else if (newProcess is IndexInterleaveAbstract)
            {
                IndexInterleaveAbstract index = newProcess as IndexInterleaveAbstract;

                foreach (PARProcess processe in index.Processes)
                {
                    string tmp = processe.ProcessID;

                    if (!counter.ContainsKey(tmp))
                    {
                        counter.Add(tmp, index.ProcessesCounter[tmp]);
                        processes.Add(processe);
                    }
                    else
                    {
                        counter[tmp] = Common.Classes.Ultility.Ultility.ProcessCounterIncrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, counter[tmp], index.ProcessesCounter[tmp]);
                    }
                }
            }
            else
            {
                string tmp = newProcess.ProcessID;
                if (!counter.ContainsKey(tmp))
                {
                    counter.Add(tmp, 1);
                    processes.Add(newProcess);
                }
                else
                {
                    counter[tmp] = Common.Classes.Ultility.Ultility.ProcessCounterIncrement(MustAbstract ? Common.Classes.Ultility.Ultility.CutNumber : -1, counter[tmp], 1);
                }
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> toReturn = new HashSet<string>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                toReturn.UnionWith(process.GetAlphabets(visitedDefinitionRefs));
            }

            return toReturn;
        }

        public override List<string> GetGlobalVariables()
        {
            if (this.ProcessesCounter == null)
            {
                return Processes[0].GetGlobalVariables();
            }

            List<string> Variables = new List<string>();

            foreach (PARProcess var in Processes)
            {
                Common.Classes.Ultility.Ultility.AddList(Variables, var.GetGlobalVariables());
            }

            return Variables;
        }

        public override List<string> GetChannels()
        {
            if (this.ProcessesCounter == null)
            {
                return Processes[0].GetChannels();
            }

            List<string> channels = new List<string>();

            foreach (PARProcess var in Processes)
            {
                Common.Classes.Ultility.Ultility.AddList(channels, var.GetChannels());
            }

            return channels;
        }

        public override string ToString()
        {
            if (this.ProcessesCounter == null)
            {
                return "(||| {" + RangeExpression.ToString() + "} @" + Processes[0] + ")";
            }

            string toReturn = "";

            Debug.Assert(ProcessesCounter != null);

            foreach (PARProcess processe in Processes)
            {
                if (ProcessesCounter[processe.ProcessID] == -1)
                {
                    toReturn += "(||| {..} @" + processe + ")|||";
                }
                else
                {
                    toReturn += "(||| {" + ProcessesCounter[processe.ProcessID] + "} @" + processe + ")|||";
                }
            }

            return toReturn.TrimEnd('|');
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            int size = Processes.Count;
            List<PARProcess> newProceses = new List<PARProcess>(size);
            Dictionary<string, int> processCounters = new Dictionary<string, int>(size);
            IndexInterleaveAbstract interleaveAbstract = null;

            if (Specification.IsParsing)
            {
                if (ProcessesCounter == null)
                {
                    return new IndexInterleaveAbstract(Processes[0].ClearConstant(constMapping), RangeExpression.ClearConstant(constMapping));
                }

                for (int i = 0; i < Processes.Count; i++)
                {
                    int outer = ProcessesCounter[Processes[i].ProcessID];
                    PARProcess newProc = Processes[i].ClearConstant(constMapping);

                    string temp = newProc.ProcessID;

                    if (!processCounters.ContainsKey(temp))
                    {
                        newProceses.Add(newProc);
                        processCounters.Add(temp, outer);
                    }
                    else
                    {
                        if (processCounters[temp] != -1)
                        {
                            processCounters[temp] = processCounters[temp] + outer;
                        }
                    }
                }

                interleaveAbstract = new IndexInterleaveAbstract(newProceses, processCounters);
                interleaveAbstract.ProcessesActualSize = new Dictionary<string, int>(processCounters);
                return interleaveAbstract;
            }

            //clear constant must be side effect free
            Dictionary<string, int> newProcessesCounter = ProcessesCounter;
            // Dictionary<string, int> newProcessesActualSize = ProcessesActualSize;

            if (ProcessesCounter == null)
            {
                Expression val = RangeExpression.ClearConstant(constMapping);
                size = (EvaluatorDenotational.Evaluate(val, null) as IntConstant).Value;

                if (size == 0)
                {
                    return new Skip();
                }
                else if (size < 0)
                {
                    throw new ParsingException("Negative range " + size + " for parameterized interleave: " + RangeExpression, 0, 0, RangeExpression.ToString());
                }

                newProcessesCounter = new Dictionary<string, int>();
                newProcessesCounter.Add(Processes[0].ProcessID, size);
            }

            for (int i = 0; i < Processes.Count; i++)
            {
                int outer = newProcessesCounter[Processes[i].ProcessID];
                PARProcess newProc = Processes[i].ClearConstant(constMapping);
                
                if (newProc is IndexInterleave)
                {
                    List<PARProcess> processes = (newProc as IndexInterleave).Processes;
                    
                    for (int j = 0; j < processes.Count; j++)
                    {
                        string temp = processes[j].ProcessID;
                        if (!processCounters.ContainsKey(temp))
                        {
                            processCounters.Add(temp, outer);
                            newProceses.Add(processes[j]);                          
                        }
                        else
                        {
                            if (processCounters[temp] != -1)
                            {
                                processCounters[temp] = processCounters[temp] + outer;
                            } 
                        }
                    }
                }
                else if (newProc is IndexInterleaveAbstract)
                {
                    IndexInterleaveAbstract intAbs = newProc as IndexInterleaveAbstract;

                    for (int j = 0; j < intAbs.Processes.Count; j++)
                    {
                        string temp = intAbs.Processes[j].ProcessID;
                        int inner = intAbs.ProcessesCounter[temp];

                        if (!processCounters.ContainsKey(temp))
                        {
                            newProceses.Add(intAbs.Processes[j]);

                            if (outer != -1 && inner != -1)
                            {
                                processCounters.Add(temp, inner * outer);                                
                            }
                            else
                            {
                                processCounters.Add(temp, -1);
                            }
                        }
                        else
                        {
                            if (processCounters[temp] != -1)
                            {
                                if (inner == -1 || outer == -1)
                                {
                                    processCounters.Add(temp,-1);                                        
                                }
                                else
                                {
                                    processCounters[temp] = processCounters[temp] + outer*inner;                                                                         
                                }
                            }                            
                        }
                    }
                }
                else
                {
                    string temp = newProc.ProcessID;

                    if (!processCounters.ContainsKey(temp))
                    {
                        newProceses.Add(newProc);
                        processCounters.Add(temp, outer);                        
                    }
                    else
                    {
                        if (processCounters[temp] != -1)
                        {
                            processCounters[temp] = processCounters[temp] + outer;
                        }
                    }
                }
            }

            interleaveAbstract =  new IndexInterleaveAbstract(newProceses, processCounters);
            interleaveAbstract.ProcessesActualSize = new Dictionary<string, int>(processCounters);
            return interleaveAbstract;
        }

        public override bool MustBeAbstracted()
        {
            if (this.ProcessesCounter == null)
            {
                return Processes[0].MustBeAbstracted();
            }

            for (int i = 0; i < Processes.Count; i++)
            {
                string tmp = Processes[i].ProcessID;

                if (ProcessesActualSize[tmp] == -1)
                {
                    return true;
                }  

                bool toReturn = Processes[i].MustBeAbstracted();
                if(toReturn)
                {
                    return true;
                }              
            }

            return false;
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return this;
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> activeTimers = new List<byte>();

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        Common.Classes.Ultility.Ultility.Union(activeTimers, Processes[i].GetActiveTimer());//eStep));
        //    }
        //    return activeTimers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            Dictionary<string, int> newProcessesCounter = new Dictionary<string, int>();
            List<PARProcess> newListProcess = new List<PARProcess>();

            bool anyNeedtimer = false;
            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess newProc = Processes[i].SetTimer(timerID, clocks);
                int processCount = ProcessesCounter[Processes[i].ProcessID];

                if (newProc == null)
                {
                    newListProcess.Add(Processes[i]);
                    newProcessesCounter.Add(Processes[i].ProcessID, processCount);
                }
                else
                {
                    anyNeedtimer = true;

                    if(newProcessesCounter.ContainsKey(newProc.ProcessID))
                    {
                        newProcessesCounter[newProc.ProcessID] = newProcessesCounter[newProc.ProcessID] + processCount;
                    }
                    else
                    {
                        newListProcess.Add(newProc);
                        newProcessesCounter.Add(newProc.ProcessID, processCount);
                    }                    
                }
            }

            //all processes are stop processes
            if (anyNeedtimer)
            {
                return new IndexInterleaveAbstract(newListProcess, newProcessesCounter);
            }

            return null;
        }

        //public override bool IsTime()
        //{
        //    if (this.ProcessesCounter == null)
        //    {
        //        return Processes[0].IsTime();
        //    }

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        string tmp = Processes[i].ProcessID;

        //        if (ProcessesActualSize[tmp] == -1)
        //        {
        //            return true;
        //        }

        //        bool toReturn = Processes[i].IsTime();
        //        if (toReturn)
        //        {
        //            return true;
        //        }
        //    }

        //    return false;
        //}
    }
}