using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class IndexInternalChoice : PARProcess
    {
        public List<PARProcess> Processes;
        public IndexedProcess IndexedProcessDefinition;

        public IndexInternalChoice(IndexedProcess definition)
        {
            IndexedProcessDefinition = definition;
        }

        public IndexInternalChoice(List<PARProcess> processes)
        {
            Processes = new List<PARProcess>();
            StringBuilder ID = new StringBuilder();

            foreach (PARProcess proc in processes)
            {
                Processes.Add(proc);
                ID.Append(Constants.INTERNAL_CHOICE);
                ID.Append(proc.ProcessID);
            }

            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID.ToString());
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            eStep.AddInequality(eStep.NewClock, TimerOperationType.Equals, 0);

            List<PARConfiguration> list = new List<PARConfiguration>();
            for (int i = 0; i < Processes.Count; i++)
            {
                list.Add(new PARConfiguration(Processes[i], Constants.TAU, "[int_choice]", eStep.GlobalEnv, false, true, eStep.TheConstraint));
            }

            return list;
        }



        public override string ToString()
        {
            if (Processes == null)
            {
                return "<>" + IndexedProcessDefinition;
            }

            string result = "(" + Processes[0];
            for (int i = 1; i < Processes.Count; i++)
            {
                result += "<>";
                result += Processes[i].ToString();
            }
            result += ")";
            return result;
        }

        public override List<string> GetGlobalVariables()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetGlobalVariables();
            }

            List<string> Variables = new List<string>();
            for (int i = 0; i < this.Processes.Count; i++)
            {
                Common.Classes.Ultility.Ultility.AddList<string>(Variables, Processes[i].GetGlobalVariables());
            }
            return Variables;
        }

        public override List<string> GetChannels()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetChannels();
            }

            List<string> channels = new List<string>();
            for (int i = 0; i < this.Processes.Count; i++)
            {
                Common.Classes.Ultility.Ultility.AddList<string>(channels, Processes[i].GetChannels());
            }
            return channels;
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetAlphabets(visitedDefinitionRefs);
            }

            HashSet<string> list = new HashSet<string>();
            for (int i = 0; i < Processes.Count; i++)
            {
                list.UnionWith(Processes[i].GetAlphabets(visitedDefinitionRefs));
            }
            return list;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            List<PARProcess> newnewListProcess = Processes;
            if (Processes == null)
            {
                if (Specification.IsParsing)
                {
                    return new IndexInternalChoice(IndexedProcessDefinition.ClearConstant(constMapping));
                }

                newnewListProcess = IndexedProcessDefinition.GetIndexedProcesses(constMapping);
            }


            List<PARProcess> newListProcess = new List<PARProcess>();
            for (int i = 0; i < newnewListProcess.Count; i++)
            {
                PARProcess newProc = newnewListProcess[i].ClearConstant(constMapping);
                newListProcess.Add(newProc);
            }
            return new IndexInternalChoice(newListProcess);
        }

        public override bool MustBeAbstracted()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.MustBeAbstracted();
            }

            for (int i = 0; i < Processes.Count; i++)
            {
                if (Processes[i].MustBeAbstracted())
                {
                    return true;
                }
            }
            return false;
        }

        //public override bool IsTime()
        //{
        //    if (Processes == null)
        //    {
        //        return IndexedProcessDefinition.Process.IsTime();
        //    }

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        if (Processes[i].IsTime())
        //        {
        //            return true;
        //        }
        //    }
        //    return false;
        //}
    }
}