using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class IndexParallel : PARProcess
    {
        public List<PARProcess> Processes;
        private HashSet<string>[] Alphabets;

        public IndexedProcess IndexedProcessDefinition;

        public IndexParallel(IndexedProcess indexedProcessDefinition)
        {
            IndexedProcessDefinition = indexedProcessDefinition;
        }

        public IndexParallel(List<PARProcess> processes) //, bool calculateAlphabet
        {
            Processes = processes;
            ProcessID = Constants.PARALLEL + Processes[0].ProcessID;
            for (int i = 1; i < Processes.Count; i++)
            {
                ProcessID += Constants.SEPARATOR + Processes[i].ProcessID;
            }
            ProcessID = PARDataStore.DataManager.InitializeProcessID(ProcessID);
        }

        private IndexParallel(List<PARProcess> processes, HashSet<string>[] alphabets)
        {
            Processes = processes;
            Alphabets = alphabets;
            ProcessID = Constants.PARALLEL + Processes[0].ProcessID; 
            for (int i = 1; i < Processes.Count; i++)
            {
                ProcessID += Constants.SEPARATOR + Processes[i].ProcessID;
            }
            ProcessID = PARDataStore.DataManager.InitializeProcessID(ProcessID);
        }

        private void IdentifySharedEventsAndVariables()
        {
            try
            {
                //the following code is used to handle the special case where (e -> e{x=1;} -> P) || e -> Q;
                //the rule is that a shared event can not have the same name as a local action.
                //CheckNameConflictBetweenSynAndLocalAction(new Dictionary<string, bool>(), new List<string>());
                Dictionary<string, int> EventSharedness = new Dictionary<string, int>();

                Alphabets = new HashSet<string>[Processes.Count];
                for (int i = 0; i < Processes.Count; i++)
                {
                    //the alphabet returned shall not have duplicates!!!!
                    HashSet<string> alphabet = Processes[i].GetAlphabets(new Dictionary<string, string>());
                    alphabet.Add(Constants.TERMINATION);
                    Alphabets[i] = alphabet;

                    foreach (string s in alphabet)
                    {
                        if (!EventSharedness.ContainsKey(s))
                        {
                            EventSharedness.Add(s, 1);
                        }
                        else
                        {
                            EventSharedness[s] = EventSharedness[s] + 1;
                        }
                    }
                }

                //remove events which is not shared.
                for (int i = 0; i < Processes.Count; i++)
                {
                    HashSet<string> alphabet = Alphabets[i];
                    foreach (KeyValuePair<string, int> pair in EventSharedness)
                    {
                        if (pair.Value == 1 && alphabet.Contains(pair.Key))
                        {
                            alphabet.Remove(pair.Key);
                        }
                    }
                    //int counter = 0;

                    //while (counter < alphabet.Count)
                    //{
                    //    string s = alphabet[counter];
                    //    if (EventSharedness[s] == 1)
                    //    {
                    //        alphabet.Remove(s);
                    //    }
                    //    else
                    //    {
                    //        counter++;
                    //    }
                    //} 
                }
            }
            catch (System.Exception)
            {
                Alphabets = null;
                throw;
            }
        }


        
        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            if (Alphabets == null)
            {
                IdentifySharedEventsAndVariables();
            }

            List<PARConfiguration> returnList = new List<PARConfiguration>();
            List<string> barrierEnabledEvents = new List<string>();
            List<string> disabled = new List<string>();
            Dictionary<string, List<PARConfiguration>> syncSteps = new Dictionary<string, List<PARConfiguration>>();

            for (int i = 0; i < Processes.Count; i++)
            {
                List<PARConfiguration> list1 = Processes[i].MoveOneStep(eStep);
                List<string> enabled = new List<string>(list1.Count);

                for (int j = 0; j < list1.Count; j++)
                {
                    PARConfiguration step = list1[j];
                    string evt = step.Event;
                    

                    if (!Alphabets[i].Contains(evt) || step.IsDataOperation)
                    {
                        //otherwise, the participating process is process-i only.
                        if (AssertionBase.CalculateParticipatingProcess)
                        {
                            step.ParticipatingProcesses = new string[] { i.ToString() };
                        }

                        List<PARProcess> newProcess = new List<PARProcess>(Processes);
                        newProcess[i] = step.Process;
                        step.Process = new IndexParallel(newProcess, Alphabets);
                        returnList.Add(step);
                    }
                    else
                    {
                        enabled.Add(evt);
                        string key = evt + Constants.SEPARATOR + i;
                        if (!syncSteps.ContainsKey(key))
                        {
                            syncSteps.Add(key, new List<PARConfiguration>());
                        }
                        syncSteps[key].Add(step);

                        if (!barrierEnabledEvents.Contains(evt)) //Alphabets[i].Contains(s) && 
                        {
                            barrierEnabledEvents.Add(evt);
                        }
                    }
                }

                foreach (string s in Alphabets[i])
                {
                    if (!enabled.Contains(s) && !disabled.Contains(s))
                    {
                        disabled.Add(s);
                    }
                }

                //to check whether there are synchoronous channel input/output
                if (SpecificationBase.HasSyncrhonousChannel)
                {
                    SynchronousChannelInputOutput(returnList, i, eStep, null, Processes);
                }
            }

            int disabledCount = disabled.Count;
            for (int i = 0; i < disabledCount; i++)
            {
                barrierEnabledEvents.Remove(disabled[i]);
            }

            List<bool> isAtomic = null;
            List<bool> isUrgent;

            //move the barrier synchronization events.
            foreach (string evt in barrierEnabledEvents)
            {
                //maps an event to the list of resulting processes.
                List<List<PARProcess>> moves = new List<List<PARProcess>>();
                moves.Add(new List<PARProcess>());

                if (SpecificationBase.HasAtomicEvent)
                {
                    isAtomic = new List<bool>();
                    isAtomic.Add(false);
                }

                isUrgent = new List<bool>();
                isUrgent.Add(false);

                List<List<Constraint>> syncedConstraints = new List<List<Constraint>>();
                syncedConstraints.Add(new List<Constraint>());

                List<string> participatingProcesses = new List<string>();

                for (int i = 0; i < Processes.Count; i++)
                {
                    if (Alphabets[i].Contains(evt))
                    {
                        participatingProcesses.Add(i.ToString());
                        List<PARConfiguration> steps = syncSteps[evt + Constants.SEPARATOR + i];
                        List<List<PARProcess>> toAdd = new List<List<PARProcess>>(moves.Count);
                        List<List<Constraint>> toAddConstraint = new List<List<Constraint>>(moves.Count);

                        foreach (PARConfiguration step in steps)
                        {
                            //if it happens that a data operation shares the same name with the sync event, ignore
                            Debug.Assert(!step.IsDataOperation);
                            if (moves[0].Count == i)
                            {
                                for (int j = 0; j < moves.Count; j++)
                                {
                                    moves[j].Add(step.Process);
                                    syncedConstraints[j].Add(step.TheConstraint);

                                    if (step.IsAtomic)
                                    {
                                        for (int k = 0; k < isAtomic.Count; k++)
                                        {
                                            isAtomic[k] = true;
                                        }
                                    }

                                    if (step.IsUrgent)
                                    {
                                        for (int k = 0; k < isUrgent.Count; k++)
                                        {
                                            isUrgent[k] = true;
                                        }
                                    }

                                }
                            }
                            else
                            {
                                //If there non-determinism, clone and then add.
                                for (int j = 0; j < moves.Count; j++)
                                {
                                    List<PARProcess> newProcList = new List<PARProcess>();
                                    List<Constraint> newConstraintList = new List<Constraint>();
                                    List<PARProcess> list = moves[j];
                                    List<Constraint> Constraintlist = syncedConstraints[j];

                                    for (int k = 0; k < list.Count - 1; k++)
                                    {
                                        newProcList.Add(list[k]);
                                        newConstraintList.Add(Constraintlist[k]);
                                    }

                                    newProcList.Add(step.Process);
                                    toAdd.Add(newProcList);

                                    newConstraintList.Add(step.TheConstraint);
                                    toAddConstraint.Add(newConstraintList);

                                    if (SpecificationBase.HasAtomicEvent)
                                    {
                                        if (!isAtomic[j])
                                        {
                                            isAtomic.Add(step.IsAtomic);
                                        }
                                        else
                                        {
                                            isAtomic.Add(true);
                                        }
                                    }

                                    if (!isUrgent[j])
                                    {
                                        isUrgent.Add(step.IsAtomic);
                                    }
                                    else
                                    {
                                        isUrgent.Add(true);
                                    }
                                }
                            }
                        }

                        moves.AddRange(toAdd);
                        syncedConstraints.AddRange(toAddConstraint);
                    }
                    else
                    {
                        for (int j = 0; j < moves.Count; j++)
                        {
                            moves[j].Add(Processes[i]);
                        }
                    }
                }

                for (int i = 0; i < moves.Count; i++)
                {
                    List<PARProcess> list = moves[i];
                    IndexParallel para = new IndexParallel(list, Alphabets);

                    Constraint newconstraint = syncedConstraints[i][0];
                    for (int k = 0; k < syncedConstraints[i].Count; k++)
                    {
                        newconstraint.Conjunction(syncedConstraints[i][k]);
                    }

                    PARConfiguration tmpStep = new PARConfiguration(para, evt, null, eStep.GlobalEnv, false, isUrgent[i], newconstraint);

                    if (SpecificationBase.HasAtomicEvent)
                    {
                        tmpStep.IsAtomic = isAtomic[i];
                    }

                    if (AssertionBase.CalculateParticipatingProcess)
                    {
                        tmpStep.ParticipatingProcesses = participatingProcesses.ToArray();
                    }

                    returnList.Add(tmpStep);
                }
            }

            return returnList;
        }



        private void SynchronousChannelInputOutput(List<PARConfiguration> returnList, int i, PARConfiguration step, string evt, List<PARProcess> clockSetProcesses)
        {
            List<PARConfigurationWithChannelData> outputs = clockSetProcesses[i].SyncOutput(step);

            foreach (PARConfigurationWithChannelData vm in outputs)
            {
                if (evt != null && vm.Event != evt)
                {
                    continue;
                }

                PARProcess output = vm.Process;

                for (int k = 0; k < clockSetProcesses.Count; k++)
                {
                    if (k != i)
                    {
                        List<PARConfiguration> syncedProcess = clockSetProcesses[k].SyncInput(vm);

                        foreach (PARConfiguration p in syncedProcess)
                        {
                            List<PARProcess> newProcess = new List<PARProcess>(clockSetProcesses);
                            newProcess[i] = output;
                            newProcess[k] = p.Process;

                            IndexParallel interleave = new IndexParallel(newProcess, Alphabets);
                            Constraint newConstraint = vm.TheConstraint.Clone();
                            newConstraint.Conjunction(p.TheConstraint);

                            PARConfiguration newStep = new PARConfiguration(interleave, vm.Event, vm.DisplayName, vm.GlobalEnv, false, vm.IsUrgent || p.IsUrgent, newConstraint);
                            newStep.IsAtomic = vm.IsAtomic || p.IsAtomic;

                            if (AssertionBase.CalculateParticipatingProcess)
                            {
                                newStep.ParticipatingProcesses = new string[] { i.ToString(), k.ToString() };
                            }

                            returnList.Add(newStep);
                        }
                    }
                }
            }
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> returnList = new List<PARConfigurationWithChannelData>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                List<PARConfigurationWithChannelData> list1 = process.SyncOutput(eStep);

                for (int j = 0; j < list1.Count; j++)
                {
                    PARConfiguration step = list1[j];

                    List<PARProcess> newProcess = new List<PARProcess>(Processes.Count);
                    newProcess.AddRange(Processes);
                    newProcess[i] = step.Process;

                    step.Process = new IndexParallel(newProcess, Alphabets);
                    returnList.Add(list1[j]);
                }
            }

            return returnList;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                List<PARConfiguration> list1 = process.SyncInput(eStep);

                for (int j = 0; j < list1.Count; j++)
                {
                    PARConfiguration step = list1[j];
                    List<PARProcess> newProcess = new List<PARProcess>(Processes.Count);
                    newProcess.AddRange(Processes);
                    newProcess[i] = step.Process;

                    step.Process = new IndexParallel(newProcess, Alphabets);
                    returnList.Add(step);
                }
            }

            return returnList;
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetAlphabets(visitedDefinitionRefs);
            }

            HashSet<string> toReturn = new HashSet<string>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                toReturn.UnionWith(process.GetAlphabets(visitedDefinitionRefs));
            }
            return toReturn;
        }

        public override List<string> GetGlobalVariables()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetGlobalVariables();
            }


            List<string> vars = new List<string>();

            foreach (PARProcess var in Processes)
            {
                Common.Classes.Ultility.Ultility.AddList(vars, var.GetGlobalVariables());
            }

            return vars;
        }

        public override List<string> GetChannels()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetChannels();
            }


            List<string> channels = new List<string>();

            foreach (PARProcess var in Processes)
            {
                Common.Classes.Ultility.Ultility.AddList(channels, var.GetChannels());
            }

            return channels;
        }

        public override string ToString()
        {
            if (Processes == null)
            {
                return "||" + IndexedProcessDefinition;
            }

            StringBuilder s = new StringBuilder();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                s.Append(process.ToString() + "\r\n||");
            }
            return "(" + s.ToString().TrimEnd('|') + ")";
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            List<PARProcess> newnewProcesses = Processes;

            List<PARProcess> newProceses = null;
            int size;
            if (Specification.IsParsing)
            {
                if (Processes == null)
                {
                    return new IndexParallel(IndexedProcessDefinition.ClearConstant(constMapping));
                }
                size = Processes.Count;
                newProceses = new List<PARProcess>(size);
                for (int i = 0; i < size; i++)
                {
                    newProceses.Add(Processes[i].ClearConstant(constMapping));
                }
            }

            if (Processes == null)
            {
                //return new IndexParallel(IndexedProcessDefinition.GetIndexedProcesses(constMapping), !Specification.IsParsing);
                newnewProcesses = IndexedProcessDefinition.GetIndexedProcesses(constMapping);
            }


            size = newnewProcesses.Count;
            newProceses = new List<PARProcess>(size); 

            for (int i = 0; i < size; i++)
            {
                PARProcess newProc = newnewProcesses[i].ClearConstant(constMapping);

                if (newProc is IndexParallel && (newProc as IndexParallel).Processes != null)
                {
                    List<PARProcess> processes = (newProc as IndexParallel).Processes;
                    newProceses.AddRange(processes);
                }

                else
                {
                    newProceses.Add(newProc);
                }
            }

            return new IndexParallel(newProceses); //, !Specification.IsParsing
        }

        public override bool MustBeAbstracted()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.MustBeAbstracted();
            }

            for (int i = 0; i < Processes.Count; i++)
            {
                if (Processes[i].MustBeAbstracted())
                {
                    return true;
                }
            }
            return false;
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return this;
        }

        //public override List<byte> GetActiveTimer()
        //{
        //    List<byte> activeTimers = new List<byte>();

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        Common.Classes.Ultility.Ultility.Union(activeTimers, Processes[i].GetActiveTimer());//eStep));
        //    }
        //    return activeTimers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            List<PARProcess> newListProcess = new List<PARProcess>();
            bool anyNeedtimer = false;

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess newProc = Processes[i].SetTimer(timerID, clocks);
                if (newProc == null)
                {
                    newListProcess.Add(Processes[i]);
                }
                else
                {
                    anyNeedtimer = true;
                    newListProcess.Add(newProc);
                }
            }

            //all processes are stop processes
            if (anyNeedtimer)
            {
                IndexParallel index = new IndexParallel(newListProcess, Alphabets);
                return index;
            }

            return null;
        }


        //public override bool IsTime()
        //{
        //    if (Processes == null)
        //    {
        //        return IndexedProcessDefinition.Process.IsTime();
        //    }

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        if (Processes[i].IsTime())
        //        {
        //            return true;
        //        }
        //    }
        //    return false;
        //}
    }
}