﻿//#define DEBUGCONSOLE

using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;

namespace PAT.PAR.LTS
{
    public class PARConfiguration : ConfigurationBase
    {
        // The current process
        public PARProcess Process;
        // ETIENNE SAYS: PROBABLY USELESS
        public bool IsUrgent;
        // The full id for the configuration
        protected string ConfigID;
        // The discrete id for the discrete part (variables + process) of the configuration
        protected string DiscreteID;
        // The current constraint on clocks and parameters
        public Constraint TheConstraint;
        // The set of active clocks in Process
        private List<byte> Clocks;
        // The clock introduced during this transition; notice that a fresh clock is introduced every transition;
        public byte NewClock;


        public PARConfiguration(PARProcess p, string e, string hiddenEvent, Valuation globalEnv, bool isDataOperation, bool isUrgent, Constraint constraint)
        {
            Process = p;
            Event = e;
            GlobalEnv = globalEnv;
            DisplayName = hiddenEvent;
            IsDataOperation = isDataOperation;
            TheConstraint = constraint;
            Clocks = new List<byte>();


            //TESTS
            //if (Constraint.NbParameters > 0)
            //{
            //    Constraint.RunTests();
            //}

        }

        public override ConfigurationBase[] MakeOneMove()
        {
            return MakeOneMoveLocal().ToArray();
        }

        /**
         * Return the number of clocks currently used
         */
        public int getNbClocks()
        {
            return Clocks.Count;
        }

        public bool IsInstantaneous()
        {
            // changed by Etienne
            //return DBM.GetTimerUpper(NewClock) == 0;
            return false;
        }



        public virtual void SetTimer()
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("--------------------------------------------------");
            System.Console.Out.WriteLine("Entering PARSimulationConfiguration::SetTimer()");
            System.Console.Out.WriteLine("  This = ");
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            if (this != null)
            {
                System.Console.Out.WriteLine(this.ToString());
            }
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Process: " + Process);
#endif
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif

            NewClock = TheConstraint.GetNewClockID();

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("NewClock = " + Constraint.ClockToString(NewClock));
#endif

            // Starting from a fresh list of clocks
            Clocks = new List<byte>();
            List<Constraint> Constraints = new List<Constraint>();

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("About to call Process.SetTimer(NewClock, new ClockInfo(GlobalEnv, Clocks, Constraints)) with NewClock = "
                + Constraint.ClockToString(NewClock));
#endif
            PARProcess newProc = Process.SetTimer(NewClock, new ClockInfo(GlobalEnv, Clocks, Constraints));

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Returned newProc: " + newProc);
#endif

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("After SetTimer(), " + Constraints.Count + " Constraints = ");
            foreach (Constraint c in Constraints)
            {
                System.Console.Out.WriteLine("  ** " + c);
            }

            System.Console.Out.WriteLine("TheConstraint = " + TheConstraint.ToString());
            System.Console.Out.WriteLine("About to keep clocks " + Constraint.SomeClocksToString(Clocks));
#endif

            TheConstraint = TheConstraint.KeepClocks(Clocks);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("TheConstraint = " + TheConstraint.ToString());
            System.Console.Out.WriteLine("About to ensure that clocks " + Constraint.SomeClocksToString(Clocks) + " are nonnegative");
#endif
            // Necessary to add x >= 0 for all clocks in 'Clocks'!
            TheConstraint.EnsureClocksNonnegative(Clocks);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("TheConstraint = " + TheConstraint.ToString());
#endif

            if (newProc != null)
            {
                Process = newProc;
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("newProc != null, hence now considering newProc.");
#endif
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Process: " + Process);
#endif
            // NEW TEST BY ETIENNE
            if (Clocks.Contains(NewClock))
            {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("Adding '" + Constraint.ClockToString(NewClock) + " = 0'.");
#endif
                TheConstraint.AddInequality(NewClock, TimerOperationType.Equals, 0);
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("TheConstraint: " + TheConstraint.ToString());
#endif
            }
            else
            {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("NewClock " + Constraint.ClockToString(NewClock) + " has not been used: forget it.");
#endif
            }

            // Delay the constraint
            TheConstraint.Delay();

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Delayed TheConstraint : " + TheConstraint.ToString());
#endif

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Before interesction, " + Constraints.Count + " Constraints = ");
            foreach (Constraint c in Constraints)
            {
                System.Console.Out.WriteLine("  ** " + c);
            }
#endif
            // Add inequalities
            TheConstraint.AddConstraints(Constraints);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Added " + Constraints.Count + " inequalities to TheConstraint : " + TheConstraint.ToString());
#endif

            //foreach (KeyValuePair<byte, int> keyValuePair in clocksBound)
            //{
            //    DBM.AddConstraint(keyValuePair.Key, TimerOperationType.LessThanOrEqualTo, keyValuePair.Value);
            //}


            //byte clockID = TheConstraint.GetNewClockID();
            //PARProcess newProc = Process.SetTimer(clockID, this);

            //if (newProc != null)
            //{
            //    Process = newProc;
            //}

            //// ETIENNE SAYS: this might be useless (?)
            //TheConstraint = TheConstraint.Clone();
            //TheConstraint.AddClock(clockID);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("About to leave PARSimulationConfiguration::SetTimer()");
            System.Console.Out.WriteLine("  This = ");
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            if (this != null)
            {
                System.Console.Out.WriteLine(this.ToString());
            }
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            System.Console.Out.WriteLine("Now really exiting PARSimulationConfiguration::SetTimer()");
            System.Console.Out.WriteLine("--------------------------------------------------");
#endif
        }





        public void AddInequality(byte clockID, TimerOperationType op, int constant)
        {
            TheConstraint.AddInequality(clockID, op, constant);
        }



        public void AddParInequality(byte clockID, TimerOperationType op, string parameter)
        {
            TheConstraint.AddParInequality(clockID, op, parameter);
        }


        // Add a constraint on the parameters to the configuration
        public void AddConstraint(Constraint c)
        {
            List<Constraint> l = new List<Constraint>();
            l.Add(c);
            TheConstraint.AddConstraints(l);
        }



        public virtual List<PARConfiguration> MakeOneMoveLocal()
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("****************************************************************************************************");
            System.Console.Out.WriteLine("Entering PARSimulationConfiguration::MakeOneMoveLocal()");
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            if (this != null)
            {
                System.Console.Out.WriteLine(this.ToString());
            }
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif

            //TheConstraint.Delay();

            List<PARConfiguration> list = Process.MoveOneStep(this);

            List<PARConfiguration> returnListAtomic = new List<PARConfiguration>();
            List<PARConfiguration> returnListNonAtomic = new List<PARConfiguration>();

            bool hasUrgent = false;

            foreach (PARConfiguration configuration in list)
            {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("--Considering a configuration");
                System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                System.Console.Out.WriteLine(configuration.ToString());
                System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif

                // Separation between atomic and non-atomic configurations
                if (configuration.IsAtomic)
                {
#if DEBUGCONSOLE
                    System.Console.Out.WriteLine("  This is an atomic configuration");
#endif
                    configuration.TheConstraint = configuration.TheConstraint.Clone();

//#if DEBUGCONSOLE
//                    System.Console.Out.WriteLine("  About to add inequality '"
//                        + Constraint.ClockToString(NewClock) + "=0'");
//#endif
//                    // ETIENNE SAYS: WHY?!
//                    configuration.TheConstraint.AddInequality(NewClock, TimerOperationType.Equals, 0);

                    if (configuration.TheConstraint.IsSatisfiable())
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("  Constraint is satisfiable");
                        System.Console.Out.WriteLine("  About to call SetTimer()");
#endif
                        configuration.SetTimer();
                        returnListAtomic.Add(configuration);
                    }
                    else
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("  Constraint is not satisfiable; skip");
#endif
                    }

                }
                else
                {
#if DEBUGCONSOLE
                    System.Console.Out.WriteLine("  This is an non-atomic configuration");
#endif
                    if (configuration.TheConstraint.IsSatisfiable())
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("  Constraint is satisfiable");
#endif
                        if (configuration.IsUrgent)
                        {
#if DEBUGCONSOLE
                            System.Console.Out.WriteLine("  Configuration is urgent");
#endif
                            hasUrgent = true;
                        }

                        returnListNonAtomic.Add(configuration);
                    }
                    else
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("  Constraint is not satisfiable; skip");
#endif
                    }
                }
            }

            // if atomic, then only return atomic
            if (returnListAtomic.Count > 0)
            {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("Found atomic configurations: return atomic");
#endif
                return returnListAtomic;
            }

            else
            {
                if (hasUrgent)
                {
#if DEBUGCONSOLE
                    System.Console.Out.WriteLine("  Found urgent configurations");
#endif
                    List<PARConfiguration> toReturn = new List<PARConfiguration>();

                    foreach (PARConfiguration configuration in returnListNonAtomic)
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("--Considering a urgent configuration");
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                        System.Console.Out.WriteLine(configuration.ToString());
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif
                        // NECESSARY ?
                        //PARConfiguration.TheConstraint = configuration.TheConstraint.Clone();
                        // IMPOSSIBLE!
                        //configuration.TheConstraint.AddInequality(NewClock, TimerOperationType.LessThanOrEqualTo, timelimit);

                        // TEST BECAME UNNECESSARY
                        //if (configuration.TheConstraint.IsSatisfiable())
                        //{
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("  About to set timer");
#endif
                        configuration.SetTimer();
                        toReturn.Add(configuration);
                        //}
                    }

                    IsDeadLock = toReturn.Count == 0;
                    return toReturn;
                }
                else
                {
#if DEBUGCONSOLE
                    System.Console.Out.WriteLine("No urgent configuration found.");
#endif
                    foreach (PARConfiguration configuration in returnListNonAtomic)
                    {
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("--Considering a non-urgent configuration");
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                        System.Console.Out.WriteLine(configuration.ToString());
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                        System.Console.Out.WriteLine("  About to call SetTimer()");
#endif
                        configuration.SetTimer();
#if DEBUGCONSOLE
                        System.Console.Out.WriteLine("  After SetTimer, configuration becomes: ");
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                        System.Console.Out.WriteLine(configuration.ToString());
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif
                    }

                    IsDeadLock = returnListNonAtomic.Count == 0;

#if DEBUGCONSOLE
                    System.Console.Out.WriteLine("Leaving PARSimulationConfiguration::MakeOneMoveLocal()");
                    System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                    if (this != null)
                    {
                        System.Console.Out.WriteLine(this.ToString());
                    }
                    System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                    System.Console.Out.WriteLine("****************************************************************************************************");
#endif
                    return returnListNonAtomic;
                }
            }
        }
        //{
        //    TheConstraint.Delay();
        //    Constraint idleConstraint = TheConstraint.Clone();
        //    List<PARConfiguration> list = Process.MoveOneStep(this);

        //    // TO DO LATER: atomic
        //    //List<PARConfiguration> returnListAtomic = new List<PARConfiguration>();
        //    //List<PARConfiguration> returnListNonAtomic = new List<PARConfiguration>();
        //    List<PARConfiguration> returnList = new List<PARConfiguration>();

        //    for (int i = 0; i < list.Count; i++)
        //    {
        //        Constraint newConstraint = list[i].TheConstraint.Clone();
        //        newConstraint.Conjunction(idleConstraint);

        //        // TO DO LATER: atomic
        //        //if (list[i].IsAtomic)
        //        //{
        //        //    newConstraint.AddConstraint(clockID, TimerOperationType.Equals, 0);
        //        //    //newDBM.Conjunction(clonedDBM);
        //        //}

        //        if (newConstraint.IsSatisfiable())
        //        {
        //            //    list[i].TimeTakenLowerBound = newDBM.GetTimerLower(timerID);
        //            //    list[i].TimeTakenUpperBound = newDBM.GetTimerUpper(timerID);

        //            List<byte> clocks = list[i].Process.GetActiveTimer();

        //            if (clocks.Count != newConstraint.GetNbClocks())
        //            {
        //                list[i].TheConstraint = newConstraint.KeepClocks(clocks);
        //            }
        //            else
        //            {
        //                list[i].TheConstraint = newConstraint;
        //            }

        //            list[i].SetTimer();

        //            // Addition to the list of configurations
        //            returnList.Add(list[i]);

        //            //    if (list[i].IsAtomic)
        //            //    {
        //            //        returnListAtomic.Add(list[i]);
        //            //    }
        //            //    else
        //            //    {
        //            //        returnListNonAtomic.Add(list[i]);
        //            //    }
        //        }
        //    }

        //    IsDeadLock = returnList.Count == 0;

        //    return returnList;
        //}




        public string GetDiscreteID()
        {
            if (DiscreteID == null)
            {
                if (GlobalEnv.IsEmpty())
                {
                    DiscreteID = Process.ProcessID;
                }
                else
                {
                    DiscreteID = GlobalEnv.GetID(Process.ProcessID);
                }
            }

            return DiscreteID;
        }



        public override string GetIDWithEvent()
        {
            // In PAR module, we MAY modify states; hence we MAY have to compute it every time
            //if (ConfigID == null)
            //{
                if (GlobalEnv.IsEmpty())
                {
                    ConfigID = Process.ProcessID + Constants.SEPARATOR + TheConstraint.GetID(Clocks);
                }
                else
                {
                    ConfigID = GlobalEnv.GetID(Process.ProcessID + Constants.SEPARATOR + TheConstraint.GetID(Clocks));
                }
            //}

            return ConfigID + Constants.SEPARATOR + Event;
        }

        public override string GetID()
        {
            // In PAR module, we MAY modify states; hence we MAY have to compute it every time
            //if (ConfigID == null)
            //{          
                if (GlobalEnv.IsEmpty())
                {
                    ConfigID = Process.ProcessID + Constants.SEPARATOR + TheConstraint.GetID(Clocks);
                }
                else
                {
                    ConfigID = GlobalEnv.GetID(Process.ProcessID + Constants.SEPARATOR + TheConstraint.GetID(Clocks));
                }
            //}

                return ConfigID;
        }



        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
#if DEBUGCONSOLE
            sb.AppendLine("%%%%%%%%%%%%%%%%%%%%");
#endif
            sb.AppendLine("Process:");
            sb.AppendLine(Process.ToString());


            if (!GlobalEnv.IsEmpty())
            {
                sb.AppendLine();
                sb.AppendLine("Environment:");
                sb.AppendLine(GlobalEnv.ToString());

            }

            sb.AppendLine();
            sb.AppendLine("List of clocks: " + Constraint.SomeClocksToString(Clocks));

            if (TheConstraint.HasClocks())
            {
                sb.AppendLine();
                sb.AppendLine("Active clocks in constraint: " + TheConstraint.ClocksToString());
            }

            // Convert each parameter to a string
            List<string> parameters_string = new List<string>(Specification.Parameters.Count);
            foreach (string name in Specification.Parameters)
            {
                parameters_string.Add(name);
            }
            // Convert everything with a separator
            string theParameters = string.Join(", ", parameters_string.ToArray());
            sb.AppendLine("Parameters: " + theParameters);

            sb.AppendLine();
            sb.AppendLine("Constraint:");
            sb.AppendLine(TheConstraint.ToString());

            sb.AppendLine();
            sb.AppendLine("Constraint onto the parameters:");
            sb.AppendLine(TheConstraint.HideAllClocks().ToString()); // Clone very important here!!

            //sb.AppendLine();
            //sb.AppendLine("Constraint id:");
            //sb.AppendLine(TheConstraint.GetID(Clocks));

            sb.AppendLine();
            sb.AppendLine("Process id:");
            sb.AppendLine(GetID());
#if DEBUGCONSOLE
            sb.AppendLine("%%%%%%%%%%%%%%%%%%%%");
#endif

            return sb.ToString();
        }



        public override string GetDisplayEvent()
        {
            string evt = base.GetDisplayEvent();
            //if (evt == Constants.INITIAL_EVENT)
            //{
                return evt;
            //}
        }
    }


    // Added a constraint list
    public sealed class ClockInfo
    {
        public List<byte> Clocks;
        public Valuation GlobalEnv;
        // List of constraints
        public List<Constraint> Constraints;


        // Trivial constructor
        public ClockInfo(Valuation globalEnv, List<byte> clocks, List<Constraint> constraints)
        {
            GlobalEnv = globalEnv;
            Clocks = clocks;
            Constraints = constraints;
        }


        // Add a clock
        public void AddClock(byte ClockID)
        {
            Clocks.Add(ClockID);
        }


        // Add an inequality to the list
        public void AddInequality(byte ClockID, Expression DelayTime)
        {
            // Get the value
            int value = (EvaluatorDenotational.Evaluate(DelayTime, GlobalEnv) as IntConstant).Value;

            Clocks.Add(ClockID);
            Constraints.Add(Constraint.inequality(ClockID, value));
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("ClockInfo > Added clock " + Constraint.ClockToString(ClockID));
            System.Console.Out.WriteLine("ClockInfo > Added inequality " + Constraint.ClockToString(ClockID) + " <= " + value);
            System.Console.Out.WriteLine("ClockInfo > Now Constraints = ");
            foreach (Constraint c in Constraints)
            {
                System.Console.Out.WriteLine("         ** " + c);
            }
#endif
        }


        // Add a parametric inequality to the list
        public void AddParInequality(byte ClockID, string parameter)
        {
            Clocks.Add(ClockID);
            Constraints.Add(Constraint.parInequality(ClockID, parameter));
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("ClockInfo > Added clock " + Constraint.ClockToString(ClockID));
            System.Console.Out.WriteLine("ClockInfo > Added par inequality " + Constraint.ClockToString(ClockID) + " <= " + parameter);
            System.Console.Out.WriteLine("ClockInfo > Now Constraints = ");
            foreach (Constraint c in Constraints)
            {
                System.Console.Out.WriteLine("         ** " + c);
            }
#endif
        }

    }
}
