using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class PARTimeOutProcess : PARProcess
    {
        public byte TimerID;
        public PARProcess FirstProcess;
        public PARProcess SecondProcess;
        public Variable Parameter;

        public PARTimeOutProcess(PARProcess firstProcess, PARProcess secondProcess, Variable parameter, byte timerID)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;
            TimerID = timerID;
            Parameter = parameter;

            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.TIMEOUT + "[" + Parameter + "]" + SecondProcess.ProcessID);
        }


        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            // Get the parameter name
            string parameter_name = Parameter.ToString();
            // Add "clock <= param"
            eStep.AddParInequality(TimerID, TimerOperationType.LessThanOrEqualTo, parameter_name);

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event == Constants.TAU)
                {
                    vm.Process = new PARTimeOutProcess(vm.Process, SecondProcess, Parameter, TimerID);
                }
            }

            Constraint constraint = eStep.TheConstraint.Clone();
            constraint.AddParInequality(TimerID, TimerOperationType.Equals, parameter_name);
            steps.Add(new PARConfiguration(SecondProcess, Constants.TAU, "[timeout[" + parameter_name + "]]", eStep.GlobalEnv, false, false, constraint));
            return steps;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
             return FirstProcess.SyncOutput(eStep);
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            return FirstProcess.SyncInput(eStep);
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, SecondProcess.GetGlobalVariables());
            //Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channel = FirstProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList<string>(channel, SecondProcess.GetChannels());

            return channel;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + FirstProcess + " timeout[" + Parameter + "] " + SecondProcess + ")";
            }
            else
            {
                return "(" + FirstProcess + " timeout[" + Parameter + "]" + Constraint.ClockToString(TimerID) + " " + SecondProcess + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new PARTimeOutProcess(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping), Parameter, TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer() //Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);    
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());//eStep));
        //    Common.Classes.Ultility.Ultility.Union(timers, SecondProcess.GetActiveTimer());//eStep));

        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = FirstProcess.SetTimer(timerID, clocks);

            // Get the parameter name
            string parameter_name = Parameter.ToString();

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddParInequality(TimerID, parameter_name);
                    return new PARTimeOutProcess(process, SecondProcess, Parameter, TimerID);
                }

                clocks.AddParInequality(timerID, parameter_name);
                return new PARTimeOutProcess(process, SecondProcess, Parameter, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddParInequality(TimerID, parameter_name);
                return null;
            }

            clocks.AddParInequality(timerID, parameter_name);
            return new PARTimeOutProcess(FirstProcess, SecondProcess, Parameter, timerID);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}