﻿//#define DEBUGCONSOLE

using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class PARWaitProcess : PARProcess
    {
        //timer id is initially 0, once it is assigned, then it > 0;
        public byte TimerID;
        public Variable Parameter;

        public PARWaitProcess(Variable parameter, byte timerID)
        {
            Parameter = parameter;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(Constants.WAIT + "[" + Parameter + "]");
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> toReturn = new List<PARConfiguration>(1);

            // Get the parameter name
            string parameter_name = Parameter.ToString();
            // Add "clock <= param"            
            eStep.AddParInequality(TimerID, TimerOperationType.LessThanOrEqualTo, parameter_name);

            Constraint newConstraint = eStep.TheConstraint.Clone();
            newConstraint.AddParInequality(TimerID, TimerOperationType.Equals, parameter_name);
            PARConfiguration step = new PARConfiguration(new Skip(), Constants.TAU, "[Wait[" + parameter_name + "]]", eStep.GlobalEnv, false, false, newConstraint);
            toReturn.Add(step);
            return toReturn;
        }



        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "Wait[" + Parameter + "]";
            }
            else
            {
                return "Wait[" + Parameter + "]" + Constraint.ClockToString(TimerID);
            }
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new PARWaitProcess(Parameter, TimerID);
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>(1);
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }

        //    return timers;
        //}

        public override List<string> GetGlobalVariables()
        {
            //return DelayTime.GetVars();
            return new List<string>();
        }

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Entering WaitProcess:SetTimer");
#endif
            // Get the parameter name
            string parameter_name = Parameter.ToString();
            
            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddParInequality(TimerID, parameter_name);
                return null;
            }
            clocks.AddParInequality(timerID, parameter_name);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("About to exit WaitProcess:SetTimer");
#endif

            return new PARWaitProcess(Parameter, timerID);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}