using System.Text;
using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class PARWaituntilProcess : PARProcess
    {
        public PARProcess Process;
        public Variable Parameter;
        public byte TimerID;

        public PARWaituntilProcess(PARProcess process, Variable parameter, byte timerID)
        {
            Process = process;
            Parameter = parameter;
            this.TimerID = timerID;
            
            StringBuilder ID = new StringBuilder(Process.ProcessID);
            ID.Append(Constants.WAITUNTIL + "[" + Parameter + "]");

            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID.ToString());
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = Process.GetGlobalVariables();
            //Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            return Process.GetChannels();
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> toReturn = new List<PARConfiguration>();
            List<PARConfiguration> steps = Process.MoveOneStep(eStep);

            // Get the parameter name
            string parameter_name = Parameter.ToString();

            foreach (PARConfiguration vm in steps)
            {
                Constraint constraint = vm.TheConstraint;

                if (vm.Event == Constants.TERMINATION)
                {
                    Constraint newConstraint = vm.TheConstraint.Clone();
                    newConstraint.AddParInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, parameter_name);
                    vm.TheConstraint = newConstraint;
                    toReturn.Add(vm);

                    newConstraint = vm.TheConstraint.Clone();


                    // DEBUG ETIENNE
                    System.Console.Out.WriteLine("About to add PAR inequality in WaitUntilProcess");

                    newConstraint.AddParInequality(TimerID, TimerOperationType.LessThanOrEqualTo, parameter_name);
                    PARConfiguration config = new PARConfiguration(new PARWaitProcess(Parameter, TimerID), Constants.TAU, "[waituntil[" + Parameter + "]]", vm.GlobalEnv, false, true, newConstraint);
                    config.IsAtomic = vm.IsAtomic;
                    toReturn.Add(config);
                }
                else
                {
                    toReturn.Add(vm);
                }
            }

            return toReturn;
        }

        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> steps = Process.SyncOutput(eStep);

            foreach (PARConfigurationWithChannelData vm in steps)
            {
                vm.Process = new PARWaituntilProcess(vm.Process, Parameter, TimerID);
            }

            return steps;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            //List<RTSConfiguration> toReturn = new List<RTSConfiguration>();
            List<PARConfiguration> steps = Process.SyncInput(eStep);

            //int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;

            foreach (PARConfiguration vm in steps)
            {
                //if (!vm.DBM.Implies(TimerID, TimerOperationType.GreaterThanOrEqualTo, value))
                //{
                vm.Process = new PARWaituntilProcess(vm.Process, Parameter, TimerID);
                //}

                //toReturn.Add(vm);
            }

            return steps;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + Process + " waituntil[" + Parameter + "])";
            }
            else
            {
                return "(" + Process + " waituntil[" + Parameter + "]" + Constraint.ClockToString(TimerID) + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return Process.GetAlphabets(visitedDefinitionRefs);
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new PARWaituntilProcess(Process.ClearConstant(constMapping), Parameter, TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, Process.GetActiveTimer());
        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = Process.SetTimer(timerID, clocks);

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddClock(TimerID);
                    return new PARWaituntilProcess(process, Parameter, TimerID);
                }

                clocks.AddClock(timerID);
                return new PARWaituntilProcess(process, Parameter, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddClock(TimerID);
                return null;
            }

            clocks.AddClock(timerID);
            return new PARWaituntilProcess(Process, Parameter, timerID);
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return Process.GetTopLevelConcurrency(visitedDef);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}
