using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class PARWithinProcess : PARProcess
    {
        public byte TimerID;
        public  PARProcess FirstProcess;
        public  Variable Parameter;

        public PARWithinProcess(PARProcess firstProcess, Variable parameter, byte timerID)
        {
            FirstProcess = firstProcess;
            Parameter = parameter;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.WITHIN + "[" + Parameter + "]");
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            // Get the parameter name
            string parameter_name = Parameter.ToString();
            // Add "clock <= param"
            eStep.AddParInequality(TimerID, TimerOperationType.LessThanOrEqualTo, parameter_name);

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event == Constants.TAU)
                {
                    vm.Process = new PARWithinProcess(vm.Process, Parameter, TimerID);
                }
            }

            return steps;
        }



        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            return FirstProcess.SyncOutput(eStep);
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            return FirstProcess.SyncInput(eStep);
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            //Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            return FirstProcess.GetChannels();
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + FirstProcess.ToString() + " within[" + Parameter + "])";
            }
            else
            {
                return "(" + FirstProcess.ToString() + " within[" + Parameter + "]" + Constraint.ClockToString(TimerID) + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return FirstProcess.GetAlphabets(visitedDefinitionRefs);
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new PARWithinProcess(FirstProcess.ClearConstant(constMapping), Parameter, TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());
            
        //    return timers;
        //}


        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = FirstProcess.SetTimer(timerID, clocks);

            // Get the parameter name
            string parameter_name = Parameter.ToString();

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddParInequality(TimerID, parameter_name);
                    return new PARWithinProcess(process, Parameter, TimerID);
                }

                clocks.AddParInequality(timerID, parameter_name);
                return new PARWithinProcess(process, Parameter, timerID); // note that timerID is not TimerID
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddParInequality(TimerID, parameter_name);
                return null;
            }

            clocks.AddParInequality(timerID, parameter_name);
            return new PARWithinProcess(FirstProcess, Parameter, timerID);
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return FirstProcess.GetTopLevelConcurrency(visitedDef);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}