//#define DEBUGCONSOLE

using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class Sequence : PARProcess
    {
        public PARProcess FirstProcess;
        public PARProcess SecondProcess;

        public Sequence(PARProcess firstProcess, PARProcess secondProcess)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;

            if (FirstProcess is Skip || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
            {
                ProcessID = SecondProcess.ProcessID;
            }
            else if (FirstProcess is Stop || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Stop))
            {
                ProcessID = Constants.STOP;
            }
            else
            {
                ProcessID = PARDataStore.DataManager.InitializeProcessID(Constants.SEQUENTIAL + FirstProcess.ProcessID + Constants.SEPARATOR + SecondProcess.ProcessID);
            }
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            if (FirstProcess is Skip ||
                (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
            {
                return SecondProcess.MoveOneStep(eStep);
            }

            List<PARConfiguration> list = FirstProcess.MoveOneStep(eStep);

            //bool terminationEnabled = false;
            for (int i = 0; i < list.Count; i++)
            {
                PARConfiguration step = list[i];

                if (step.Event == Constants.TERMINATION)
                {
                    //terminationEnabled = true;
                    step.Event = Constants.TAU;
                    step.Process = SecondProcess;
                    // idleConstraint.AddInequality(newClock, TimerOperationType.Equals, 0);
                }
                else
                {
                    step.Process = new Sequence(step.Process, SecondProcess);
                }
            }

            //if (terminationEnabled)
            //{
            //    idleConstraint.AddInequality(newClock, TimerOperationType.Equals, 0);
            //}

            return list;
        }



        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            if (FirstProcess is Skip || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
            {
                return SecondProcess.SyncOutput(eStep);
            }

            List<PARConfigurationWithChannelData> list = FirstProcess.SyncOutput(eStep);

            for (int i = 0; i < list.Count; i++)
            {
                PARConfiguration step = list[i];
                step.Process = new Sequence(step.Process, this.SecondProcess);
            }

            return list;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            if (FirstProcess is Skip || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
            {
                return SecondProcess.SyncInput(eStep);
            }

            List<PARConfiguration> list = FirstProcess.SyncInput(eStep);

            for (int i = 0; i < list.Count; i++)
            {
                PARConfiguration step = list[i];
                step.Process = new Sequence(step.Process, SecondProcess);
            }
            
            return list;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList<string>(Variables, SecondProcess.GetGlobalVariables());

            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channels = FirstProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList<string>(channels, SecondProcess.GetChannels());
            return channels;
        }

        public override string ToString()
        {
            return "(" + FirstProcess.ToString() + ";" + SecondProcess.ToString() + ")";
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            if (FirstProcess is Skip || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
            {
                return SecondProcess.GetAlphabets(visitedDefinitionRefs);
            }

            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

 
        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            if (FirstProcess is Skip || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
            {
                return SecondProcess.ClearConstant(constMapping);
            }
            
            return new Sequence(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping));
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()
        //{
        //    if (FirstProcess is Skip || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
        //    {
        //        return SecondProcess.GetActiveTimer();
        //    }

        //    return FirstProcess.GetActiveTimer();
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Entering Sequence:SetTimer");
#endif

            if (FirstProcess is Skip || (FirstProcess is AtomicProcess && (FirstProcess as AtomicProcess).Process is Skip))
            {
                return SecondProcess.SetTimer(timerID, clocks);
            }

            PARProcess first = FirstProcess.SetTimer(timerID, clocks);
            if(first == null)
            {
                return null;
            }

            Sequence sequence = new Sequence(first, SecondProcess);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("About to exit Sequence:SetTimer");
#endif
            
            return sequence;
        }

        //public override bool IsTime()
        //{
        //    return FirstProcess.IsTime() || SecondProcess.IsTime();
        //}
    }
}