using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class TimeOutProcess : PARProcess
    {
        public byte TimerID;
        public PARProcess FirstProcess;
        public PARProcess SecondProcess;
        public Expression DelayTime;

        public TimeOutProcess(PARProcess firstProcess, PARProcess secondProcess, Expression delay, byte timerID)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;
            DelayTime = delay;
            TimerID = timerID;

            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.TIMEOUT + "[" + DelayTime.ExpressionID + "]" + SecondProcess.ProcessID);
        }


        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            //get the int constant delay value
            int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;

            eStep.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value);

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event == Constants.TAU)
                {
                    vm.Process = new TimeOutProcess(vm.Process, SecondProcess, DelayTime, TimerID);
                }
            }

            Constraint constraint = eStep.TheConstraint.Clone();
            constraint.AddInequality(TimerID, TimerOperationType.Equals, value);
            steps.Add(new PARConfiguration(SecondProcess, Constants.TAU, "[timeout[" + value + "]]", eStep.GlobalEnv, false, false, constraint));
            return steps;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
             return FirstProcess.SyncOutput(eStep);
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            return FirstProcess.SyncInput(eStep);
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, SecondProcess.GetGlobalVariables());
            Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channel = FirstProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList<string>(channel, SecondProcess.GetChannels());

            return channel;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + FirstProcess + " timeout[" + DelayTime + "] " + SecondProcess + ")";
            }
            else
            {
                return "(" + FirstProcess + " timeout[" + DelayTime + "]" + Constraint.ClockToString(TimerID) + " " + SecondProcess + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new TimeOutProcess(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping), DelayTime.ClearConstant(constMapping), TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer() //Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);    
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());//eStep));
        //    Common.Classes.Ultility.Ultility.Union(timers, SecondProcess.GetActiveTimer());//eStep));

        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = FirstProcess.SetTimer(timerID, clocks);

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddInequality(TimerID, DelayTime);
                    return new TimeOutProcess(process, SecondProcess, DelayTime, TimerID);
                }

                clocks.AddInequality(timerID, DelayTime);
                return new TimeOutProcess(process, SecondProcess, DelayTime, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddInequality(TimerID, DelayTime);
                return null;
            }

            clocks.AddInequality(timerID, DelayTime);
            return new TimeOutProcess(FirstProcess, SecondProcess, DelayTime, timerID);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}