using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class TimeOutRangeProcess : PARProcess
    {
        public byte TimerID;
        public PARProcess FirstProcess;
        public PARProcess SecondProcess;
        public Expression FromTime;
        public Expression ToTime;

        public TimeOutRangeProcess(PARProcess firstProcess, PARProcess secondProcess, Expression from, Expression to, byte timerID)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;
            FromTime = from;
            ToTime = to;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.TIMEOUT + TimerID + "[" + FromTime.ExpressionID + "," + ToTime.ExpressionID + "]" + SecondProcess.ProcessID);
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            //get the int constant delay value
            int value1 = (EvaluatorDenotational.Evaluate(FromTime, eStep.GlobalEnv) as IntConstant).Value;
            int value2 = (EvaluatorDenotational.Evaluate(ToTime, eStep.GlobalEnv) as IntConstant).Value;

            //idleConstraint.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value2);

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event == Constants.TAU)
                {
                    vm.Process = new TimeOutRangeProcess(vm.Process, SecondProcess, FromTime, ToTime, TimerID);
                }
            }

            Constraint constraint = eStep.TheConstraint.Clone();
            constraint.AddInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, value1);
            steps.Add(new PARConfiguration(SecondProcess, Constants.TAU, "[timeout[" + value1 + "," + value2 + "]]", eStep.GlobalEnv, false, false, constraint));
            return steps;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            int value1 = (EvaluatorDenotational.Evaluate(FromTime, eStep.GlobalEnv) as IntConstant).Value;

            //List<RTSConfigurationWithChannelData> toReturn = new List<RTSConfigurationWithChannelData>();
            List<PARConfigurationWithChannelData> steps = FirstProcess.SyncOutput(eStep);

            foreach (PARConfigurationWithChannelData vm in steps)
            {
                //RTSValuation newEnv = vm.GlobalEnv.GetDBMClone();
                //newEnv.DBM.AddConstraint(TimerID, TimerOperationType.LessThanOrEqualTo, value2);
                //vm.GlobalEnv = newEnv;    
                Constraint newConstraint = vm.TheConstraint.Clone();
                newConstraint.AddInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, value1);
                vm.TheConstraint = newConstraint;
                //toReturn.Add(vm);
            }

            return steps; 
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            //get the int constant delay value
            int value1 = (EvaluatorDenotational.Evaluate(FromTime, eStep.GlobalEnv) as IntConstant).Value;

            //List<RTSConfiguration> toReturn = new List<RTSConfiguration>();
            List<PARConfiguration> steps = FirstProcess.SyncInput(eStep);

            foreach (PARConfiguration vm in steps)
            {
                Constraint newConstraint = vm.TheConstraint.Clone();
                newConstraint.AddInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, value1);
                vm.TheConstraint = newConstraint;
                //toReturn.Add(vm);
            }

            return steps;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, SecondProcess.GetGlobalVariables());
            Common.Classes.Ultility.Ultility.AddList(Variables, FromTime.GetVars());
            Common.Classes.Ultility.Ultility.AddList(Variables, ToTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channel = FirstProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList(channel, SecondProcess.GetChannels());

            return channel;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + FirstProcess + " timeout[" + FromTime + "," + ToTime + "] " + SecondProcess.ToString() + ")";
            }
            else
            {
                return "(" + FirstProcess + " timeout[" + FromTime + "," + ToTime + "]" + Constraint.ClockToString(TimerID) + " " + SecondProcess.ToString() + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new TimeOutRangeProcess(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping), FromTime.ClearConstant(constMapping), ToTime.ClearConstant(constMapping), TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        public override List<byte> GetActiveTimer()//Configuration eStep)
        {
            List<byte> timers = new List<byte>();
            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                timers.Add(TimerID);    
            }

            Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());//eStep));
            Common.Classes.Ultility.Ultility.Union(timers, SecondProcess.GetActiveTimer());//eStep));

            return timers;
        }

        public override PARProcess SetTimer(byte timerID, PARConfiguration eStep)
        {
            PARProcess process = FirstProcess.SetTimer(timerID, eStep);
 
            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    return new TimeOutRangeProcess(process, SecondProcess, FromTime, ToTime, TimerID);
                }

                return new TimeOutRangeProcess(process, SecondProcess, FromTime, ToTime, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                return null;
            }

            return new TimeOutRangeProcess(FirstProcess, SecondProcess, FromTime, ToTime, timerID);
        }

        public override bool IsTime()
        {
            return true;
        }
    }
}