﻿using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class WaitRangeProcess : PARProcess
    {
        public byte TimerID;
        public Expression FromTime;
        public Expression ToTime;

        public WaitRangeProcess(Expression from, Expression to, byte timerID)
        {
            FromTime = from;
            ToTime = to;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(Constants.WAIT + TimerID + "[" + FromTime.ExpressionID + "," + ToTime.ExpressionID + "]");
        }




        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            //get the int constant delay value
            int value1 = (EvaluatorDenotational.Evaluate(FromTime, eStep.GlobalEnv) as IntConstant).Value;
            int value2 = (EvaluatorDenotational.Evaluate(ToTime, eStep.GlobalEnv) as IntConstant).Value;

            List<PARConfiguration> toReturn = new List<PARConfiguration>(1);
            // idleConstraint.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value2);

            Constraint newConstraint = eStep.TheConstraint.Clone();
            newConstraint.AddInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, value1);

            PARConfiguration step = new PARConfiguration(new Skip(), Constants.TAU, "[Wait[" + value1 + "," + value2 + "]]", eStep.GlobalEnv, false, false, newConstraint);
            toReturn.Add(step);
            return toReturn;
        }




        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FromTime.GetVars();
            Common.Classes.Ultility.Ultility.AddList(Variables, ToTime.GetVars());
            return Variables;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "Wait[" + FromTime + "," + ToTime + "]";
            }
            else
            {
                return "Wait[" + FromTime + "," + ToTime + "]" + Constraint.ClockToString(TimerID);
            }
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new WaitRangeProcess(FromTime.ClearConstant(constMapping), ToTime.ClearConstant(constMapping), TimerID);
        }

        public override List<byte> GetActiveTimer()//Configuration eStep)
        {
            List<byte> timers = new List<byte>(1);
            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                timers.Add(TimerID);
            }

            return timers;
        }

        public override PARProcess SetTimer(byte timerID, PARConfiguration eStep)
        {
            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                return null;
            }

            return new WaitRangeProcess(FromTime, ToTime, timerID);
        }

        public override bool IsTime()
        {
            return true;
        }
    }
}